package daruma.bus.processor.csv;

import java.io.File;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xerces.util.XMLChar;

/**
 * DARUMAのスキーマを定義するクラスです
 * 
 * @author shumi
 * 
 */
public class MispSchema implements Cloneable {

	/**
	 * ロガー
	 */
	static private Log log = LogFactory.getLog(MispSchema.class);

	/**
	 * 改行コード
	 */
	public final static String CRLF = "\n"; //$NON-NLS-1$

	/*
	 * 下記のフラグはDaRuMaスキーマの特有機能
	 */
	/**
	 * FeatureTypeは必要項目かどうかのフラグ
	 */
	private static boolean FLG_FEATURETYPE_MUST = true;

	/**
	 * スキーマIDのデータ型はxs:IDかxs:anyURIのフラグ
	 */
	private static boolean FLG_ID_ANYURITYPE = true;

	/**
	 * インポートされるスキーマの名前空間接頭辞は必要かどうかのフラグ
	 */
	private static boolean FLG_IMPORT_NEED_PREFIX = false;

	/**
	 * スキーマid
	 */
	private String id = ""; //$NON-NLS-1$

	/**
	 * 名前空間リスト
	 */
	private List namespaceList = new ArrayList();

	/**
	 * ターゲット名前空間
	 */
	private String targetNamespace = ""; //$NON-NLS-1$

	/**
	 * 注釈
	 */
	private String annotation = ""; //$NON-NLS-1$

	/**
	 * インポートリスト
	 */
	private List importList = new ArrayList();

	/**
	 * 複合型リスト
	 */
	private List complexTypeList = new ArrayList();

	/**
	 * 大域要素リスト
	 */
	private List rootElementList = new ArrayList();

	/**
	 * 大域要素リスト
	 */
	private List rootAttributeList = new ArrayList();

	/**
	 * スキーマのファイルパス
	 */
	private String filePath = ""; //$NON-NLS-1$

	/**
	 * スキーマのデフォルトキャラクタセット
	 */
	private String dfsCharSet = ""; //$NON-NLS-1$

	/**
	 * スキーマのコンストラクタ
	 * 
	 */
	public MispSchema() {
		/*
		 * DaRuMaスキーマでは１番目の大域要素が必要
		 */
		MispSchemaElement rootEm = addRootElement();
		rootEm.flgFeature = true;
	}

	/**
	 * DARUMAスキーマの名前空間を定義するクラス
	 * 
	 * @author shumi
	 * 
	 */
	public class MispSchemaNamespace implements Cloneable {
		/**
		 * 名前空間接頭辞
		 */
		private String prefix = ""; //$NON-NLS-1$

		/**
		 * 名前空間URI
		 */
		private String uri = ""; //$NON-NLS-1$

		/**
		 * 名前空間のインデックス(内部用)
		 */
		private int index = -1;

		/**
		 * 名前空間の検証を行う
		 * 
		 * @param blFirst
		 *            全エラーを返すかか最初のエラーのみを返すフラグ
		 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
		 */
		public String validate(boolean blFirst) {
			String validate = ""; //$NON-NLS-1$

			validate += validatePrefix(blFirst);
			if ((validate.length() > 0) && blFirst)
				return validate;

			validate += validateUri(blFirst);
			if ((validate.length() > 0) && blFirst)
				return validate;

			return validate;
		}

		/**
		 * 名前空間接頭辞の検証を行う
		 * 
		 * @param blFirst
		 *            全エラーを返すかか最初のエラーのみを返すフラグ
		 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
		 */
		public String validatePrefix(boolean blFirst) {
			String validate = ""; //$NON-NLS-1$

			if (prefix.length() <= 0)
				validate += MispSchemaMessages.getString("MispSchema.1") + CRLF; //$NON-NLS-1$
			if ((validate.length() > 0) && blFirst)
				return validate;

			validate += chkXsQName(prefix);
			if ((validate.length() > 0) && blFirst)
				return validate;

			return validate;
		}

		/**
		 * 名前空間URIの検証を行う
		 * 
		 * @param blFirst
		 *            全エラーを返すかか最初のエラーのみを返すフラグ
		 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
		 */
		public String validateUri(boolean blFirst) {
			String validate = ""; //$NON-NLS-1$

			if (uri.length() <= 0)
				validate += MispSchemaMessages.getString("MispSchema.2") + CRLF; //$NON-NLS-1$
			if ((validate.length() > 0) && blFirst)
				return validate;

			validate += chkXsAnyURI(uri);
			if ((validate.length() > 0) && blFirst)
				return validate;

			return validate;
		}

		/**
		 * インデックスを返す
		 * 
		 * @return インデックス
		 */
		public int getIndex() {
			return index;
		}

		/**
		 * 名前空間接頭辞を取得する
		 * 
		 * @return 名前空間接頭辞
		 */
		public String getPrefix() {
			return prefix;
		}

		/**
		 * 名前空間接頭辞を設定する
		 * 
		 * @param prefix
		 *            設定する名前空間接頭辞
		 */
		public void setPrefix(String prefix) {
			this.prefix = prefix;
		}

		/**
		 * 名前空間接頭辞を設定する、スキーマの中に該当名前空間接頭辞への参照も更新する
		 * 
		 * @param prefix
		 *            設定する名前空間接頭辞
		 */
		public void setPrefixReference(String prefix) {
			String oldPrefix = this.prefix;
			this.prefix = prefix;

			if ((oldPrefix.length() > 0) && (prefix.length() > 0)) {
				for (int cnt = 0; cnt < complexTypeList.size(); cnt++) {
					for (int cnti = 0; cnti < complexTypeList.size(); cnti++) {
						MispSchemaComplexType ct = (MispSchemaComplexType) complexTypeList
								.get(cnti);
						for (int cntj = 0; cntj < ct.attributeList.size(); cntj++) {
							MispSchemaAttribute attr = (MispSchemaAttribute) ct.attributeList
									.get(cntj);
							if (attr.getNamePrefix().equals(oldPrefix))
								attr.setNamePrefix(prefix);
							if (attr.getTypePrefix().equals(oldPrefix))
								attr.setTypePrefix(prefix);
						}

						for (int cntj = 0; cntj < ct.elementList.size(); cntj++) {
							MispSchemaElement em = (MispSchemaElement) ct.elementList
									.get(cntj);
							if (em.getNamePrefix().equals(oldPrefix))
								em.setNamePrefix(prefix);
							if (em.getTypePrefix().equals(oldPrefix))
								em.setTypePrefix(prefix);
						}
					}
					MispSchemaComplexType ct2 = (MispSchemaComplexType) complexTypeList
							.get(cnt);
					if (ct2.getNamePrefix().equals(oldPrefix)) {
						ct2.setNamePrefix(prefix);
					}
				}
			}
		}

		/**
		 * 名前空間URIを取得する
		 * 
		 * @return 名前空間URI
		 */
		public String getUri() {
			return uri;
		}

		/**
		 * 名前空間URIを設定する
		 * 
		 * @param uri
		 *            設定する名前空間URI
		 */
		public void setUri(String uri) {
			this.uri = uri;
		}

		/**
		 * 名前空間をコピーする
		 * 
		 * @param ns
		 *            コピー先の名前空間
		 */
		public void copyTo(MispSchemaNamespace ns) {
			ns.prefix = new String(this.prefix);
			ns.uri = new String(this.uri);
		}

		/**
		 * 名前空間が同じかどうかを比較する
		 * 
		 * @return 比較された名前空間
		 */
		public boolean equals(Object obj) {
			if (obj == null)
				return false;
			if (!(obj instanceof MispSchemaNamespace))
				return false;
			MispSchemaNamespace ns = (MispSchemaNamespace) obj;
			if (!ns.prefix.equals(this.prefix))
				return false;
			if (!ns.uri.equals(this.uri))
				return false;
			return true;
		}

		/**
		 * 文字列を返す
		 * 
		 * @return 文字列
		 */
		public String toString() {
			return indexStr(index + 1) + ":" + this.uri; //$NON-NLS-1$
		}
	}

	/**
	 * DARUMAスキーマのインポートを定義するクラス
	 * 
	 * @author shumi
	 * 
	 */
	public class MispSchemaImport implements Cloneable {
		/**
		 * インポートするスキーマの名前空間URI
		 */
		private String uri = ""; //$NON-NLS-1$

		/**
		 * インポートするスキーマのロケーション
		 */
		private String location = ""; //$NON-NLS-1$

		/**
		 * XML:includeフラグ
		 */
		private boolean includeFlg = false;

		/**
		 * インポートファイルの読込む結果
		 */
		private String loadResult = MispSchemaMessages
				.getString("MispSchema.82"); //$NON-NLS-1$

		/**
		 * インデックス(内部用)
		 */
		private int index = -1;

		/**
		 * インポートするスキーマの検証を行う
		 * 
		 * @param blFirst
		 *            全エラーを返すかか最初のエラーのみを返すフラグ
		 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
		 */
		public String validate(boolean blFirst) {
			String validate = ""; //$NON-NLS-1$

			validate += validateUri(blFirst);
			if ((validate.length() > 0) && blFirst)
				return validate;

			validate += validateLocation(blFirst);
			if ((validate.length() > 0) && blFirst)
				return validate;

			return validate;
		}

		/**
		 * インポートするスキーマの名前空間URIの検証を行う
		 * 
		 * @param blFirst
		 *            全エラーを返すかか最初のエラーのみを返すフラグ
		 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
		 */
		public String validateUri(boolean blFirst) {
			String validate = ""; //$NON-NLS-1$

			if (uri.length() <= 0)
				validate += MispSchemaMessages.getString("MispSchema.0") + CRLF; //$NON-NLS-1$
			if ((validate.length() > 0) && blFirst)
				return validate;

			validate += chkXsAnyURI(uri);
			if ((validate.length() > 0) && blFirst)
				return validate;

			if (FLG_IMPORT_NEED_PREFIX) {
				if (!existsNamespace(uri))
					validate += MispSchemaMessages.getString("MispSchema.3"); //$NON-NLS-1$
				if ((validate.length() > 0) && blFirst)
					return validate;
			}

			return validate;
		}

		/**
		 * インポートするスキーマのロケーションの検証を行う
		 * 
		 * @param blFirst
		 *            全エラーを返すかか最初のエラーのみを返すフラグ
		 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
		 */
		public String validateLocation(boolean blFirst) {
			String validate = ""; //$NON-NLS-1$

			if (location.length() > 0) {
				/*
				 * validate += chkXsAnyURI(location); if ((validate.length() >
				 * 0) && blFirst) return validate;
				 */
			}

			if (loadResult.length() > 0) {
				validate += loadResult;
				if ((validate.length() > 0) && blFirst)
					return validate;
			}

			return validate;
		}

		/**
		 * インデックスを返す
		 * 
		 * @return インデックス
		 */
		public int getIndex() {
			return index;
		}

		/**
		 * インポートするスキーマの名前空間URIを取得する
		 * 
		 * @return インポートするスキーマの名前空間URI
		 */
		public String getLocation() {
			return location;
		}

		/**
		 * インポートするスキーマの名前空間URIを設定する
		 * 
		 * @param location
		 *            インポートするスキーマの名前空間URI
		 */
		public void setLocation(String location) {
			this.location = location.replaceAll("\\\\", "/"); //$NON-NLS-1$ //$NON-NLS-2$
		}

		/**
		 * includeFlgを取得する
		 * 
		 * @return Returns the includeFlg.
		 */
		public boolean isIncludeFlg() {
			return includeFlg;
		}

		/**
		 * includeFlgをセットする
		 * 
		 * @param includeFlg
		 *            The includeFlg to set.
		 */
		public void setIncludeFlg(boolean includeFlg) {
			this.includeFlg = includeFlg;
		}

		/**
		 * インポートするスキーマの名前空間URIを取得する
		 * 
		 * @return インポートするスキーマの名前空間URI
		 */
		public String getUri() {
			return uri;
		}

		/**
		 * インポートするスキーマの名前空間URIを設定する
		 * 
		 * @param uri
		 *            インポートするスキーマの名前空間URI
		 */
		public void setUri(String uri) {
			this.uri = uri;
		}

		/**
		 * loadResultを取得する
		 * 
		 * @return Returns the loadResult.
		 */
		public String getLoadResult() {
			return loadResult;
		}

		/**
		 * loadResultをセットする
		 * 
		 * @param loadResult
		 *            The loadResult to set.
		 */
		public void setLoadResult(String loadResult) {
			this.loadResult = loadResult;
		}

		/**
		 * スキーマをコピーする
		 * 
		 * @param im
		 *            コピー先のスキーマ
		 */
		public void copyTo(MispSchemaImport im) {
			im.uri = this.uri;
			im.location = this.location;
			im.includeFlg = this.includeFlg;
			im.loadResult = this.loadResult;
		}

		/**
		 * スキーマが同じかどうかを比較する
		 * 
		 * @return 比較されたスキーマ
		 */
		public boolean equals(Object obj) {
			if (obj == null)
				return false;
			if (!(obj instanceof MispSchemaImport))
				return false;
			MispSchemaImport im = (MispSchemaImport) obj;
			if (!im.uri.equals(this.uri))
				return false;
			if (!im.location.equals(this.location))
				return false;
			return true;
		}

		/**
		 * 文字列を返す
		 * 
		 * @return 文字列
		 */
		public String toString() {
			return indexStr(index + 1) + ":" + this.uri; //$NON-NLS-1$
		}
	}

	/**
	 * DARUMAスキーマの複合型を定義するクラス
	 * 
	 * @author shumi
	 * 
	 */
	public class MispSchemaComplexType implements Cloneable {
		/**
		 * 複合型名
		 */
		private String name = ""; //$NON-NLS-1$

		/**
		 * 派生型
		 */
		private String extension = ""; //$NON-NLS-1$

		/**
		 * 属性リスト
		 */
		private List attributeList = new ArrayList();

		/**
		 * 要素リスト
		 */
		private List elementList = new ArrayList();

		/**
		 * ファイルパス
		 */
		private String filePath = ""; //$NON-NLS-1$

		/**
		 * インポート(コピーされない)
		 */
		MispSchemaImport im = null;

		/**
		 * kind(コピーされない)
		 */
		private int kind = KIND_INTERNAL;

		/**
		 * スキーマ(内部用)
		 */
		private MispSchema schema;

		/**
		 * インデックス(内部用)
		 */
		private int index = -1;

		/**
		 * 複合型の検証を行う
		 * 
		 * @param blFirst
		 *            全エラーを返すかか最初のエラーのみを返すフラグ
		 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
		 */
		public String validate(boolean blFirst) {
			String validate = ""; //$NON-NLS-1$

			if (im == null) {
				validate += validateName(blFirst);
				if ((validate.length() > 0) && blFirst)
					return validate;

				validate += validateExtension(blFirst);
				if ((validate.length() > 0) && blFirst)
					return validate;

				validate += validateAttributeList(blFirst);
				if ((validate.length() > 0) && blFirst)
					return validate;

				validate += validateElementList(blFirst);
				if ((validate.length() > 0) && blFirst)
					return validate;
			}

			return validate;
		}

		/**
		 * filePathを取得する
		 * 
		 * @return Returns the filePath.
		 */
		public String getFilePath() {
			if (filePath.length() == 0)
				return schema.getFilePath();
			return filePath;
		}

		/**
		 * filePathをセットする
		 * 
		 * @param filePath
		 *            The filePath to set.
		 */
		public void setFilePath(String filePath) {
			this.filePath = filePath;
		}

		/**
		 * 複合型名のprefixを取得する
		 * 
		 * @return Returns the prefix.
		 */
		public String getNamePrefix() {
			return MispSchema.getPrefix(this.name);
		}

		/**
		 * 複合型名のprefixをセットする
		 * 
		 * @param prefix
		 *            The prefix to set.
		 */
		public void setNamePrefix(String prefix) {
			this.name = MispSchema.setPrefix(this.name, prefix);
		}

		/**
		 * prefixをセットする
		 * 
		 * @param prefix
		 *            The prefix to set.
		 */
		public void setPrefixAll(String prefix) {
			this.name = MispSchema.setPrefix(this.name, prefix);
			if (this.extension.length() > 0)
				this.extension = MispSchema.setPrefix(this.extension, prefix);
			for (int cnt = 0; cnt < attributeList.size(); cnt++) {
				MispSchemaAttribute attr = (MispSchemaAttribute) attributeList
						.get(cnt);
				if (attr.getNamePrefix().length() == 0)
					attr.setNamePrefix(prefix);
				if (attr.getTypePrefix().length() == 0)
					attr.setTypePrefix(prefix);
			}
			for (int cnt = 0; cnt < elementList.size(); cnt++) {
				MispSchemaElement em = (MispSchemaElement) elementList.get(cnt);
				if (em.getNamePrefix().length() == 0)
					em.setNamePrefix(prefix);
				if (em.getTypePrefix().length() == 0)
					em.setTypePrefix(prefix);
			}
		}

		/**
		 * 複合型名の検証を行う
		 * 
		 * @param blFirst
		 *            全エラーを返すかか最初のエラーのみを返すフラグ
		 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
		 */
		public String validateName(boolean blFirst) {
			String validate = ""; //$NON-NLS-1$

			if (name.length() <= 0)
				validate += MispSchemaMessages.getString("MispSchema.4") + CRLF; //$NON-NLS-1$
			if ((validate.length() > 0) && blFirst)
				return validate;

			validate += chkXsQName(name);
			if ((validate.length() > 0) && blFirst)
				return validate;

			for (int cnt = 0; cnt < complexTypeList.size(); cnt++) {
				if (cnt != index) {
					if (((MispSchemaComplexType) complexTypeList.get(cnt))
							.getName().equals(name)) {
						validate += MispSchemaMessages
								.getString("MispSchema.5") + CRLF; //$NON-NLS-1$
					}
					if ((validate.length() > 0) && blFirst)
						return validate;
				}
			}

			return validate;
		}

		/**
		 * 派生型の検証を行う
		 * 
		 * @param blFirst
		 *            全エラーを返すかか最初のエラーのみを返すフラグ
		 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
		 */
		public String validateExtension(boolean blFirst) {
			String validate = ""; //$NON-NLS-1$

			if (extension.length() > 0) {
				validate += chkXsQName(name);
				if ((validate.length() > 0) && blFirst)
					return validate;
			}

			return validate;
		}

		/**
		 * 属性リストの検証を行う
		 * 
		 * @param blFirst
		 *            全エラーを返すかか最初のエラーのみを返すフラグ
		 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
		 */
		public String validateAttributeList(boolean blFirst) {
			return MispSchemaUtil.validateList(attributeList, blFirst);
		}

		/**
		 * 要素リストの検証を行う
		 * 
		 * @param blFirst
		 *            全エラーを返すかか最初のエラーのみを返すフラグ
		 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
		 */
		public String validateElementList(boolean blFirst) {
			return MispSchemaUtil.validateList(elementList, blFirst);
		}

		/**
		 * インデックスを返す
		 * 
		 * @return インデックス
		 */
		public int getIndex() {
			return index;
		}

		/**
		 * 属性リストサイズを取得する
		 * 
		 * @return 属性リストサイズ
		 */
		public int getAttributeSize() {
			return attributeList.size();
		}

		/**
		 * 属性配列を取得する
		 * 
		 * @return 属性配列
		 */
		public Object[] getAttributeArray() {
			return attributeList.toArray();
		}

		/**
		 * 指定されたインデックスの属性を取得する
		 * 
		 * @return 属性
		 */
		public MispSchemaAttribute getAttribute(int index) {
			return (MispSchemaAttribute) attributeList.get(index);
		}

		/**
		 * 指定された名前の属性を取得する
		 * 
		 * @return 属性の配列
		 */
		public Object[] findAttributes(String name) {
			List rc = new ArrayList();

			for (int cnt = 0; cnt < attributeList.size(); cnt++) {
				MispSchemaAttribute attr = (MispSchemaAttribute) attributeList
						.get(cnt);
				if (attr.name.equals(name)) {
					rc.add(attr);
				}
			}
			return rc.toArray();
		}

		/**
		 * 属性をリストに追加する
		 * 
		 * @param index
		 *            追加位置
		 * @return 追加された属性
		 */
		public MispSchemaAttribute addAttribute(int index) {
			MispSchemaAttribute attr = new MispSchemaAttribute();
			attr.index = attributeList.size();
			attr.flgRoot = false;
			attr.parentCt = this;

			if ((index >= 0) && (index < attributeList.size())) {
				attributeList.add(index, attr);
				updateAttributeIndex(0);
			} else
				attributeList.add(attr);
			return attr;
		}

		/**
		 * 属性をリストから削除する
		 * 
		 * @param index
		 *            削除する属性のインデックス
		 */
		public void delAttribute(int index) {
			if ((index >= 0) && (index < attributeList.size())) {
				attributeList.remove(index);
				updateAttributeIndex(index);
			}
		}

		/**
		 * 属性をリストから削除する
		 * 
		 * @param attr
		 *            削除する属性
		 */
		public void delAttribute(MispSchemaAttribute attr) {
			delAttribute(attributeList.indexOf(attr));
		}

		/**
		 * 属性を移動する
		 * 
		 * @param attr
		 *            移動する属性
		 * @param delta
		 *            移動量
		 */
		public void moveAttribute(MispSchemaAttribute attr, int delta) {
			int oldIndex = attributeList.indexOf(attr);

			if ((oldIndex >= 0) && (oldIndex < attributeList.size())) {
				int newIndex = oldIndex + delta;
				if (newIndex < 0)
					newIndex = 0;
				if (newIndex > attributeList.size())
					newIndex = attributeList.size();
				if ((newIndex >= 0) && (newIndex < attributeList.size())) {
					if (oldIndex != newIndex) {
						Object org = attributeList.get(newIndex);
						attributeList.set(newIndex, attr);
						attributeList.set(oldIndex, org);
						updateAttributeIndex(0);
					}

				}
			}
		}

		/**
		 * 属性リストのインデックスを更新
		 * 
		 * @param start
		 *            更新する属性の開始インデックス
		 */
		private void updateAttributeIndex(int start) {
			if (start < 0)
				start = 0;
			for (int cnt = start; cnt < attributeList.size(); cnt++) {
				((MispSchemaAttribute) attributeList.get(cnt)).index = cnt;
			}
		}

		/**
		 * 要素リストサイズを取得する
		 * 
		 * @return 要素リストサイズ
		 */
		public int getElementSize() {
			return elementList.size();
		}

		/**
		 * 要素配列を取得する
		 * 
		 * @return 要素配列
		 */
		public Object[] getElementArray() {
			return elementList.toArray();
		}

		/**
		 * 指定されたインデックスの要素を取得する
		 * 
		 * @return 要素
		 */
		public MispSchemaElement getElement(int index) {
			return (MispSchemaElement) elementList.get(index);
		}

		/**
		 * 指定された名前の属性を取得する
		 * 
		 * @return 属性の配列
		 */
		public Object[] findElements(String name) {
			List rc = new ArrayList();
			for (int cnt = 0; cnt < elementList.size(); cnt++) {
				MispSchemaElement em = (MispSchemaElement) elementList.get(cnt);
				if (em.name.equals(name))
					rc.add(em);
			}
			return rc.toArray();
		}

		/**
		 * 要素をリストに追加する
		 * 
		 * @param index
		 *            追加位置
		 * @return 追加する要素
		 */
		public MispSchemaElement addElement(int index) {
			MispSchemaElement em = new MispSchemaElement();
			em.index = elementList.size();
			em.flgRoot = false;
			em.parentCt = this;
			if ((index >= 0) && (index < elementList.size())) {
				elementList.add(index, em);
				updateElementIndex(0);
			} else
				elementList.add(em);
			return em;
		}

		/**
		 * 要素を移動する
		 * 
		 * @param em
		 *            移動する要素
		 * @param delta
		 *            移動量
		 */
		public void moveElement(MispSchemaElement em, int delta) {
			int oldIndex = elementList.indexOf(em);

			if ((oldIndex >= 0) && (oldIndex < elementList.size())) {
				int newIndex = oldIndex + delta;
				if (newIndex < 0)
					newIndex = 0;
				if (newIndex > elementList.size())
					newIndex = elementList.size();
				if ((newIndex >= 0) && (newIndex < elementList.size())) {
					if (oldIndex != newIndex) {
						Object org = elementList.get(newIndex);
						elementList.set(newIndex, em);
						elementList.set(oldIndex, org);
						updateElementIndex(0);
					}

				}
			}
		}

		/**
		 * 要素をリストから削除する
		 * 
		 * @param index
		 *            削除する要素のインデックス
		 */
		public void delElement(int index) {
			if ((index >= 0) && (index < elementList.size())) {
				elementList.remove(index);
				updateElementIndex(index);
			}
		}

		/**
		 * 要素をリストから削除する
		 * 
		 * @param em
		 *            削除する要素
		 */
		public void delElement(MispSchemaElement em) {
			delElement(elementList.indexOf(em));
		}

		/**
		 * 要素リストのインデックスを更新
		 * 
		 * @param start
		 *            更新する要素の開始インデックス
		 */
		private void updateElementIndex(int start) {
			if (start < 0)
				start = 0;
			for (int cnt = start; cnt < elementList.size(); cnt++) {
				((MispSchemaElement) elementList.get(cnt)).index = cnt;
			}
		}

		/**
		 * 属性、要素リストを取得する
		 * 
		 * @param flgExt
		 *            基底型的属性、要素取得フラグ
		 * @param flgNormal
		 *            複合型の属性、要素取得フラグ
		 * @param flgAttr
		 *            属性取得フラグ
		 * @param flgEm
		 *            要素取得フラグ
		 * @return 属性、要素リスト
		 */
		public List getAttributeElementList(boolean flgExt, boolean flgNormal,
				boolean flgAttr, boolean flgEm) {
			List attributeElementList = new ArrayList();
			MispSchemaComplexType extCt = null;

			if (this.extension.length() > 0)
				extCt = findComplexTypeByName(extension);

			if (flgExt && (extCt != null) && flgAttr)
				attributeElementList.addAll(extCt.getAttributeElementList(true,
						true, true, false));

			if (flgNormal && flgAttr)
				attributeElementList.addAll(attributeList);

			if (flgExt && (extCt != null) && flgEm)
				attributeElementList.addAll(extCt.getAttributeElementList(true,
						true, false, true));

			if (flgNormal && flgEm)
				attributeElementList.addAll(elementList);
			return attributeElementList;
		}

		/**
		 * 属性、要素サイズ配列を取得する
		 * 
		 * @param flgExt
		 *            基底型的属性、要素取得フラグ
		 * @param flgNormal
		 *            複合型の属性、要素取得フラグ
		 * @param flgAttr
		 *            属性取得フラグ
		 * @param flgEm
		 *            要素取得フラグ
		 * @return 属性、要素のサイズ
		 */
		public int getAttributeElementSize(boolean flgExt, boolean flgNormal,
				boolean flgAttr, boolean flgEm) {
			return getAttributeElementList(flgExt, flgNormal, flgAttr, flgEm)
					.size();
		}

		/**
		 * 派生型を取得する
		 * 
		 * @return 派生型
		 */
		public String getExtension() {
			return extension;
		}

		/**
		 * 派生型を設定する
		 * 
		 * @param extension
		 *            派生型
		 */
		public void setExtension(String extension) {
			this.extension = extension;
		}

		/**
		 * 複合型名を取得する
		 * 
		 * @return 複合型名
		 */
		public String getName() {
			return name;
		}

		/**
		 * 複合型名を設定する
		 * 
		 * @param name
		 *            複合型名
		 */
		public void setName(String name) {
			this.name = name;
		}

		/**
		 * imを取得する
		 * 
		 * @return Returns the im.
		 */
		public MispSchemaImport getIm() {
			return im;
		}

		/**
		 * kindを取得する
		 * 
		 * @return Returns the kind.
		 */
		public int getKind() {
			return kind;
		}

		/**
		 * 複合型名を設定する、スキーマの中に該当名前空間接頭辞への参照も更新する
		 * 
		 * @param newName
		 *            複合型名
		 */
		public void setNameUpdateReference(String newName) {
			String oldName = this.name;
			this.name = newName;
			for (int cnti = 0; cnti < rootElementList.size(); cnti++) {
				MispSchemaElement em = (MispSchemaElement) rootElementList
						.get(cnti);
				if (em.type.equals(oldName))
					em.type = newName;
			}

			for (int cnti = 0; cnti < complexTypeList.size(); cnti++) {
				MispSchemaComplexType ct = (MispSchemaComplexType) complexTypeList
						.get(cnti);
				for (int cntj = 0; cntj < ct.elementList.size(); cntj++) {
					MispSchemaElement em = (MispSchemaElement) ct.elementList
							.get(cntj);
					if (em.type.equals(oldName))
						em.type = newName;
				}
			}

		}

		/**
		 * スキーマを取得する
		 * 
		 * @return Returns the schema.
		 */
		public MispSchema getSchema() {
			return schema;
		}

		/**
		 * 複合型をコピーする
		 * 
		 * @param ct
		 *            コピー先の複合型
		 */
		public void copyTo(MispSchemaComplexType ct) {
			ct.name = new String(this.name);
			ct.extension = new String(this.extension);
			ct.attributeList.clear();
			MispSchemaUtil.copyMispSchemaComplexTypeList(ct,
					this.attributeList, false);
			ct.elementList.clear();
			MispSchemaUtil.copyMispSchemaComplexTypeList(ct, this.elementList,
					false);
			ct.filePath = this.filePath;
		}

		/**
		 * 複合型をコピーする
		 * 
		 * @param ct
		 *            コピー先の複合型
		 */
		public void copyToWithoutRef(MispSchemaComplexType ct) {
			ct.name = new String(this.name);
			ct.extension = new String(this.extension);
			ct.attributeList.clear();
			MispSchemaUtil.copyMispSchemaComplexTypeList(ct,
					this.attributeList, true);
			ct.elementList.clear();
			MispSchemaUtil.copyMispSchemaComplexTypeList(ct, this.elementList,
					true);
			ct.filePath = this.filePath;
		}

		/**
		 * 複合型が同じかどうかを比較する
		 * 
		 * @return 比較された複合型
		 */
		public boolean equals(Object obj) {
			if (obj == null)
				return false;
			if (!(obj instanceof MispSchemaComplexType))
				return false;
			MispSchemaComplexType ct = (MispSchemaComplexType) obj;
			if (!ct.name.equals(this.name))
				return false;
			if (!ct.extension.equals(this.extension))
				return false;
			if (!MispSchemaUtil
					.equalsList(ct.attributeList, this.attributeList))
				return false;
			if (!MispSchemaUtil.equalsList(ct.elementList, this.elementList))
				return false;
			return true;
		}

		/**
		 * 文字列を返す
		 * 
		 * @return 文字列
		 */
		public String toString() {
			return indexStr(index + 1) + ":" + this.name; //$NON-NLS-1$
		}

	}

	/**
	 * DARUMAスキーマの属性を定義するクラス
	 * 
	 * @author shumi
	 * 
	 */
	public class MispSchemaAttribute implements Cloneable {
		/**
		 * 大域属性フラグ
		 */
		boolean flgRoot;

		/**
		 * 属性名
		 */
		String name = ""; //$NON-NLS-1$

		/**
		 * 属性のデータ型
		 */
		String type = ""; //$NON-NLS-1$

		/**
		 * Refフラグ
		 */
		boolean flgRef = false;

		/**
		 * インデックスフラグ
		 */
		boolean indexedFlg = false;

		/**
		 * 親複合型(内部用)
		 */
		MispSchemaComplexType parentCt;

		/**
		 * インデックスを返す(内部用)
		 */
		private int index = -1;

		/**
		 * 属性の検証を行う
		 * 
		 * @param blFirst
		 *            全エラーを返すかか最初のエラーのみを返すフラグ
		 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
		 */
		public String validate(boolean blFirst) {
			String validate = ""; //$NON-NLS-1$

			validate += validateName(blFirst);
			if ((validate.length() > 0) && blFirst)
				return validate;

			validate += validateType(blFirst);
			if ((validate.length() > 0) && blFirst)
				return validate;

			return validate;
		}

		/**
		 * 属性名の検証を行う
		 * 
		 * @param blFirst
		 *            全エラーを返すかか最初のエラーのみを返すフラグ
		 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
		 */
		public String validateName(boolean blFirst) {
			String validate = ""; //$NON-NLS-1$

			if (name.length() <= 0)
				validate += MispSchemaMessages.getString("MispSchema.6") + CRLF; //$NON-NLS-1$
			if ((validate.length() > 0) && blFirst)
				return validate;

			validate += chkXsQName(name);
			if ((validate.length() > 0) && blFirst)
				return validate;

			for (int cnt = 0; cnt < parentCt.attributeList.size(); cnt++) {
				if (cnt != index) {
					if (((MispSchemaAttribute) parentCt.attributeList.get(cnt))
							.getName().equals(name)) {
						validate += MispSchemaMessages
								.getString("MispSchema.7") + CRLF; //$NON-NLS-1$
						if ((validate.length() > 0) && blFirst)
							return validate;
					}
				}
			}

			return validate;
		}

		/**
		 * データ型の検証を行う
		 * 
		 * @param blFirst
		 *            全エラーを返すかか最初のエラーのみを返すフラグ
		 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
		 */
		public String validateType(boolean blFirst) {
			String validate = ""; //$NON-NLS-1$

			if (type.length() <= 0)
				validate += MispSchemaMessages.getString("MispSchema.8") + CRLF; //$NON-NLS-1$
			if ((validate.length() > 0) && blFirst)
				return validate;

			validate += chkXsQName(type);
			if ((validate.length() > 0) && blFirst)
				return validate;

			if (!isValidAttributeType(type))
				validate += MispSchemaMessages.getString("MispSchema.9") + CRLF; //$NON-NLS-1$
			if ((validate.length() > 0) && blFirst)
				return validate;

			return validate;
		}

		/**
		 * インデックスを返す
		 * 
		 * @return インデックス
		 */
		public int getIndex() {
			return index;
		}

		/**
		 * flgRefを取得する
		 * 
		 * @return Returns the flgRef.
		 */
		public boolean isFlgRef() {
			return flgRef;
		}

		/**
		 * flgRefをセットする
		 * 
		 * @param flgRef
		 *            The flgRef to set.
		 */
		public void setFlgRef(boolean flgRef) {
			this.flgRef = flgRef;
		}

		/**
		 * 属性名を取得する
		 * 
		 * @return 属性名
		 */
		public String getName() {
			if (flgRef) {
				MispSchemaAttribute attr = findRefAttributeByName(this.name);
				return (attr != null) ? attr.getName() : ""; //$NON-NLS-1$
			} else {
				return name;
			}
		}

		/**
		 * 属性名を設定する
		 * 
		 * @param name
		 *            属性名
		 */
		public void setName(String name) {
			if (flgRef) {
				// ありえない
			} else {
				this.name = name;
			}
		}

		/**
		 * prefixを取得する
		 * 
		 * @return Returns the prefix.
		 */
		public String getNamePrefix() {
			if (flgRef) {
				MispSchemaAttribute attr = findRefAttributeByName(this.name);
				return (attr != null) ? attr.getNamePrefix() : ""; //$NON-NLS-1$
			} else {
				return MispSchema.getPrefix(this.name);
			}
		}

		/**
		 * prefixをセットする
		 * 
		 * @param prefix
		 *            The prefix to set.
		 */
		public void setNamePrefix(String prefix) {
			if (flgRef) {
				// ありえない
			} else {
				this.name = MispSchema.setPrefix(this.name, prefix);
			}
		}

		/**
		 * データ型を取得する
		 * 
		 * @return データ型
		 */
		public String getType() {
			if (flgRef) {
				MispSchemaAttribute attr = findRefAttributeByName(this.name);
				return (attr != null) ? attr.getType() : ""; //$NON-NLS-1$
			} else {
				return type;
			}
		}

		/**
		 * データ型を設定する
		 * 
		 * @param type
		 *            データ型
		 */
		public void setType(String type) {
			if (flgRef) {
				// ありえない
			} else {
				this.type = type;
			}
		}

		/**
		 * prefixを取得する
		 * 
		 * @return Returns the prefix.
		 */
		public String getTypePrefix() {
			if (flgRef) {
				MispSchemaAttribute attr = findRefAttributeByName(this.name);
				return (attr != null) ? attr.getTypePrefix() : ""; //$NON-NLS-1$
			} else {
				return MispSchema.getPrefix(this.type);
			}
		}

		/**
		 * prefixをセットする
		 * 
		 * @param prefix
		 *            The prefix to set.
		 */
		public void setTypePrefix(String prefix) {
			if (flgRef) {
				// ありえない
			} else {
				this.type = MispSchema.setPrefix(this.type, prefix);
			}
		}

		/**
		 * indexedFlgを取得する
		 * 
		 * @return Returns the indexedFlg.
		 */
		public boolean isIndexedFlg() {
			return indexedFlg;
		}

		/**
		 * indexedFlgをセットする
		 * 
		 * @param indexedFlg
		 *            The indexedFlg to set.
		 */
		public void setIndexedFlg(boolean indexedFlg) {
			this.indexedFlg = indexedFlg;
		}

		/**
		 * 属性をコピーする
		 * 
		 * @param attr
		 *            コピー先の属性
		 * 
		 */
		public void copyTo(MispSchemaAttribute attr) {
			attr.flgRef = this.flgRef;
			attr.name = this.name;
			attr.type = this.type;

			attr.flgRoot = this.flgRoot;
			attr.indexedFlg = this.indexedFlg;
		}

		/**
		 * 属性をコピーする
		 * 
		 * @param attr
		 *            コピー先の属性
		 * 
		 */
		public void copyToWithoutRef(MispSchemaAttribute attr) {
			attr.flgRef = false;
			attr.name = new String(this.getName());
			attr.type = new String(this.getType());

			attr.flgRoot = this.flgRoot;
			attr.indexedFlg = this.indexedFlg;
		}

		/**
		 * 属性が同じかどうかを比較する
		 * 
		 * @return 比較された属性
		 */
		public boolean equals(Object obj) {
			if (obj == null)
				return false;
			if (!(obj instanceof MispSchemaAttribute))
				return false;
			MispSchemaAttribute attr = (MispSchemaAttribute) obj;
			if (!attr.name.equals(this.name))
				return false;
			if (!attr.type.equals(this.type))
				return false;
			if (attr.indexedFlg != this.indexedFlg)
				return false;
			return true;
		}

		/**
		 * 文字列を返す
		 * 
		 * @return 文字列
		 */
		public String toString() {
			return this.name + "/" + this.type; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		}

		/**
		 * 複合型を返す
		 * 
		 * @return 複合型
		 */
		public MispSchemaComplexType getParentCt() {
			return parentCt;
		}
	}

	/**
	 * DARUMAスキーマの名前空間を定義するクラス
	 * 
	 * @author shumi
	 * 
	 */
	public class MispSchemaElement implements Cloneable {
		/**
		 * 大域要素フラグ
		 */
		boolean flgRoot = false;

		/**
		 * フィーチャタイプフラグ
		 */
		boolean flgFeature = false;

		/**
		 * 要素名
		 */
		String name = ""; //$NON-NLS-1$

		/**
		 * データ型
		 */
		String type = ""; //$NON-NLS-1$

		/**
		 * minOccurs
		 */
		String minOccurs = "1"; //$NON-NLS-1$

		/**
		 * maxOccurs
		 */
		String maxOccurs = "1"; //$NON-NLS-1$

		/**
		 * インデックスフラグ
		 */
		boolean indexedFlg = false;

		/**
		 * 親複合型(内部用)
		 */
		MispSchemaComplexType parentCt;

		/**
		 * Refフラグ(内部用)
		 */
		boolean flgRef = false;

		/**
		 * インデックス(内部用)
		 */
		private int index = -1;

		/**
		 * 要素の検証を行う
		 * 
		 * @param blFirst
		 *            全エラーを返すかか最初のエラーのみを返すフラグ
		 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
		 */

		public String validate(boolean blFirst) {
			String validate = ""; //$NON-NLS-1$

			validate += validateName(blFirst);
			if ((validate.length() > 0) && blFirst)
				return validate;

			validate += validateType(blFirst);
			if ((validate.length() > 0) && blFirst)
				return validate;

			validate += validateMinOccurs(blFirst);
			if ((validate.length() > 0) && blFirst)
				return validate;

			validate += validateMaxOccurs(blFirst);
			if ((validate.length() > 0) && blFirst)
				return validate;

			return validate;
		}

		public String validateMaxOccurs(boolean blFirst) {
			String validate = ""; //$NON-NLS-1$

			if (name.length() <= 0)
				validate += MispSchemaMessages.getString("MispSchema.6") + CRLF; //$NON-NLS-1$
			if ((validate.length() > 0) && blFirst)
				return validate;

			if (!maxOccurs.equals("unbounded")) { //$NON-NLS-1$
				validate += chkNum(maxOccurs);
				if ((validate.length() > 0) && blFirst) {
					return validate;
				}
			}

			return validate;
		}

		public String validateMinOccurs(boolean blFirst) {
			String validate = ""; //$NON-NLS-1$

			if (name.length() <= 0)
				validate += MispSchemaMessages.getString("MispSchema.6") + CRLF; //$NON-NLS-1$
			if ((validate.length() > 0) && blFirst)
				return validate;

			validate += chkNum(minOccurs);
			if ((validate.length() > 0) && blFirst)
				return validate;

			return validate;
		}

		/**
		 * 要素名の検証を行う
		 * 
		 * @param blFirst
		 *            全エラーを返すかか最初のエラーのみを返すフラグ
		 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
		 */
		public String validateName(boolean blFirst) {
			String validate = ""; //$NON-NLS-1$

			if (flgFeature)
				return validateFeatureTypeElementName(blFirst);

			if (name.length() <= 0)
				validate += MispSchemaMessages.getString("MispSchema.11") + CRLF; //$NON-NLS-1$
			if ((validate.length() > 0) && blFirst)
				return validate;

			validate += chkXsQName(name);
			if ((validate.length() > 0) && blFirst)
				return validate;

			for (int cnt = 0; cnt < parentCt.elementList.size(); cnt++) {
				if (cnt != index) {
					if (((MispSchemaElement) parentCt.elementList.get(cnt))
							.getName().equals(name)) {
						validate += MispSchemaMessages
								.getString("MispSchema.12") + CRLF; //$NON-NLS-1$
						if ((validate.length() > 0) && blFirst)
							return validate;
					}
				}
			}

			return validate;
		}

		/**
		 * データ型の検証を行う
		 * 
		 * @param blFirst
		 *            全エラーを返すかか最初のエラーのみを返すフラグ
		 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
		 */

		public String validateType(boolean blFirst) {
			String validate = ""; //$NON-NLS-1$

			if (flgFeature)
				return validateFeatureTypeElementType(blFirst);

			if (type.length() <= 0)
				validate += MispSchemaMessages.getString("MispSchema.13") + CRLF; //$NON-NLS-1$
			if ((validate.length() > 0) && blFirst)
				return validate;

			validate += chkXsQName(type);
			if ((validate.length() > 0) && blFirst)
				return validate;

			if (type.equals(parentCt.name)) {
				validate += MispSchemaMessages.getString("MispSchema.14") + CRLF; //$NON-NLS-1$
				if ((validate.length() > 0) && blFirst)
					return validate;
			}

			if (!isValidElementType(type))
				validate += MispSchemaMessages.getString("MispSchema.15") + CRLF; //$NON-NLS-1$
			if ((validate.length() > 0) && blFirst)
				return validate;

			return validate;
		}

		/**
		 * フィーチャタイプ要素名の検証を行う
		 * 
		 * @param blFirst
		 *            全エラーを返すかか最初のエラーのみを返すフラグ
		 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
		 */
		private String validateFeatureTypeElementName(boolean blFirst) {
			String validate = ""; //$NON-NLS-1$

			if (FLG_FEATURETYPE_MUST) {
				if (name.length() <= 0)
					validate += MispSchemaMessages.getString("MispSchema.16") + CRLF; //$NON-NLS-1$
				if ((validate.length() > 0) && blFirst)
					return validate;
			}

			if (FLG_FEATURETYPE_MUST || (name.length() > 0)) {
				validate += chkXsNCName(name);
				if ((validate.length() > 0) && blFirst)
					return validate;
			}

			return validate;
		}

		/**
		 * フィーチャタイプデータ型の検証を行う
		 * 
		 * @param blFirst
		 *            全エラーを返すかか最初のエラーのみを返すフラグ
		 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
		 */
		private String validateFeatureTypeElementType(boolean blFirst) {
			String validate = ""; //$NON-NLS-1$

			if (FLG_FEATURETYPE_MUST || (name.length() > 0)) {
				if (type.length() <= 0)
					validate += MispSchemaMessages.getString("MispSchema.17") + CRLF; //$NON-NLS-1$
				if ((validate.length() > 0) && blFirst)
					return validate;

				validate += chkXsQName(type);
				if ((validate.length() > 0) && blFirst)
					return validate;

				if (!isValidFeatureElementType(type))
					validate += MispSchemaMessages.getString("MispSchema.18") + CRLF; //$NON-NLS-1$
				if ((validate.length() > 0) && blFirst)
					return validate;
			}
			return validate;
		}

		/**
		 * インデックスを返す
		 * 
		 * @return インデックス
		 */
		public int getIndex() {
			return index;
		}

		/**
		 * flgRefを取得する
		 * 
		 * @return Returns the flgRef.
		 */
		public boolean isFlgRef() {
			return flgRef;
		}

		/**
		 * flgRefをセットする
		 * 
		 * @param flgRef
		 *            The flgRef to set.
		 */
		public void setFlgRef(boolean flgRef) {
			this.flgRef = flgRef;
		}

		/**
		 * 要素名を取得する
		 * 
		 * @return 要素名
		 */
		public String getName() {
			if (flgRef) {
				MispSchemaElement em = findRefElementByName(this.name);
				return (em != null) ? em.getName() : ""; //$NON-NLS-1$
			} else {
				return name;
			}
		}

		/**
		 * 要素名を設定する
		 * 
		 * @param name
		 *            要素名
		 */
		public void setName(String name) {
			if (flgRef) {
				// ありえない
			} else {
				this.name = name;
			}
		}

		/**
		 * prefixを取得する
		 * 
		 * @return Returns the prefix.
		 */
		public String getNamePrefix() {
			if (flgRef) {
				MispSchemaElement em = findRefElementByName(this.name);
				return (em != null) ? em.getNamePrefix() : ""; //$NON-NLS-1$
			} else {
				return MispSchema.getPrefix(this.name);
			}
		}

		/**
		 * prefixをセットする
		 * 
		 * @param prefix
		 *            The prefix to set.
		 */
		public void setNamePrefix(String prefix) {
			if (flgRef) {
				// ありえない
			} else {
				this.name = MispSchema.setPrefix(this.name, prefix);
			}
		}

		/**
		 * データ型を取得する
		 * 
		 * @return データ型
		 */
		public String getType() {
			if (flgRef) {
				MispSchemaElement em = findRefElementByName(this.name);
				return (em != null) ? em.getType() : ""; //$NON-NLS-1$
			} else {
				return type;
			}
		}

		/**
		 * データ型を取得する
		 * 
		 * @return データ型
		 */
		public String getXsdType() {
			String type = getType();
			if ((type != null) && type.equals(Const.DFS_TYPE_ATTATCHEDFILE))
				type = Const.XSD_TYPE_ATTATCHEDFILE;
			return type;
		}

		/**
		 * データ型を設定する
		 * 
		 * @param type
		 *            データ型
		 */
		public void setType(String type) {
			if (flgRef) {
				// ありえない
			} else {
				this.type = type;
			}
		}

		/**
		 * prefixを取得する
		 * 
		 * @return Returns the prefix.
		 */
		public String getTypePrefix() {
			if (flgRef) {
				MispSchemaElement em = findRefElementByName(this.name);
				return (em != null) ? em.getTypePrefix() : ""; //$NON-NLS-1$
			} else {
				return MispSchema.getPrefix(this.type);
			}
		}

		/**
		 * prefixをセットする
		 * 
		 * @param prefix
		 *            The prefix to set.
		 */
		public void setTypePrefix(String prefix) {
			if (flgRef) {
				// ありえない
			} else {
				this.type = MispSchema.setPrefix(this.type, prefix);
			}
		}

		/**
		 * maxOccursを取得する
		 * 
		 * @return Returns the maxOccurs.
		 */
		public String getMaxOccurs() {
			if (flgRef) {
				MispSchemaElement em = findRefElementByName(this.name);
				return (em != null) ? em.getMaxOccurs() : ""; //$NON-NLS-1$
			} else {
				return maxOccurs;
			}
		}

		/**
		 * maxOccursを取得する
		 * 
		 * @return Returns the maxOccurs.
		 */
		public int getMaxOccursVal() {
			if (flgRef) {
				MispSchemaElement em = findRefElementByName(this.name);
				return (em != null) ? em.getMaxOccursVal() : -1;
			} else {
				try {
					return Integer.parseInt(maxOccurs);
				} catch (Exception e) {
				}
				return -1;
			}
		}

		/**
		 * maxOccursを設定する
		 * 
		 * @param maxOccurs
		 *            The maxOccurs to set.
		 */
		public void setMaxOccurs(String maxOccurs) {
			if (flgRef) {
				// ありえない
			} else {
				this.maxOccurs = maxOccurs;
			}
		}

		/**
		 * minOccursを取得する
		 * 
		 * @return Returns the minOccurs.
		 */
		public String getMinOccurs() {
			if (flgRef) {
				MispSchemaElement em = findRefElementByName(this.name);
				return (em != null) ? em.getMinOccurs() : ""; //$NON-NLS-1$
			} else {
				return minOccurs;
			}
		}

		/**
		 * minOccursを取得する
		 * 
		 * @return Returns the minOccurs.
		 */
		public int getMinOccursVal() {
			if (flgRef) {
				MispSchemaElement em = findRefElementByName(this.name);
				return (em != null) ? em.getMinOccursVal() : -1;
			} else {
				try {
					return Integer.parseInt(minOccurs);
				} catch (Exception e) {
				}
				return -1;
			}
		}

		/**
		 * minOccursを設定する
		 * 
		 * @param minOccurs
		 *            The minOccurs to set.
		 */
		public void setMinOccurs(String minOccurs) {
			if (flgRef) {
				// ありえない
			} else {
				this.minOccurs = minOccurs;
			}
		}

		/**
		 * indexedFlgを取得する
		 * 
		 * @return Returns the indexedFlg.
		 */
		public boolean isIndexedFlg() {
			if (flgRef) {
				MispSchemaElement em = findRefElementByName(this.name);
				return (em != null) ? em.isIndexedFlg() : false;
			} else {
				return indexedFlg;
			}
		}

		/**
		 * indexedFlgをセットする
		 * 
		 * @param indexedFlg
		 *            The indexedFlg to set.
		 */
		public void setIndexedFlg(boolean indexedFlg) {
			if (flgRef) {
				// ありえない
			} else {
				this.indexedFlg = indexedFlg;
			}

		}

		/**
		 * 要素をコピーする
		 * 
		 * @param em
		 *            コピー先の要素
		 */
		public void copyTo(MispSchemaElement em) {
			em.flgRef = this.flgRef;
			em.name = this.name;
			em.type = this.type;
			em.maxOccurs = this.maxOccurs;
			em.minOccurs = this.minOccurs;

			em.flgRoot = this.flgRoot;
			em.flgFeature = this.flgFeature;
			em.indexedFlg = this.isIndexedFlg();
		}

		/**
		 * 要素をコピーする
		 * 
		 * @param em
		 *            コピー先の要素
		 */
		public void copyToWithoutRef(MispSchemaElement em) {
			em.flgRef = false;
			em.name = new String(this.getName());
			em.type = new String(this.getType());
			em.maxOccurs = new String(this.getMaxOccurs());
			em.minOccurs = new String(this.getMinOccurs());

			em.flgRoot = this.flgRoot;
			em.flgFeature = this.flgFeature;
			em.indexedFlg = this.isIndexedFlg();
		}

		/**
		 * 要素が同じかどうかを比較する
		 * 
		 * @return 比較された要素
		 */
		public boolean equals(Object obj) {
			if (obj == null)
				return false;
			if (!(obj instanceof MispSchemaElement))
				return false;
			MispSchemaElement em = (MispSchemaElement) obj;
			if (!em.name.equals(this.name))
				return false;
			if (!em.type.equals(this.type))
				return false;
			if (!em.maxOccurs.equals(this.maxOccurs))
				return false;
			if (!em.minOccurs.equals(this.minOccurs))
				return false;
			if (em.indexedFlg != this.indexedFlg)
				return false;

			return true;
		}

		/**
		 * 文字列を返す
		 * 
		 * @return 文字列
		 */
		public String toString() {
			return this.name + "/" + this.type; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
		}

		/**
		 * 複合型取得
		 * 
		 * @return 複合型
		 */
		public MispSchemaComplexType findComplexType() {
			int grp = MispSchema.this.getElementTypeGrp(type);
			if ((grp == GRP_COMPLEX_INTERNAL) || (grp == GRP_COMPLEX_EXTERNAL))
				return findComplexTypeByName(type);
			return null;
		}

		/**
		 * 複合型を返す
		 * 
		 * @return 複合型
		 */
		public MispSchemaComplexType getParentCt() {
			return parentCt;
		}
	}

	/**
	 * 複合型リストサイズを取得する
	 * 
	 * @return 複合型リストサイズ
	 */
	public int getComplexTypeSize() {
		return complexTypeList.size();
	}

	/**
	 * 複合型配列を取得する
	 * 
	 * @return 複合型配列
	 */
	public Object[] getComplexTypeArray() {
		return complexTypeList.toArray();
	}

	/**
	 * 複合型配列を取得する
	 * 
	 * @return 複合型配列
	 */
	public Object[] getComplexTypeArray(boolean flgInternal, boolean flgExt) {
		List list2 = new ArrayList();
		for (int cnt = 0; cnt < getComplexTypeSize(); cnt++) {
			MispSchemaComplexType ct = getComplexType(cnt);
			if (ct.im == null) {
				if (flgInternal)
					list2.add(ct);
			} else {
				if (flgExt)
					list2.add(ct);
			}
		}
		return list2.toArray();
	}

	/**
	 * 指定されたインデックスの複合型を取得する
	 * 
	 * @return 複合型
	 */
	public MispSchemaComplexType getComplexType(int index) {
		return (MispSchemaComplexType) complexTypeList.get(index);
	}

	/**
	 * 複合型をリストに追加する
	 * 
	 * @return 追加された複合型
	 */
	public MispSchemaComplexType addComplexType(MispSchemaImport im) {
		MispSchemaComplexType ct = new MispSchemaComplexType();
		ct.index = complexTypeList.size();
		ct.schema = this;
		if (im == null) {
			for (int cnt = 0; cnt < complexTypeList.size(); cnt++) {
				if (((MispSchemaComplexType) complexTypeList.get(cnt)).kind != KIND_INTERNAL) {
					ct.index = cnt;
					break;
				}
			}
			complexTypeList.add(ct.index, ct);
		} else {
			ct.im = im;
			if (im != null)
				ct.kind = im.includeFlg ? KIND_INCLUDE : KIND_IMPORT;
			complexTypeList.add(ct);

		}
		updateComplexTypeIndex(ct.index);
		return ct;
	}

	/**
	 * 複合型をリストから削除する
	 * 
	 * @param index
	 *            削除する複合型のインデックス
	 */
	public void delComplexType(int index) {
		if ((index >= 0) && (index < complexTypeList.size())) {
			complexTypeList.remove(index);
			updateComplexTypeIndex(index);
		}
	}

	/**
	 * 複合型のインデックスを更新する
	 * 
	 * @return 更新する複合型の開始インデックス
	 */
	private void updateComplexTypeIndex(int start) {
		if (start < 0)
			start = 0;
		for (int cnt = start; cnt < complexTypeList.size(); cnt++) {
			((MispSchemaComplexType) complexTypeList.get(cnt)).index = cnt;
		}
	}

	/**
	 * スキーマIDを取得する
	 * 
	 * @return スキーマID
	 */
	public String getId() {
		return id;
	}

	/**
	 * スキーマIDを設定する
	 * 
	 * @param id
	 *            設定するスキーマID
	 */
	public void setId(String id) {
		this.id = id;
	}

	/**
	 * スキーマリストサイズを取得する
	 * 
	 * @return スキーマリストサイズ
	 */
	public int getImportSize() {
		return importList.size();
	}

	/**
	 * スキーマ配列を取得する
	 * 
	 * @return スキーマ配列
	 */

	public Object[] getImportArray() {
		return importList.toArray();
	}

	/**
	 * 指定されたインデックスのスキーマを取得する
	 * 
	 * @return スキーマ
	 */
	public MispSchemaImport getImport(int index) {
		return (MispSchemaImport) importList.get(index);
	}

	/**
	 * スキーマをリストに追加する
	 * 
	 * @return 追加されたスキーマ
	 */
	public MispSchemaImport addImport() {
		MispSchemaImport im = new MispSchemaImport();
		im.index = importList.size();
		importList.add(im);
		return im;
	}

	/**
	 * スキーマをリストからさ削除する
	 * 
	 * @param index
	 *            削除するスキーマのインデックス
	 */
	public void delImport(int index) {
		if ((index >= 0) && (index < importList.size())) {
			MispSchemaImport im = (MispSchemaImport) importList.get(index);
			for (int cnt = complexTypeList.size(); cnt-- > 0;) {
				MispSchemaComplexType ct = (MispSchemaComplexType) complexTypeList
						.get(cnt);
				if (ct.im == im)
					complexTypeList.remove(cnt);
			}
			importList.remove(index);
			updateImportIndex(index);
		}
	}

	/**
	 * スキーマのインデックスを更新する
	 * 
	 * @return 更新するスキーマの開始インデックス
	 */
	private void updateImportIndex(int start) {
		if (start < 0)
			start = 0;
		for (int cnt = start; cnt < importList.size(); cnt++) {
			((MispSchemaImport) importList.get(cnt)).index = cnt;
		}
	}

	/**
	 * 名前空間リストサイズを取得する
	 * 
	 * @return 名前空間リストサイズ
	 */
	public int getNamespaceSize() {
		return namespaceList.size();
	}

	/**
	 * 名前空間配列を取得する
	 * 
	 * @return 名前空間配列
	 */
	public Object[] getNamespaceArray() {
		return namespaceList.toArray();
	}

	/**
	 * 指定されたインデックスの名前空間を取得する
	 * 
	 * @return 名前空間
	 */
	public MispSchemaNamespace getNamespace(int index) {
		return (MispSchemaNamespace) namespaceList.get(index);
	}

	/**
	 * 名前空間をリストに追加する
	 * 
	 * @return 追加された名前空間
	 */
	public MispSchemaNamespace addNamespace() {
		MispSchemaNamespace ns = new MispSchemaNamespace();
		ns.index = namespaceList.size();
		namespaceList.add(ns);
		return ns;
	}

	/**
	 * 
	 */
	public String getAutoPrefix() {
		String prefix;
		int cntj = 0;
		while (true) {
			boolean found = false;
			for (int cnti = 0; cnti < namespaceList.size(); cnti++) {
				MispSchemaNamespace temp = (MispSchemaNamespace) namespaceList
						.get(cnti);
				if (temp.getPrefix().equals(
						MispSchemaMessages.getString("MispSchema.63") + cntj)) { //$NON-NLS-1$
					found = true;
					break;
				}
			}
			if (!found) {
				prefix = "ns" + cntj; //$NON-NLS-1$
				break;
			}
			cntj++;
		}
		return prefix;
	}

	/**
	 * 名前空間をリストに追加する
	 * 
	 * @return 追加された名前空間
	 */
	public MispSchemaNamespace addNamespaceAutoPrefix(String uri) {
		MispSchemaNamespace ns = new MispSchemaNamespace();
		ns.index = namespaceList.size();
		namespaceList.add(ns);
		ns.uri = uri;
		ns.prefix = getAutoPrefix();
		return ns;
	}

	/**
	 * 名前空間をリストからさ削除する
	 * 
	 * @param index
	 *            削除する名前空間のインデックス
	 */
	public void delNamespace(int index) {
		if ((index >= 0) && (index < namespaceList.size())) {
			namespaceList.remove(index);
			updateNamespaceIndex(index);
		}
	}

	/**
	 * 名前空間のインデックスを更新する
	 * 
	 * @return 更新する名前空間の開始インデックス
	 */
	private void updateNamespaceIndex(int start) {
		if (start < 0)
			start = 0;
		for (int cnt = start; cnt < namespaceList.size(); cnt++) {
			((MispSchemaNamespace) namespaceList.get(cnt)).index = cnt;
		}
	}

	/**
	 * ターゲット名前空間を取得する
	 * 
	 * @return ターゲット名前空間
	 */
	public String getTargetNamespace() {
		return targetNamespace;
	}

	/**
	 * ターゲット名前空間を設定する
	 * 
	 * @param targetNamespace
	 *            設定するターゲット名前空間
	 */
	public void setTargetNamespace(String targetNamespace) {
		this.targetNamespace = targetNamespace;
	}

	/**
	 * 注釈を取得する
	 * 
	 * @return 注釈
	 */
	public String getAnnotation() {
		return annotation;
	}

	/**
	 * 注釈を設定する
	 * 
	 * @param annotation
	 *            設定する注釈
	 */
	public void setAnnotation(String annotation) {
		this.annotation = annotation;
	}

	/**
	 * 大域要素リストサイズを取得する
	 * 
	 * @return 大域要素リストサイズ
	 */
	public int getRootElementSize() {
		return rootElementList.size();
	}

	/**
	 * 大域要素配列を取得する
	 * 
	 * @return 大域要素配列
	 */
	public Object[] getRootElementArray() {
		return rootElementList.toArray();
	}

	/**
	 * フィーチャタイプ要素を取得する
	 * 
	 * @return フィーチャタイプ要素
	 */
	public MispSchemaElement getFeatureElement() {
		return getRootElement(0);
	}

	/**
	 * 指定されたインデックスの大域要素を取得する
	 * 
	 * @return 大域要素
	 */
	public MispSchemaElement getRootElement(int index) {
		return (MispSchemaElement) rootElementList.get(index);
	}

	/**
	 * 大域要素をリストに追加する
	 * 
	 * @return 追加された大域要素
	 */
	public MispSchemaElement addRootElement() {
		MispSchemaElement rootEm = new MispSchemaElement();
		rootEm.index = rootElementList.size();
		rootEm.flgRoot = true;
		rootEm.parentCt = null;
		rootElementList.add(rootEm);
		return rootEm;
	}

	/**
	 * 大域属性リストサイズを取得する
	 * 
	 * @return 大域属性リストサイズ
	 */
	public int getRootAttributeSize() {
		return rootAttributeList.size();
	}

	/**
	 * 大域属性配列を取得する
	 * 
	 * @return 大域属性配列
	 */
	public Object[] getRootAttributeArray() {
		return rootAttributeList.toArray();
	}

	/**
	 * 指定されたインデックスの大域属性を取得する
	 * 
	 * @return 大域属性
	 */
	public MispSchemaAttribute getRootAttribute(int index) {
		return (MispSchemaAttribute) rootAttributeList.get(index);
	}

	/**
	 * 大域属性をリストに追加する
	 * 
	 * @return 追加された大域属性
	 */
	public MispSchemaAttribute addRootAttribute() {
		MispSchemaAttribute rootAttr = new MispSchemaAttribute();
		rootAttr.index = rootAttributeList.size();
		rootAttr.flgRoot = true;
		rootAttr.parentCt = null;
		rootAttributeList.add(rootAttr);
		return rootAttr;
	}

	/**
	 * 指定された名前空間URIがリストに含まれてるかどうかを判断する
	 * 
	 * @param uri
	 *            名前空間URI
	 * @return 含まれている場合はTrue それ以外はFalseを返す
	 */
	public boolean existsNamespace(String uri) {
		MispSchemaNamespace ns = findNamespaceByUri(uri);
		return ns != null;
	}

	/**
	 * 指定された名前空間接頭辞がリストに含まれてるかどうかを判断する
	 * 
	 * @param prefix
	 *            名前空間接頭辞
	 * @return 含まれている場合はTrue それ以外はFalseを返す
	 */
	public boolean existsNamespacePrefix(String prefix) {
		MispSchemaNamespace ns = findNamespaceByPrefix(prefix);
		return ns != null;
	}

	/**
	 * 指定された名前空間URIがインポートスキーマリストに含まれてるかどうかを判断する
	 * 
	 * @param uri
	 *            名前空間URI
	 * @return 含まれている場合はTrue それ以外はFalseを返す
	 */
	public boolean existsImportNamespace(String uri) {
		MispSchemaImport im = findImportByUri(uri);
		return im != null;
	}

	/**
	 * 指定された名前空間接頭辞がインポートスキーマリストに含まれてるかどうかを判断する
	 * 
	 * @param prefix
	 *            名前空間接頭辞
	 * @return 含まれている場合はTrue それ以外はFalseを返す
	 */
	public boolean existsImportNamespacePrefix(String prefix) {
		MispSchemaNamespace ns = findNamespaceByPrefix(prefix);
		if (ns != null) {
			MispSchemaImport im = findImportByUri(ns.uri);
			return im != null;
		}
		return false;
	}

	/**
	 * 指定された要素のデータ型が有効かどうかをチェックする
	 * 
	 * @param typeName
	 *            データ型名
	 * @return 有効な場合はTrue それ以外はFalseを返す
	 */
	public boolean isValidElementType(String typeName) {
		return getElementTypeGrp(typeName) != GRP_UNKNOWN;
	}

	/**
	 * 指定されたフィーチャタイプ要素のデータ型が有効かどうかをチェックする
	 * 
	 * @param typeName
	 *            データ型名
	 * @return 有効な場合はTrue それ以外はFalseを返す
	 */
	public boolean isValidFeatureElementType(String typeName) {
		int grp = getElementTypeGrp(typeName);
		return ((grp == GRP_COMPLEX_INTERNAL) || (grp == GRP_COMPLEX_EXTERNAL));
	}

	/**
	 * 指定された属性のデータ型が有効かどうかをチェックする
	 * 
	 * @param typeName
	 *            データ型名
	 * @return 有効な場合はTrue それ以外はFalseを返す
	 */
	public boolean isValidAttributeType(String typeName) {
		return getAttributeTypeGrp(typeName) != GRP_UNKNOWN;

	}

	/**
	 * 解決できないデータ型
	 */
	final public static int GRP_UNKNOWN = -1;

	/**
	 * ビルトインデータ型
	 */
	final public static int GRP_BUILTIN = 0;

	/**
	 * インポートされた名前空間のデータ型
	 */
	final public static int GRP_IMPORT_OTHER = 1;

	/**
	 * 複合データ型
	 */
	final public static int GRP_COMPLEX_INTERNAL = 2;

	/**
	 * 複合データ型
	 */
	final public static int GRP_COMPLEX_EXTERNAL = 3;

	/**
	 * 複合データ型
	 */
	final public static int KIND_INTERNAL = 1;

	/**
	 * 複合データ型
	 */
	final public static int KIND_IMPORT = 2;

	/**
	 * 複合データ型
	 */
	final public static int KIND_INCLUDE = 3;

	/**
	 * 指定された属性のデータ型のグループを取得する
	 * 
	 * @param typeName
	 *            データ型名
	 * @return データ型のグループ
	 */
	public int getAttributeTypeGrp(String typeName) {
		if (MispSchemaBuiltIn.isBuiltinAttributeType(typeName))
			return GRP_BUILTIN;
		int index = typeName.indexOf(":"); //$NON-NLS-1$
		if (index > 0) {
			if (existsImportNamespacePrefix(typeName.substring(0, index))) {
				return GRP_IMPORT_OTHER;
			}
		}
		return GRP_UNKNOWN;
	}

	/**
	 * 指定された要素のデータ型のグループを取得する
	 * 
	 * @param typeName
	 *            データ型名
	 * @return データ型のグループ
	 */
	public int getElementTypeGrp(String typeName) {
		if (MispSchemaBuiltIn.isBuiltinElementType(typeName))
			return GRP_BUILTIN;
		int index = typeName.indexOf(":"); //$NON-NLS-1$

		MispSchemaComplexType ct = findComplexTypeByName(typeName);
		if (ct != null)
			return (ct.kind == KIND_INTERNAL) ? GRP_COMPLEX_INTERNAL
					: GRP_COMPLEX_EXTERNAL;
		else if (index > 0) {
			if (existsImportNamespacePrefix(typeName.substring(0, index))) {
				return GRP_IMPORT_OTHER;
			}
		}
		return GRP_UNKNOWN;
	}

	private MispSchemaElement findRefElementByName(String name) {
		String prefix1 = MispSchema.getPrefix(name);
		String local1 = MispSchema.getLocalName(name);
		String url1 = null;

		if (prefix1.length() > 0) {
			MispSchemaNamespace ns1 = findNamespaceByPrefix(prefix1);
			if (ns1 != null)
				url1 = ns1.getUri();
		} else
			url1 = this.targetNamespace;

		for (int cnt = 0; cnt < rootElementList.size(); cnt++) {
			MispSchemaElement et = (MispSchemaElement) rootElementList.get(cnt);
			String prefix2 = et.getNamePrefix();
			String local2 = MispSchema.getLocalName(et.getName());
			if (local1.equals(local2)) {
				String url2 = null;
				if (prefix2.length() > 0) {
					MispSchemaNamespace ns2 = findNamespaceByPrefix(prefix2);
					if (ns2 != null)
						url2 = ns2.getUri();
				} else
					url2 = this.targetNamespace;
				if ((url1 != null) && (url2 != null) && url1.equals(url2)) {
					return et;
				}
			}
		}
		log.warn("findRefElementByName fail:" + name); //$NON-NLS-1$
		return null;
	}

	private MispSchemaAttribute findRefAttributeByName(String name) {
		String prefix1 = MispSchema.getPrefix(name);
		String local1 = MispSchema.getLocalName(name);
		String url1 = null;

		if (prefix1.length() > 0) {
			MispSchemaNamespace ns1 = findNamespaceByPrefix(prefix1);
			if (ns1 != null)
				url1 = ns1.getUri();
		} else
			url1 = this.targetNamespace;

		for (int cnt = 0; cnt < rootAttributeList.size(); cnt++) {
			MispSchemaAttribute attr = (MispSchemaAttribute) rootAttributeList
					.get(cnt);
			String prefix2 = attr.getNamePrefix();
			String local2 = MispSchema.getLocalName(attr.getName());
			if (local1.equals(local2)) {
				String url2 = null;
				if (prefix2.length() > 0) {
					MispSchemaNamespace ns2 = findNamespaceByPrefix(prefix2);
					if (ns2 != null)
						url2 = ns2.getUri();
				} else
					url2 = this.targetNamespace;
				if ((url1 != null) && (url2 != null) && url1.equals(url2)) {
					return attr;
				}
			}
		}
		log.warn("findRefAttributeByName fail:" + name); //$NON-NLS-1$
		return null;
	}

	/**
	 * 指定された名前の複合型を検索する
	 * 
	 * @param typeName
	 *            複合型名
	 * @return 見つかった場合該当複合型、それ以外はNULLを返す
	 */
	public MispSchemaComplexType findComplexTypeByName(String typeName) {
		Object[] cts = findComplexTypesByName(typeName, null);
		if (cts.length > 0)
			return (MispSchemaComplexType) cts[0];
		else
			return null;
	}

	/**
	 * 指定された名前の複合型を検索する
	 * 
	 * @param typeName
	 *            複合型名
	 * @return 複合型の配列
	 */
	public Object[] findComplexTypesByName(String typeName) {
		return findComplexTypesByName(typeName, null);
	}

	/**
	 * 指定された名前の複合型を検索する
	 * 
	 * @param typeName
	 *            複合型名
	 * @return 複合型の配列
	 */
	public Object[] findComplexTypesByName(String typeName, String location) {
		List rc = new ArrayList();
		String prefix1 = MispSchema.getPrefix(typeName);
		String local1 = MispSchema.getLocalName(typeName);
		String url1 = null;
		if (prefix1.length() > 0) {
			MispSchemaNamespace ns1 = findNamespaceByPrefix(prefix1);
			if (ns1 != null)
				url1 = ns1.getUri();
		} else
			url1 = this.targetNamespace;

		for (int cnt = 0; cnt < complexTypeList.size(); cnt++) {
			MispSchemaComplexType ct = (MispSchemaComplexType) complexTypeList
					.get(cnt);
			if ((location != null) && !location.equals(ct.filePath))
				continue;
			String prefix2 = ct.getNamePrefix();
			String local2 = MispSchema.getLocalName(ct.getName());
			if (local1.equals(local2)) {
				String url2 = null;
				if (prefix2.length() > 0) {
					MispSchemaNamespace ns2 = findNamespaceByPrefix(prefix2);
					if (ns2 != null)
						url2 = ns2.getUri();
				} else
					url2 = this.targetNamespace;
				if ((url1 != null) && (url2 != null) && url1.equals(url2)) {
					rc.add(ct);
				}
			}
		}
		return rc.toArray();
	}

	/**
	 * 指定された名前空間URIの名前空間を検索する
	 * 
	 * @param uri
	 *            名前空間URI
	 * @return 見つかった場合該当複合型、それ以外はNULLを返す
	 */
	public MispSchemaNamespace findNamespaceByUri(String uri) {
		for (int cnt = 0; cnt < namespaceList.size(); cnt++) {
			MispSchemaNamespace ns = (MispSchemaNamespace) namespaceList
					.get(cnt);
			if (ns.uri.equals(uri))
				return ns;
		}
		return null;
	}

	/**
	 * 指定された名前空間接頭辞の名前空間を検索する
	 * 
	 * @param prefix
	 *            名前空間接頭辞
	 * @return 見つかった場合該当複合型、それ以外はNULLを返す
	 */
	public MispSchemaNamespace findNamespaceByPrefix(String prefix) {
		for (int cnt = 0; cnt < namespaceList.size(); cnt++) {
			MispSchemaNamespace ns = (MispSchemaNamespace) namespaceList
					.get(cnt);
			if (ns.prefix.equals(prefix))
				return ns;
		}
		return null;
	}

	/**
	 * 指定された名前空間URIのインポートスキーマを検索する
	 * 
	 * @param uri
	 *            名前空間URI
	 * @return 見つかった場合該当複合型、それ以外はNULLを返す
	 */
	public MispSchemaImport findImportByUri(String uri) {
		for (int cnt = 0; cnt < importList.size(); cnt++) {
			MispSchemaImport im = (MispSchemaImport) importList.get(cnt);
			if (im.uri.equals(uri))
				return im;
		}
		return null;
	}

	/**
	 * 有効な属性データ型配列を返す
	 * 
	 * @return データ型配列
	 */
	public String[] getValidAttributeTypes() {
		String[] list1 = MispSchemaBuiltIn.getBuiltinAttributeTypes();
		return list1;
	}

	/**
	 * 有効な要素データ型配列を返す
	 * 
	 * @return データ型配列
	 */
	public String[] getValidElementTypes(MispSchemaElement em) {
		String[] list1 = MispSchemaBuiltIn.getBuiltinElementTypes();
		List list2 = new ArrayList();
		for (int cnt = 0; cnt < list1.length; cnt++) {
			list2.add(list1[cnt]);
		}
		for (int cnt = 0; cnt < getComplexTypeSize(); cnt++) {
			MispSchemaComplexType ct = getComplexType(cnt);
			if (ct.getName().equals(em.getParentCt().getName()))
				continue;
			if (ct.getName().equals(this.getFeatureElement().getName()))
				continue;
			list2.add(ct.getName());
		}
		return (String[]) list2.toArray(new String[0]);
	}

	/**
	 * 有効なフィーチャタイプ要素のデータ型配列を返す
	 * 
	 * @return データ型配列
	 */
	public String[] getValidFeatureElementTypes() {
		List list2 = new ArrayList();
		for (int cnt = 0; cnt < getComplexTypeSize(); cnt++) {
			MispSchemaComplexType ct = getComplexType(cnt);
			if (ct.im == null)
				list2.add(ct.getName());
		}
		return (String[]) list2.toArray(new String[0]);
	}

	public Object clone() {
		MispSchema schema = new MispSchema();
		copyTo(schema);
		return schema;
	}

	public Object copyTo(MispSchema schema) {
		schema.id = new String(this.id);
		schema.targetNamespace = new String(this.targetNamespace);
		schema.annotation = new String(this.annotation);
		schema.rootElementList.clear();
		MispSchemaUtil.copyMispSchemaList(schema, this.rootElementList);
		schema.rootAttributeList.clear();
		MispSchemaUtil.copyMispSchemaList(schema, this.rootAttributeList);
		schema.namespaceList.clear();
		MispSchemaUtil.copyMispSchemaList(schema, this.namespaceList);
		schema.importList.clear();
		MispSchemaUtil.copyMispSchemaList(schema, this.importList);
		schema.complexTypeList.clear();
		MispSchemaUtil.copyMispSchemaList(schema, this.complexTypeList);
		schema.filePath = new String(this.filePath);
		schema.dfsCharSet = new String(this.dfsCharSet);
		return schema;
	}

	public boolean equals(Object obj) {
		if (obj == null)
			return false;
		if (!(obj instanceof MispSchema))
			return false;
		MispSchema schema = (MispSchema) obj;
		if (!schema.id.equals(this.id))
			return false;
		if (!schema.targetNamespace.equals(this.targetNamespace))
			return false;
		if (!schema.annotation.equals(this.annotation))
			return false;
		if (!MispSchemaUtil.equalsList(schema.rootElementList,
				this.rootElementList))
			return false;
		if (!MispSchemaUtil.equalsList(schema.rootAttributeList,
				this.rootAttributeList))
			return false;
		if (!MispSchemaUtil.equalsList2(schema.namespaceList,
				this.namespaceList))
			return false;
		if (!MispSchemaUtil.equalsList(schema.importList, this.importList))
			return false;
		if (!MispSchemaUtil.equalsList(schema.complexTypeList,
				this.complexTypeList))
			return false;
		return true;
	}

	/**
	 * スキーマの検証を行う
	 * 
	 * @param blFirst
	 *            全エラーを返すかか最初のエラーのみを返すフラグ
	 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
	 */
	public String validate(boolean blFirst) {
		String validate = ""; //$NON-NLS-1$

		/* ID */
		validate += validateId(blFirst);
		if ((validate.length() > 0) && blFirst)
			return validate;

		/* Namespcae */
		validate += validateNamespaceList(blFirst);
		if ((validate.length() > 0) && blFirst)
			return validate;

		/* Import */
		validate += validateImportList(blFirst);
		if ((validate.length() > 0) && blFirst)
			return validate;

		/* Targetnamespace */
		validate += validateTargetNamespace(blFirst);
		if ((validate.length() > 0) && blFirst)
			return validate;

		/* Annotation */
		validate += validateAnnotation(blFirst);
		if ((validate.length() > 0) && blFirst)
			return validate;

		/* FeatureElement */
		validate += validateRootElementList(blFirst);
		if ((validate.length() > 0) && blFirst)
			return validate;

		/* ElementType */
		validate += validateComplexTypeList(blFirst);
		if ((validate.length() > 0) && blFirst)
			return validate;

		return validate;

	}

	/**
	 * 複合型リストの検証を行う
	 * 
	 * @param blFirst
	 *            全エラーを返すかか最初のエラーのみを返すフラグ
	 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
	 */
	public String validateComplexTypeList(boolean blFirst) {
		return MispSchemaUtil.validateList(complexTypeList, blFirst);
	}

	/**
	 * スキーマIDの検証を行う
	 * 
	 * @param blFirst
	 *            全エラーを返すかか最初のエラーのみを返すフラグ
	 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
	 */
	public String validateId(boolean blFirst) {
		String validate = ""; //$NON-NLS-1$

		if (id.length() <= 0)
			validate += MispSchemaMessages.getString("MispSchema.19") + CRLF; //$NON-NLS-1$
		if ((validate.length() > 0) && blFirst)
			return validate;

		if (FLG_ID_ANYURITYPE)
			validate += chkXsAnyURI(id);
		else
			validate += chkXsID(id);
		if ((validate.length() > 0) && blFirst)
			return validate;

		return validate;

	}

	/**
	 * インポートスキーマリストの検証を行う
	 * 
	 * @param blFirst
	 *            全エラーを返すかか最初のエラーのみを返すフラグ
	 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
	 */
	public String validateImportList(boolean blFirst) {
		return MispSchemaUtil.validateList(importList, blFirst);
	}

	/**
	 * 名前空間リストの検証を行う
	 * 
	 * @param blFirst
	 *            全エラーを返すかか最初のエラーのみを返すフラグ
	 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
	 */
	public String validateNamespaceList(boolean blFirst) {
		return MispSchemaUtil.validateList(namespaceList, blFirst);
	}

	/**
	 * ターゲット名前空間の検証を行う
	 * 
	 * @param blFirst
	 *            全エラーを返すかか最初のエラーのみを返すフラグ
	 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
	 */
	public String validateTargetNamespace(boolean blFirst) {
		String validate = ""; //$NON-NLS-1$

		if (targetNamespace.length() <= 0)
			validate += MispSchemaMessages.getString("MispSchema.20") + CRLF; //$NON-NLS-1$
		if ((validate.length() > 0) && blFirst)
			return validate;

		validate += chkXsAnyURI(targetNamespace);
		if ((validate.length() > 0) && blFirst)
			return validate;

		return validate;
	}

	/**
	 * ターゲット名前空間の検証を行う
	 * 
	 * @param blFirst
	 *            全エラーを返すかか最初のエラーのみを返すフラグ
	 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
	 */
	public String validateAnnotation(boolean blFirst) {
		String validate = ""; //$NON-NLS-1$
		return validate;
	}

	/**
	 * 大域要素リストの検証を行う
	 * 
	 * @param blFirst
	 *            全エラーを返すかか最初のエラーのみを返すフラグ
	 * @return エラーがある場合該当エラーを返す それ以外は空文字を返す
	 */
	public String validateRootElementList(boolean blFirst) {
		return MispSchemaUtil.validateList(rootElementList, blFirst);
	}

	/**
	 * xs:Name型を検証する
	 * 
	 * @param val
	 *            検証する文字列
	 * @return 正しい場合True, それ以外はFalseを返す
	 */
	public static String chkXsName(String val) {
		if (val.length() == 0) {
			return MispSchemaMessages.getString("MispSchema.21") + CRLF; //$NON-NLS-1$
		}
		if (!XMLChar.isNameStart(val.charAt(0))) {
			return MispSchemaMessages.getString("MispSchema.22") + CRLF; //$NON-NLS-1$
		}
		for (int i = 1; i < val.length(); i++) {
			if (!XMLChar.isName(val.charAt(i))) {
				return MispSchemaMessages.getString("MispSchema.23") + CRLF; //$NON-NLS-1$
			}
		}
		return ""; //$NON-NLS-1$
	}

	/**
	 * xs:NCName型を検証する
	 * 
	 * @param val
	 *            検証する文字列
	 * @return 正しい場合True, それ以外はFalseを返す
	 */
	public static String chkXsNCName(String val) {
		if (val.length() == 0) {
			return MispSchemaMessages.getString("MispSchema.24") + CRLF; //$NON-NLS-1$
		}
		if (!XMLChar.isNCNameStart(val.charAt(0))) {
			return MispSchemaMessages.getString("MispSchema.25") + CRLF; //$NON-NLS-1$
		}
		for (int i = 1; i < val.length(); i++) {
			if (!XMLChar.isNCName(val.charAt(i))) {
				return MispSchemaMessages.getString("MispSchema.26") + CRLF; //$NON-NLS-1$
			}
		}
		return ""; //$NON-NLS-1$
	}

	/**
	 * xs:QName型を検証する
	 * 
	 * @param val
	 *            検証する文字列
	 * @return 正しい場合True, それ以外はFalseを返す
	 */
	public static String chkXsQName(String val) {
		int index = val.indexOf(':');
		int lastIndex = val.lastIndexOf(':');
		int length = val.length();

		if (val.length() == 0) {
			return MispSchemaMessages.getString("MispSchema.27") + CRLF; //$NON-NLS-1$
		}
		if (index == 0 || index == length - 1 || lastIndex != index) {
			return MispSchemaMessages.getString("MispSchema.28") + CRLF; //$NON-NLS-1$
		} else {
			int start = 0;

			// Namespace in XML production [6]
			if (index > 0) {
				// check that prefix is NCName
				if (!XMLChar.isNCNameStart(val.charAt(start))) {
					return MispSchemaMessages.getString("MispSchema.29") + CRLF; //$NON-NLS-1$
				}
				for (int i = 1; i < index; i++) {
					if (!XMLChar.isNCName(val.charAt(i))) {
						return MispSchemaMessages.getString("MispSchema.30") + CRLF; //$NON-NLS-1$
					}
				}
				start = index + 1;
			}

			// check local part
			if (!XMLChar.isNCNameStart(val.charAt(start))) {
				return MispSchemaMessages.getString("MispSchema.31") + CRLF; //$NON-NLS-1$
			}

			for (int i = start + 1; i < length; i++) {
				if (!XMLChar.isNCName(val.charAt(i))) {
					return MispSchemaMessages.getString("MispSchema.32") + CRLF; //$NON-NLS-1$
				}
			}
		}

		return ""; //$NON-NLS-1$
	}

	/**
	 * xs:ID型を検証する
	 * 
	 * @param val
	 *            検証する文字列
	 * @return 正しい場合True, それ以外はFalseを返す
	 */
	public static String chkXsID(String val) {
		if (val.length() == 0) {
			return MispSchemaMessages.getString("MispSchema.33") + CRLF; //$NON-NLS-1$
		}
		if (!XMLChar.isNCNameStart(val.charAt(0))) {
			return MispSchemaMessages.getString("MispSchema.34") + CRLF; //$NON-NLS-1$
		}

		for (int i = 1; i < val.length(); i++) {
			if (!XMLChar.isNCName(val.charAt(i))) {
				return MispSchemaMessages.getString("MispSchema.35") + CRLF; //$NON-NLS-1$
			}
		}
		return ""; //$NON-NLS-1$
	}

	/**
	 * xs:anyURI型を検証する
	 * 
	 * @param val
	 *            検証する文字列
	 * @return 正しい場合True, それ以外はFalseを返す
	 */
	public static String chkXsAnyURI(String val) {
		try {
			URI objUri = new URI(val);
			if (objUri != null) {
			}
		} catch (URISyntaxException e) {
			return MispSchemaMessages.getString("MispSchema.36") + CRLF; //$NON-NLS-1$
		}
		return ""; //$NON-NLS-1$
	}

	public static String chkNum(String val) {
		if (!val.matches("[0-9]*")) { //$NON-NLS-1$
			return MispSchemaMessages.getString("MispSchema.42") + CRLF; //$NON-NLS-1$
		}
		return ""; //$NON-NLS-1$
	}

	/**
	 * 文字列を返す
	 * 
	 * @return 文字列
	 */
	public String toString() {
		return this.id;
	}

	/**
	 * localNameを取得する
	 * 
	 * @return Returns the localName.
	 */
	public static String getLocalName(String orgName) {
		int index = orgName.indexOf(':');
		if (index >= 0)
			return orgName.substring(index + 1);
		else
			return orgName;
	}

	/**
	 * localNameをセットする
	 * 
	 * @param localName
	 *            The localName to set.
	 */
	public static String setLocalName(String orgName, String localName) {
		String newName;
		int index = orgName.indexOf(':');
		if (index >= 0)
			newName = orgName.substring(0, index + 1) + localName;
		else
			newName = localName;
		return newName;
	}

	/**
	 * prefixを取得する
	 * 
	 * @return Returns the prefix.
	 */
	public static String getPrefix(String orgName) {
		int index = orgName.indexOf(':');
		if (index >= 0)
			return orgName.substring(0, index);
		else
			return ""; //$NON-NLS-1$
	}

	/**
	 * prefixをセットする
	 * 
	 * @param prefix
	 *            The prefix to set.
	 */
	public static String setPrefix(String orgName, String prefix) {
		String newName;
		int index = orgName.indexOf(':');
		if (index >= 0)
			newName = prefix + ":" + orgName.substring(index + 1); //$NON-NLS-1$
		else
			newName = prefix + ":" + orgName; //$NON-NLS-1$
		return newName;
	}

	private static String indexStr(int index) {
		String temp = "0000" + index; //$NON-NLS-1$
		temp = temp.substring(temp.length() - 4, temp.length());
		return temp;
	}

	/**
	 * filePathを取得する
	 * 
	 * @return Returns the filePath.
	 */
	public String getFilePath() {
		return filePath;
	}

	/**
	 * filePathをセットする
	 * 
	 * @param filePath
	 *            The filePath to set.
	 */
	public void setFilePath(String filePath) {
		this.filePath = filePath.replaceAll("\\\\", "/"); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * ファイルパスをセットする、スキーマの中に該当ファイルパスへの参照も更新する
	 * 
	 * @param newFilePath
	 *            The filePath to set.
	 */
	public void setFilePathUpdateReference(String newFilePath) {
		for (int cnt = 0; cnt < importList.size(); cnt++) {
			MispSchemaImport im = (MispSchemaImport) importList.get(cnt);
			if (im.location.startsWith("http://")
					|| isAbsolutePath(im.location)) {
			} else {
				im.location = makeFullPath(getBaseDir(), im.location);
			}
		}
		this.filePath = newFilePath;
	}

	/**
	 * ベースフォルダを設定する
	 * 
	 * @return ベースフォルダ
	 */
	public String getBaseDir() {
		String baseDir = null;

		if ((filePath != null) && (filePath.length() > 0)) {
			if (filePath.startsWith("http://")) { //$NON-NLS-1$
				int temp = filePath.lastIndexOf("/"); //$NON-NLS-1$
				baseDir = filePath.substring(0, temp);
				if (!baseDir.endsWith("/")) //$NON-NLS-1$
					baseDir += "/"; //$NON-NLS-1$
			} else {
				File file = new File(filePath);
				baseDir = file.getParent();
			}
		}
		if ((baseDir == null) || (baseDir.length() == 0))
			baseDir = "./"; //$NON-NLS-1$
		baseDir = baseDir.replaceAll("\\\\", "/"); //$NON-NLS-1$ //$NON-NLS-2$
		return baseDir;
	}

	/**
	 * デフォルトキャラクタセットを取得する
	 * 
	 * @return Returns the dfsCharSet.
	 */
	public String getDfsCharSet() {
		return dfsCharSet;
	}

	/**
	 * デフォルトキャラクタセットをセットする
	 * 
	 * @param dfsCharSet
	 *            The dfsCharSet to set.
	 */
	public void setDfsCharSet(String dfsCharSet) {
		this.dfsCharSet = dfsCharSet;
	}

	public void adjustImportPath() {
		String baseDir = this.getBaseDir();
		for (int cnt = 0; cnt < importList.size(); cnt++) {
			MispSchemaImport im = (MispSchemaImport) importList.get(cnt);
			if (im.location.startsWith(baseDir)) {
				im.location = im.location.substring(baseDir.length());
				while (im.location.startsWith("/")) { //$NON-NLS-1$
					im.location = im.location.substring(1);
				}
			}
		}
	}

	/**
	 * インポートリストに絶対パスで指定されてるファイルあるかどうかをチェックする
	 * 
	 * @return 絶対パスで指定されるファイルがある場合はtrue それ以外はfalseを返す
	 */
	public boolean hasAbsolutePathImport() {
		for (int cnt = 0; cnt < importList.size(); cnt++) {
			MispSchemaImport im = (MispSchemaImport) importList.get(cnt);
			if (MispSchema.isAbsolutePath(im.location))
				return true;
		}
		return false;
	}

	/**
	 * 複合型のコンパレータを定義するクラス
	 * 
	 * @author shumi
	 * 
	 */
	class MispSchemaComplexTypeComparator implements Comparator {
		/*
		 * (non-Javadoc)
		 * 
		 * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
		 */
		public int compare(Object o1, Object o2) {
			int rc = 0;
			if ((o1 instanceof MispSchemaComplexType)
					&& (o2 instanceof MispSchemaComplexType)) {
				MispSchemaComplexType ct1 = (MispSchemaComplexType) o1;
				MispSchemaComplexType ct2 = (MispSchemaComplexType) o2;
				if ((ct1.kind == KIND_INTERNAL) && (ct2.kind == KIND_INTERNAL))
					rc = new Integer(ct1.index)
							.compareTo(new Integer(ct2.index));
				else {
					if (ct1.kind == KIND_INTERNAL)
						rc = -1;
					if (ct2.kind == KIND_INTERNAL)
						rc = 1;
					rc = ct1.getName().compareTo(ct2.getName());
				}
			}
			return rc;
		}

	}

	/**
	 * 複合型リストをソートする
	 * 
	 */
	public void sortComplexType() {
		Collections
				.sort(complexTypeList, new MispSchemaComplexTypeComparator());
		updateComplexTypeIndex(0);
	}

	/**
	 * 絶対パスかどうかをチェックする
	 * 
	 * @param location
	 * @return locationが絶対パスの場合はtrue, それ以外はfalseを返す
	 */
	public static boolean isAbsolutePath(String location) {
		location = location.replaceAll("\\\\", "/");
		return location.startsWith("/") || location.matches("^[A-Za-z]:/.*"); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * ファイルのフルパスを生成する
	 * 
	 * @param dir
	 *            フォルダ名
	 * @param file
	 *            ファイル名
	 * @return フルパスを返す
	 */
	public static String makeFullPath(String dir, String file) {
		return dir + (dir.endsWith("/") ? "" : "/") + file.replaceAll("\\\\", "/");
	}
}
