package daruma.client.schemaeditor.ui.views;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.eclipse.core.runtime.Path;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.FileDialog;

import daruma.bus.processor.csv.MispSchema;
import daruma.bus.processor.csv.MispSchemaAccess;
import daruma.bus.processor.csv.MispSchema.MispSchemaAttribute;
import daruma.bus.processor.csv.MispSchema.MispSchemaComplexType;
import daruma.bus.processor.csv.MispSchema.MispSchemaElement;
import daruma.bus.processor.csv.MispSchema.MispSchemaImport;
import daruma.bus.processor.csv.MispSchema.MispSchemaNamespace;
import daruma.client.schemaeditor.Messages;
import daruma.client.schemaeditor.SchemaEditorPlugin;

/**
 * スキーマエディタ操作を制御するクラス
 * 
 * @author shumi
 * 
 */
public class SchemaEditorViewOperation {
	/**
	 * ロガー
	 */
	static private Log log = LogFactory.getLog(SchemaEditorViewOperation.class);

	/**
	 * 操作なしまたはキャンセルされた
	 */
	public static int OP_NONE = 0;

	/**
	 * スキーマID変更操作
	 */
	public static final int OP_SCHEMA_ID = 1;

	/**
	 * ターゲット名前空間変更操作
	 */
	public static final int OP_SCHEMA_TNS = 2;

	/**
	 * FeatureType名変更操作
	 */
	public static final int OP_SCHEMA_FT_NAME = 3;

	/**
	 * FeatureType型変更操作
	 */
	public static final int OP_SCHEMA_FT_TYPE = 4;

	/**
	 * 要素名変更
	 */
	public static final int OP_EM_NAME = 5;

	/**
	 * 要素型変更
	 */
	public static final int OP_EM_TYPE = 6;

	/**
	 * 要素上へ移動
	 */
	public static final int OP_EM_MOVEUP = 7;

	/**
	 * 要素下へ移動
	 */
	public static final int OP_EM_MOVEDOWN = 8;

	/**
	 * 属性名変更
	 */
	public static final int OP_ATTR_NAME = 9;

	/**
	 * 属性型変更
	 */
	public static final int OP_ATTR_TYPE = 10;

	/**
	 * 属性上へ移動
	 */
	public static final int OP_ATTR_MOVEUP = 11;

	/**
	 * 属性下へ移動
	 */
	public static final int OP_ATTR_MOVEDOWN = 12;

	/**
	 * 名前空間追加
	 */
	public static final int OP_NS_ADD = 13;

	/**
	 * 名前空間削除
	 */
	public static final int OP_NS_DEL = 14;

	/**
	 * 名前空間Prefix変更
	 */
	public static final int OP_NS_PREFIX = 15;

	/**
	 * 名前空間URI変更
	 */
	public static final int OP_NS_URI = 16;

	/**
	 * インポートスキーマ追加
	 */
	public static final int OP_IM_ADD = 17;

	/**
	 * インポートスキーマ追加
	 */
	public static final int OP_IM_ADD2 = 18;

	/**
	 * インポートスキーマ削除
	 */
	public static final int OP_IM_DEL = 19;

	/**
	 * インポートスキーマロケーション変更
	 */
	public static final int OP_IM_LOCATION = 20;

	/**
	 * インポートスキーマURI変更
	 */
	public static final int OP_IM_URI = 21;

	/**
	 * 複合型追加
	 */
	public static final int OP_CT_ADD = 22;

	/**
	 * 複合型削除
	 */
	public static final int OP_CT_DEL = 23;

	/**
	 * 複合型名変更
	 */
	public static final int OP_CT_NAME = 24;

	/**
	 * 複合型基底型変更
	 */
	public static final int OP_CT_EXTENSION = 25;

	/**
	 * 複合型要素追加
	 */
	public static final int OP_CT_ET_ADD = 26;

	/**
	 * 複合型要素作成
	 */
	public static final int OP_CT_ET_DEL = 27;

	/**
	 * 複合型属性追加
	 */
	public static final int OP_CT_ATTR_ADD = 28;

	/**
	 * 複合型属性作成
	 */
	public static final int OP_CT_ATTR_DEL = 29;

	/**
	 * 新しい複合型
	 */
	public static final int OP_EM_NEW_CT = 30;

	/**
	 * 注釈変更操作
	 */
	public static final int OP_SCHEMA_ANNOTATION = 31;

	public static final int OP_EM_MIN_OCCURS = 32;

	public static final int OP_EM_MAX_OCCURS = 33;

	public static final int OP_EM_INDEXED = 34;

	public static final int OP_ATTR_INDEXED = 35;

	/**
	 * インポートスキーマリロード
	 */
	public static final int OP_IM_RELOAD = 36;

	/**
	 * インポートスキーマリロケーション
	 */
	public static final int OP_IM_RELOCATION = 37;

	/**
	 * インポートXSD
	 */
	public static final int OP_SCHEMA_IMPORT_XSD = 38;

	/**
	 * 最大インデックス番号
	 */
	private static int MAX_INDEX_NUM = 100;

	/**
	 * 最大操作記録数
	 */
	private static int MAX_OP_NUM = 100;

	/**
	 * スキーマリスト
	 */
	private List schemaList = new ArrayList();

	/**
	 * 操作リスト
	 */
	private List operationList = new ArrayList();

	/**
	 * カレントスキーマの位置
	 */
	private int currSchemaPos = 0;

	/**
	 * プロパティ
	 */
	private SchemaEditorViewProp prop;

	/**
	 * ビュー
	 */
	private SchemaEditorView view;

	/**
	 * スキーマエディタ操作クラスを生成する
	 * 
	 * @param prop
	 */
	public SchemaEditorViewOperation(SchemaEditorView view,
			SchemaEditorViewProp prop) {
		this.prop = prop;
		this.view = view;
	}

	public void reset(MispSchema schema) {
		currSchemaPos = 0;
		schemaList.clear();
		schemaList.add(schema.clone());
		operationList.clear();
	}

	/**
	 * テンプレートファイルからDARUMAスキーマを新規生成する
	 * 
	 * @param path
	 *            テンプレートファイル
	 * @return 生成されたスキーマ
	 */
	public MispSchema makeNewSchemaFromTemplate(String path) {
		MispSchema schema = null;
		MispSchemaAccess access = view.getAccess();
		try {
			URL url = SchemaEditorPlugin.getDefault().find(new Path(path));
			InputStream fis = url.openStream();
			schema = access.loadSchemaFromCsv(new InputStreamReader(fis),
					MispSchemaAccess.DFS_DFT_SEPARATOR,
					MispSchemaAccess.DFS_DFT_PARENTHESIS,
					MispSchemaAccess.DFS_DFT_COMMENT);
		} catch (Exception e) {
			log.error("File to open template file:" + path); //$NON-NLS-1$
			schema = null;
		}
		if (schema == null)
			schema = makeDftMispSchema();

		return schema;

	}

	/**
	 * 新規DARUMAスキーマを生成する
	 * 
	 * @return 生成されたスキーマ
	 */
	private MispSchema makeDftMispSchema() {

		MispSchema schema = new MispSchema();
		MispSchemaNamespace ns;
		MispSchemaImport im;
		MispSchemaComplexType ct;

		schema.setId(Messages.getString("DefaultMispSchema.0")); //$NON-NLS-1$
		schema.setTargetNamespace(Messages.getString("DefaultMispSchema.1")); //$NON-NLS-1$

		ns = schema.addNamespace();
		ns.setPrefix(Messages.getString("DefaultMispSchema.2")); //$NON-NLS-1$
		ns.setUri(Messages.getString("DefaultMispSchema.3")); //$NON-NLS-1$

		im = schema.addImport();
		im.setUri(Messages.getString("DefaultMispSchema.4")); //$NON-NLS-1$
		im.setLocation(Messages.getString("DefaultMispSchema.5")); //$NON-NLS-1$

		schema.getFeatureElement().setName(
				Messages.getString("DefaultMispSchema.8")); //$NON-NLS-1$
		schema.getFeatureElement().setType(
				Messages.getString("DefaultMispSchema.9")); //$NON-NLS-1$
		ct = schema.addComplexType(null);
		ct.setName(Messages.getString("DefaultMispSchema.9")); //$NON-NLS-1$

		/*
		 * ns = schema.addNamespace(); ns.setPrefix("rcbase");
		 * ns.setUri("http://staff.aist.go.jp/i.noda/Rescue/RandomCity/base");
		 * 
		 * MispSchemaImport im = schema.addImport();
		 * im.setUri("http://staff.aist.go.jp/i.noda/Rescue/RandomCity/base");
		 * im
		 * .setLocation("http://staff.aist.go.jp/i.noda/Rescue/RandomCity/base/common.xsd");
		 */
		return schema;
	}

	/**
	 * スキーマに名前空間を追加する
	 * 
	 * @param schema
	 *            スキーマ
	 * @return 追加された名前空間
	 */
	private MispSchemaNamespace addNamespace(MispSchema schema) {
		MispSchemaNamespace ns = schema.addNamespace();
		int index = schema.getNamespaceSize();
		ns.setPrefix(Messages.getString("DefaultMispSchema.10") + index); //$NON-NLS-1$
		ns.setUri(Messages.getString("DefaultMispSchema.11") + index); //$NON-NLS-1$
		return ns;
	}

	/**
	 * スキーマにインポートスキーマを追加する
	 * 
	 * @param schema
	 *            スキーマ
	 * @return 追加されたインポートスキーマ
	 */
	private MispSchemaImport addImport(MispSchema schema) {
		MispSchemaImport im = schema.addImport();
		int index = schema.getImportSize();
		im.setUri(Messages.getString("DefaultMispSchema.12") + index); //$NON-NLS-1$
		return im;
	}

	/**
	 * スキーマにインポートスキーマを追加する
	 * 
	 * @param schema
	 *            スキーマ
	 * @return 追加されたインポートスキーマ
	 */
	private MispSchemaImport loadImport(MispSchema schema, MispSchemaImport im,
			boolean fileDlg) {
		MispSchemaAccess access = view.getAccess();
		String ext1 = MispSchemaAccess.DFS_DFT_FILE_EXT;
		String ext2 = MispSchemaAccess.XSD_DFT_FILE_EXT;

		String file;
		if (fileDlg) {
			FileDialog openDialog = new FileDialog(this.view.getSite()
					.getShell(), SWT.OPEN);
			openDialog.setFilterExtensions(new String[] {
					"*." + ext1, "*." + ext2 }); //$NON-NLS-1$ //$NON-NLS-2$
			file = openDialog.open();
		} else {
			file = im.getLocation();
		}

		if (file != null) {
			if (file.endsWith("." + ext1) || file.endsWith("." + ext2)) //$NON-NLS-1$ //$NON-NLS-2$
			{
				if (im == null)
					im = schema.addImport();
				im.setLocation(file);
				if (!access.loadImportSchema(schema, im, schema.getBaseDir(),
						schema.getDfsCharSet(), true)) {
					schema.delImport(im.getIndex());
					im = null;
					MessageDialog.openError(view.getSite().getShell(), Messages
							.getString("SchemaEditorViewProvider.27"), //$NON-NLS-1$
							Messages.getString("SchemaEditorViewOperation.37")); //$NON-NLS-1$
				}
			} else {
				MessageDialog.openError(view.getSite().getShell(), Messages
						.getString("SchemaEditorViewProvider.27"), //$NON-NLS-1$
						Messages.getString("SchemaEditorViewOperation.39")); //$NON-NLS-1$
			}
		}
		return im;
	}

	/**
	 * スキーマに複合型を追加する
	 * 
	 * @param schema
	 *            スキーマ
	 * @return 追加された複合型
	 */
	private MispSchemaComplexType addComplexType(MispSchema schema) {
		MispSchemaComplexType ct = schema.addComplexType(null);
		String prefix = schema.getFeatureElement().getName()
				+ Messages.getString("DefaultMispSchema.13"); //$NON-NLS-1$
		int newIndex = 1;
		LOOP: while (newIndex < MAX_INDEX_NUM) {
			String temp = prefix + newIndex;
			for (int cntj = 0; cntj < schema.getComplexTypeSize(); cntj++) {
				MispSchema.MispSchemaComplexType chk = schema
						.getComplexType(cntj);
				if (temp.equals(chk.getName())) {
					newIndex++;
					continue LOOP;
				}
			}
			break;
		}

		ct.setName(prefix + newIndex);
		return ct;

	}

	/**
	 * 複合型に属性を追加する
	 * 
	 * @param schema
	 *            スキーマ
	 * @param currCt
	 *            複合型
	 * @param 追加位置
	 * @return 追加された属性
	 */
	private MispSchemaAttribute addAttribute(MispSchema schema,
			MispSchemaComplexType currCt, int index) {
		MispSchemaAttribute attr = currCt.addAttribute(index);
		String prefix = Messages.getString("DefaultMispSchema.14"); //$NON-NLS-1$
		int newIndex = 1;
		LOOP: while (newIndex < MAX_INDEX_NUM) {
			String temp = prefix + newIndex;
			for (int cntj = 0; cntj < currCt.getAttributeSize(); cntj++) {
				MispSchema.MispSchemaAttribute chk = currCt.getAttribute(cntj);
				if (temp.equals(chk.getName())) {
					newIndex++;
					continue LOOP;
				}
			}
			break;
		}

		attr.setName(prefix + newIndex);
		attr.setType(prop.getDftAttrType());
		return attr;
	}

	/**
	 * 複合型に要素を追加する
	 * 
	 * @param schema
	 *            スキーマ
	 * @param currCt
	 *            複合型
	 * @param 追加位置
	 * @return 追加された要素
	 */
	private MispSchemaElement addElement(MispSchema schema,
			MispSchemaComplexType currCt, int index) {
		MispSchemaElement et = currCt.addElement(index);
		String prefix = Messages.getString("DefaultMispSchema.15"); //$NON-NLS-1$
		int newIndex = 1;
		LOOP: while (newIndex < MAX_INDEX_NUM) {
			String temp = prefix + newIndex;
			for (int cntj = 0; cntj < currCt.getElementSize(); cntj++) {
				MispSchema.MispSchemaElement chk = currCt.getElement(cntj);
				if (temp.equals(chk.getName())) {
					newIndex++;
					continue LOOP;
				}
			}
			break;
		}

		et.setName(prefix + newIndex);
		et.setType(prop.getDftElementType());
		return et;
	}

	private void logOp(MispSchema schema, int op) {
		log.info("Operation:" + getOpName(op)); //$NON-NLS-1$
		for (int cnt = schemaList.size(); cnt-- > (currSchemaPos + 1);) {
			schemaList.remove(cnt);
			operationList.remove(cnt - 1);
		}

		if (schemaList.size() >= MAX_OP_NUM) {
			schemaList.remove(0);
			operationList.remove(0);
		}
		schemaList.add(schema.clone());
		operationList.add(getOpName(op));
		currSchemaPos = schemaList.size() - 1;
	}

	public static String getOpName(int op) {
		String opName = ""; //$NON-NLS-1$
		switch (op) {
		case OP_SCHEMA_ID:
			opName = Messages.getString("SchemaEditorViewOperation.2"); //$NON-NLS-1$
			break;
		case OP_SCHEMA_TNS:
			opName = Messages.getString("SchemaEditorViewOperation.3"); //$NON-NLS-1$
			break;
		case OP_SCHEMA_ANNOTATION:
			opName = Messages.getString(Messages
					.getString("SchemaEditorViewOperation.1")); //$NON-NLS-1$
			break;
		case OP_EM_MIN_OCCURS:
			opName = Messages.getString("SchemaEditorViewOperation.0"); //$NON-NLS-1$
			break;
		case OP_EM_MAX_OCCURS:
			opName = Messages.getString("SchemaEditorViewOperation.31"); //$NON-NLS-1$
			break;
		case OP_EM_INDEXED:
			opName = Messages.getString("SchemaEditorViewOperation.32"); //$NON-NLS-1$
			break;
		case OP_ATTR_INDEXED:
			opName = Messages.getString("SchemaEditorViewOperation.33"); //$NON-NLS-1$
			break;
		case OP_SCHEMA_FT_NAME:
			opName = Messages.getString("SchemaEditorViewOperation.4"); //$NON-NLS-1$
			break;
		case OP_SCHEMA_FT_TYPE:
			opName = Messages.getString("SchemaEditorViewOperation.5"); //$NON-NLS-1$
			break;
		case OP_EM_NAME:
			opName = Messages.getString("SchemaEditorViewOperation.6"); //$NON-NLS-1$
			break;
		case OP_EM_TYPE:
			opName = Messages.getString("SchemaEditorViewOperation.7"); //$NON-NLS-1$
			break;
		case OP_EM_MOVEUP:
			opName = Messages.getString("SchemaEditorViewOperation.8"); //$NON-NLS-1$
			break;
		case OP_EM_MOVEDOWN:
			opName = Messages.getString("SchemaEditorViewOperation.9"); //$NON-NLS-1$
			break;
		case OP_EM_NEW_CT:
			opName = Messages.getString("SchemaEditorViewOperation.10"); //$NON-NLS-1$
			break;
		case OP_ATTR_NAME:
			opName = Messages.getString("SchemaEditorViewOperation.11"); //$NON-NLS-1$
			break;
		case OP_ATTR_TYPE:
			opName = Messages.getString("SchemaEditorViewOperation.12"); //$NON-NLS-1$
			break;
		case OP_ATTR_MOVEUP:
			opName = Messages.getString("SchemaEditorViewOperation.13"); //$NON-NLS-1$
			break;
		case OP_ATTR_MOVEDOWN:
			opName = Messages.getString("SchemaEditorViewOperation.14"); //$NON-NLS-1$
			break;
		case OP_NS_ADD:
			opName = Messages.getString("SchemaEditorViewOperation.15"); //$NON-NLS-1$
			break;
		case OP_NS_DEL:
			opName = Messages.getString("SchemaEditorViewOperation.16"); //$NON-NLS-1$
			break;
		case OP_NS_URI:
			opName = Messages.getString("SchemaEditorViewOperation.17"); //$NON-NLS-1$
			break;
		case OP_NS_PREFIX:
			opName = Messages.getString("SchemaEditorViewOperation.18"); //$NON-NLS-1$
			break;

		case OP_IM_ADD:
			opName = Messages.getString("SchemaEditorViewOperation.19"); //$NON-NLS-1$
			break;
		case OP_IM_ADD2:
			opName = Messages.getString(Messages
					.getString("SchemaEditorViewOperation.40")); //$NON-NLS-1$
			break;
		case OP_IM_RELOAD:
			opName = Messages.getString(Messages
					.getString("SchemaEditorViewOperation.41")); //$NON-NLS-1$
			break;
		case OP_IM_RELOCATION:
			opName = Messages.getString(Messages
					.getString("SchemaEditorViewOperation.42")); //$NON-NLS-1$
			break;
		case OP_IM_DEL:
			opName = Messages.getString("SchemaEditorViewOperation.20"); //$NON-NLS-1$
			break;
		case OP_IM_URI:
			opName = Messages.getString("SchemaEditorViewOperation.21"); //$NON-NLS-1$
			break;
		case OP_IM_LOCATION:
			opName = Messages.getString("SchemaEditorViewOperation.22"); //$NON-NLS-1$
			break;

		case OP_CT_ADD:
			opName = Messages.getString("SchemaEditorViewOperation.23"); //$NON-NLS-1$
			break;
		case OP_CT_DEL:
			opName = Messages.getString("SchemaEditorViewOperation.24"); //$NON-NLS-1$
			break;
		case OP_CT_NAME:
			opName = Messages.getString("SchemaEditorViewOperation.25"); //$NON-NLS-1$
			break;
		case OP_CT_EXTENSION:
			opName = Messages.getString("SchemaEditorViewOperation.26"); //$NON-NLS-1$
			break;
		case OP_CT_ET_ADD:
			opName = Messages.getString("SchemaEditorViewOperation.27"); //$NON-NLS-1$
			break;
		case OP_CT_ET_DEL:
			opName = Messages.getString("SchemaEditorViewOperation.28"); //$NON-NLS-1$
			break;
		case OP_CT_ATTR_ADD:
			opName = Messages.getString("SchemaEditorViewOperation.29"); //$NON-NLS-1$
			break;
		case OP_CT_ATTR_DEL:
			opName = Messages.getString("SchemaEditorViewOperation.30"); //$NON-NLS-1$
			break;

		case OP_SCHEMA_IMPORT_XSD:
			opName = Messages.getString("SchemaEditorViewOperation.43"); //$NON-NLS-1$
		}
		return opName;
	}

	public MispSchema editUndoRedo(boolean flgRedo) {
		if (isCanEditUndoRedo(flgRedo)) {
			if (flgRedo) {
				currSchemaPos++;
				log.info("Operation:Redo:" + currSchemaPos); //$NON-NLS-1$
			} else {
				currSchemaPos--;
				log.info("Operation:Undo:" + currSchemaPos); //$NON-NLS-1$
			}
			MispSchema newSchema = (MispSchema) schemaList.get(currSchemaPos);
			return (MispSchema) newSchema.clone();
		}
		return null;
	}

	public boolean isCanEditUndoRedo(boolean flgRedo) {
		if (flgRedo) {
			return ((currSchemaPos >= 0) && (currSchemaPos < (schemaList.size() - 1)));
		} else {
			return ((currSchemaPos > 0) && (currSchemaPos <= (schemaList.size() - 1)));
		}
	}

	public String getEditUndoRedoOpName(boolean flgRedo) {
		if (isCanEditUndoRedo(flgRedo)) {
			if (flgRedo)
				return (String) operationList.get(currSchemaPos);
			else
				return (String) operationList.get(currSchemaPos - 1);
		}
		return "";

	}

	public int opSchema(MispSchema schema, int op, Object param) {
		boolean done = false;
		switch (op) {
		case OP_SCHEMA_ID:
			if (!schema.getId().equals((String) param)) {
				schema.setId((String) param);
				done = true;
			}
			break;
		case OP_SCHEMA_TNS:
			if (!schema.getTargetNamespace().equals((String) param)) {
				schema.setTargetNamespace((String) param);
				done = true;
			}
			break;
		case OP_SCHEMA_ANNOTATION:
			if (!schema.getAnnotation().equals((String) param)) {
				schema.setAnnotation((String) param);
				done = true;
			}
			break;
		case OP_SCHEMA_FT_NAME:
			if (!schema.getFeatureElement().getName().equals((String) param)) {
				schema.getFeatureElement().setName((String) param);
				done = true;
			}
			break;
		case OP_SCHEMA_FT_TYPE:
			MispSchemaElement em = schema.getFeatureElement();
			if (!em.getType().equals((String) param)) {
				/*
				 * MispSchemaComplexType ct = em.findComplexType(); if (ct !=
				 * null) { if (ct.getSchema().findComplexTypeByName((String)
				 * param) != null) { } else ct.setNameUpdateReference((String)
				 * param); }
				 */
				em.setType((String) param);
				done = true;

			}
			break;
		case OP_SCHEMA_IMPORT_XSD:
			((MispSchema) param).copyTo(schema);
			done = true;
			break;
		}

		if (done) {
			logOp(schema, op);
			return op;
		} else {
			return OP_NONE;
		}
	}

	public int opElement(MispSchema schema, MispSchemaElement em, int op,
			Object param) {
		boolean done = false;
		switch (op) {
		case OP_EM_NAME:
			if (!em.getName().equals((String) param)) {
				if (em.getParentCt().findElements((String) param).length > 0) {
					MessageDialog.openError(view.getSite().getShell(), Messages
							.getString("SchemaEditorViewProvider.27"), //$NON-NLS-1$
							Messages.getString("SchemaEditorViewProvider.26")); //$NON-NLS-1$
				} else {
					em.setName((String) param);
					done = true;
				}
			}
			break;
		case OP_EM_TYPE:
			if (!em.getType().equals((String) param)) {
				em.setType((String) param);
				done = true;
			}
			break;
		case OP_EM_MIN_OCCURS:
			if (!em.getMinOccurs().equals((String) param)) {
				em.setMinOccurs((String) param);
				done = true;
			}
			break;
		case OP_EM_MAX_OCCURS:
			if (!em.getMaxOccurs().equals((String) param)) {
				em.setMaxOccurs((String) param);
				done = true;
			}
			break;
		case OP_EM_INDEXED:
			if (em.isIndexedFlg() != ((Boolean) param).booleanValue()) {
				em.setIndexedFlg(((Boolean) param).booleanValue());
				done = true;
			}
			break;
		case OP_EM_MOVEUP:
		case OP_EM_MOVEDOWN:
			em.getParentCt().moveElement(em, (op == OP_EM_MOVEUP) ? -1 : 1);
			done = true;
			break;
		case OP_EM_NEW_CT:
			MispSchemaComplexType ct = addComplexType(schema);
			em.setType(ct.getName());
			done = true;
			break;

		}

		if (done) {
			logOp(schema, op);
			return op;
		} else {
			return OP_NONE;
		}
	}

	public int opAttribute(MispSchema schema, MispSchemaAttribute attr, int op,
			Object param) {
		boolean done = false;
		switch (op) {
		case OP_ATTR_NAME:
			if (!attr.getName().equals((String) param)) {
				if (attr.getParentCt().findAttributes((String) param).length > 0) {
					MessageDialog.openError(view.getSite().getShell(), Messages
							.getString("SchemaEditorViewProvider.29"), //$NON-NLS-1$
							Messages.getString("SchemaEditorViewProvider.28")); //$NON-NLS-1$
				} else {
					attr.setName((String) param);
					done = true;
				}
			}
			break;
		case OP_ATTR_TYPE:
			if (!attr.getType().equals((String) param)) {
				attr.setType((String) param);
				done = true;
			}
			break;
		case OP_ATTR_INDEXED:
			if (attr.isIndexedFlg() != ((Boolean) param).booleanValue()) {
				attr.setIndexedFlg(((Boolean) param).booleanValue());
				done = true;
			}
			break;
		case OP_ATTR_MOVEUP:
		case OP_ATTR_MOVEDOWN:
			attr.getParentCt().moveAttribute(attr,
					(op == OP_ATTR_MOVEUP) ? -1 : 1);
			done = true;
			break;
		}

		if (done) {
			logOp(schema, op);
			return op;
		} else {
			return OP_NONE;
		}
	}

	public int opNamespace(MispSchema schema, MispSchemaNamespace ns, int op,
			Object param) {
		boolean done = false;
		switch (op) {
		case OP_NS_ADD:
			ns = addNamespace(schema);
			done = true;
			break;
		case OP_NS_DEL:
			schema.delNamespace(ns.getIndex());
			done = true;
			break;
		case OP_NS_PREFIX:
			if (!ns.getPrefix().equals((String) param)) {
				ns.setPrefixReference((String) param);
				done = true;
			}
			break;
		case OP_NS_URI:
			if (!ns.getUri().equals((String) param)) {
				ns.setUri((String) param);
				done = true;
			}
			break;
		}

		if (done) {
			logOp(schema, op);
			return op;
		} else {
			return OP_NONE;
		}
	}

	public int opImport(MispSchema schema, MispSchemaImport im, int op,
			Object param) {
		boolean done = false;
		switch (op) {
		case OP_IM_ADD:
			im = addImport(schema);
			done = true;
			break;
		case OP_IM_ADD2:
			im = loadImport(schema, null, true);
			done = true;
			break;
		case OP_IM_RELOAD:
			im = loadImport(schema, im, false);
			done = true;
			break;
		case OP_IM_RELOCATION:
			im = loadImport(schema, im, true);
			done = true;
			break;
		case OP_IM_DEL:
			schema.delImport(im.getIndex());
			done = true;
			break;
		case OP_IM_LOCATION:
			if (!im.getLocation().equals((String) param)) {
				im.setLocation((String) param);
				done = true;
			}
			break;
		case OP_IM_URI:
			if (!im.getUri().equals((String) param)) {
				im.setUri((String) param);
				done = true;
			}
			break;
		}

		if (done) {
			logOp(schema, op);
			return op;
		} else {
			return OP_NONE;
		}
	}

	public int opComplexType(MispSchema schema, MispSchemaComplexType ct,
			int op, Object param) {
		boolean done = false;
		switch (op) {
		case OP_CT_ADD:
			ct = addComplexType(schema);
			done = true;
			break;
		case OP_CT_DEL:
			schema.delComplexType(ct.getIndex());
			done = true;
			break;
		case OP_CT_EXTENSION:
			if (!ct.getExtension().equals((String) param)) {
				ct.setExtension((String) param);
				done = true;
			}
			break;
		case OP_CT_NAME:
			if (!ct.getName().equals((String) param)) {
				if (ct.getSchema().findComplexTypeByName((String) param) != null) {
					MessageDialog.openError(view.getSite().getShell(), Messages
							.getString("SchemaEditorViewProvider.31"), //$NON-NLS-1$
							Messages.getString("SchemaEditorViewProvider.30")); //$NON-NLS-1$
				} else {
					ct.setNameUpdateReference((String) param);
					done = true;
				}
			}
			break;
		case OP_CT_ET_ADD:
			addElement(schema, ct, ((Integer) param).intValue());
			done = true;
			break;
		case OP_CT_ET_DEL:
			ct.delElement((MispSchemaElement) param);
			done = true;
			break;
		case OP_CT_ATTR_ADD:
			addAttribute(schema, ct, ((Integer) param).intValue());
			done = true;
			break;
		case OP_CT_ATTR_DEL:
			ct.delAttribute((MispSchemaAttribute) param);
			done = true;
			break;
		}

		if (done) {
			logOp(schema, op);
			return op;
		} else {
			return OP_NONE;
		}
	}
}
