package application.runtime;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ConfigurableApplicationContext;
import org.springframework.context.support.FileSystemXmlApplicationContext;

import application.startup.SystemProps;


public class ServiceComponentLoader {

    static private Log log = LogFactory.getLog(ServiceComponentLoader.class);

    // <ServiceComponentConfig ,ApplicationContext>
    private Map componentContextMap = new HashMap();

    public boolean loadServiceComponents(ApplicationContext systemContext) {
        try {
            Map configFileMap = retrieveComponentConfigFiles();
            if (configFileMap != null) {
                Iterator itComponentIDs = configFileMap.keySet().iterator();
                while (itComponentIDs.hasNext()) {
                    // create component config
                    File curComponentDir = (File) itComponentIDs.next();
                    log.info("---- [" + curComponentDir.getName() + "] ServiceComponent is loading... ----");

                    File configPath = new File(curComponentDir, System
                            .getProperty(SystemProps.PROPS_SYS_COMPONENT_CONFIGDIR_NAME));
                    File libPath = new File(curComponentDir, System.getProperty(SystemProps.PROPS_SYS_COMPONENT_LIBDIR_NAME));
                    File configFile = new File(configFileMap.get(curComponentDir).toString());

                    ServiceComponentConfig config = new ServiceComponentConfig();
                    config.setComponentID(curComponentDir.getName());
                    config.setConfigPath(configPath.getCanonicalPath());
                    config.setLibraryPath(libPath.getCanonicalPath());

                    log.info("Config XML: " + configFile.getCanonicalPath() + " ]");
                    log.info("ComponentID: " + config.getComponentID() + " ]");
                    log.info("Config Files: " + config.getConfigPath() + " ]");
                    log.info("Libraries: " + config.getLibraryPath() + " ]");

                    ClassLoader cl = config.getClassLoader();
                    ClassLoader oldCl = Thread.currentThread().getContextClassLoader();
                    if (configFile != null && configFile.exists() && cl != null) {
                        Thread.currentThread().setContextClassLoader(cl);
                        ConfigurableApplicationContext context = new FileSystemXmlApplicationContext(new String[] { configFile
                                .getCanonicalPath() }, systemContext);
                        componentContextMap.put(config, context);
                    }
                    Thread.currentThread().setContextClassLoader(oldCl);

                    log.info("---- [" + curComponentDir.getName() + "] ServiceComponent has loaded. ----");

                }
            }

            return true;

        } catch (Exception e) {
            log.error(e.getMessage(), e);
            return false;
        }
    }

    /**
     * すべてのコンポーネントの設定ファイル名の取得
     * 
     * @return ディレクトリのファイルオブジェクトをkeyとした、設定ファイルのPATHを返す。
     */
    private Map retrieveComponentConfigFiles() throws IOException {

        Map configFileMap = new HashMap();

        String componentCfgPath = System.getProperty(SystemProps.PROPS_SYS_COMPONENT_PATH);
        File componentRootDir = new File(componentCfgPath);
        if (componentRootDir.exists() && componentRootDir.isDirectory()) {
            File[] componentDirs = componentRootDir.listFiles();
            for (int curIndex = 0; curIndex < componentDirs.length; curIndex++) {
                // Check component directories
                File curComponentDir = componentDirs[curIndex];
                if (curComponentDir.isDirectory()) {
                    File[] configFiles = curComponentDir.listFiles(new FilenameFilter() {
                        public boolean accept(File dir, String name) {
                            return name.equals(System.getProperty(SystemProps.PROPS_SYS_COMPONENT_CONFIG_NAME));
                        }
                    });
                    for (int curCfgFileIndex = 0; curCfgFileIndex < configFiles.length; curCfgFileIndex++) {
                        // check the component config file in a directory
                        File curFile = configFiles[curCfgFileIndex];
                        if (curFile.isFile()) {
                            configFileMap.put(curComponentDir, curFile.getCanonicalPath());
                        }
                    }
                }
            }
        }
        if (configFileMap.size() > 0) {
            return configFileMap;
        }
        return null;
    }

    public Set getComponentContextSet() {

        Set contextSet = new HashSet(componentContextMap.values());
        return contextSet;

    }

    public ConfigurableApplicationContext getComponentContext(String componentID) {
        Iterator itConfigs = componentContextMap.keySet().iterator();
        while (itConfigs.hasNext()) {
            ServiceComponentConfig config = (ServiceComponentConfig) itConfigs.next();
            if (config.getComponentID().equals(componentID)) {
                ConfigurableApplicationContext context = (ConfigurableApplicationContext) componentContextMap.get(config);
                return context;
            }
        }
        return null;
    }

    public Map getServiceComponent(String componentID) {

        ConfigurableApplicationContext context = getComponentContext(componentID);
        if (context != null) {
            Map serviceComponentMap = context.getBeansOfType(ServiceComponent.class);
            return serviceComponentMap;
        }
        else
            return null;

    }

    public Set getServiceComponentConfigSet() {

        return componentContextMap.keySet();
    }

    /**
     * ServiceComponentConfigを返す。
     * 
     * @param componentID
     *            コンポーネントのID
     * @return ServiceComponentConfig オブジェクト、ComponentIDが正しくない場合は nullを返す。
     */
    public ServiceComponentConfig getServiceComponentConfig(String componentID) {

        //Map resultMap = new HashMap();
        Iterator itConfigs = componentContextMap.keySet().iterator();
        while (itConfigs.hasNext()) {
            ServiceComponentConfig config = (ServiceComponentConfig) itConfigs.next();
            if (config.getComponentID().equals(componentID)) {
                return config;
            }
        }
        return null;
    }

    /**
     * Component Context からServiceComponentのインスタンスを取得する
     * 
     * @return
     */
    public Map getServiceComponentMap() {

        Map resultMap = new HashMap();
        Iterator itContexts = componentContextMap.values().iterator();
        while (itContexts.hasNext()) {
            ConfigurableApplicationContext context = (ConfigurableApplicationContext) itContexts.next();
            Map serviceComponentMap = context.getBeansOfType(ServiceComponent.class);
            resultMap.putAll(serviceComponentMap);
        }
        return resultMap;

    }

    public boolean unloadServiceComponent(String componentID) {
        ServiceComponentConfig key = getServiceComponentConfig(componentID);
        ConfigurableApplicationContext context = (ConfigurableApplicationContext) this.componentContextMap.get(key);
        if (context != null) {
            context.close();
            this.componentContextMap.remove(key);
            return true;
        }
        return false;

    }

    public boolean unloadServiceComponents() {
        Object[] keys = componentContextMap.keySet().toArray();
        for (int index = 0; index < keys.length; index++) {
            ServiceComponentConfig curKey = (ServiceComponentConfig) keys[index];
            if (!unloadServiceComponent(curKey.getComponentID()))
                return false;
        }
        return true;

    }

    // /**
    // * すべてのコンポーネントの設定ファイル名の取得
    // *
    // * @return ディレクトリのファイルオブジェクトをkeyとした、設定ファイルのPATHを返す。
    // */
    // private static Map retrieveComponentConfigFiles() throws IOException {
    //
    // Map configFileMap = new HashMap();
    //
    // String componentCfgPath =
    // System.getProperty(SystemProps.PROPS_SYS_COMPONENT_PATH);
    // File componentRootDir = new File(componentCfgPath);
    // if (componentRootDir.exists() && componentRootDir.isDirectory()) {
    // File[] componentDirs = componentRootDir.listFiles();
    // for (int curIndex = 0; curIndex < componentDirs.length; curIndex++) {
    // // Check component directories
    // File curComponentDir = componentDirs[curIndex];
    // if (curComponentDir.isDirectory()) {
    // File[] configFiles = curComponentDir.listFiles(new FilenameFilter() {
    // public boolean accept(File dir, String name) {
    // return
    // name.equals(System.getProperty(SystemProps.PROPS_SYS_COMPONENT_CONFIG_NAME));
    // }
    // });
    // for (int curCfgFileIndex = 0; curCfgFileIndex < configFiles.length;
    // curCfgFileIndex++) {
    // // check the component config file in a directory
    // File curFile = configFiles[curCfgFileIndex];
    // if (curFile.isFile()) {
    // configFileMap.put(curComponentDir,curFile.getCanonicalPath());
    // }
    // }
    // }
    // }
    // }
    // if (configFileMap.size() > 0) {
    // return configFileMap;
    // }
    // return null;
    // }

    // /**
    // * Component Context からServiceComponentのインスタンスを取得する
    // *
    // * @return
    // */
    // static public Map getServiceComponents() {
    //
    // Map resultMap = componentLoader.getServiceComponentMap();
    // return resultMap;
    //
    // }

    public boolean shutdownComponents() {
        try {
            Set componentConfigSet = getServiceComponentConfigSet();
            Iterator itConfigs = componentConfigSet.iterator();
            while (itConfigs.hasNext()) {
                ServiceComponentConfig config = (ServiceComponentConfig) itConfigs.next();
                ClassLoader oldCl = Thread.currentThread().getContextClassLoader();
                Thread.currentThread().setContextClassLoader(config.getClassLoader());
                Map componentMap = getServiceComponent(config.getComponentID());
                Iterator itComponents = componentMap.values().iterator();
                while (itComponents.hasNext()) {
                    ServiceComponent component = (ServiceComponent) itComponents.next();
                    component.stop();
                }
                Thread.currentThread().setContextClassLoader(oldCl);
            }
            return true;

        } catch (ServiceComponentException e) {
            log.error(e.getMessage(), e);
            return false;
        }
    }

    public boolean startComponents(boolean onlyAutoStartComponent) {

        try {
            Set componentConfigSet = getServiceComponentConfigSet();
            Iterator itConfigs = componentConfigSet.iterator();
            while (itConfigs.hasNext()) {
                ServiceComponentConfig config = (ServiceComponentConfig) itConfigs.next();
                ClassLoader oldCl = Thread.currentThread().getContextClassLoader();
                Thread.currentThread().setContextClassLoader(config.getClassLoader());
                Map componentMap = getServiceComponent(config.getComponentID());
                Iterator itComponents = componentMap.values().iterator();
                while (itComponents.hasNext()) {
                    ServiceComponent component = (ServiceComponent) itComponents.next();
                    if (!onlyAutoStartComponent)
                        component.start();
                    else {
                        if (component.isAutoStart())
                            component.start();
                    }
                }
                Thread.currentThread().setContextClassLoader(oldCl);
            }
            return true;

        } catch (ServiceComponentException e) {
            log.error(e.getMessage(), e);
            return false;
        }
    }

}
