package application.runtime;

import java.io.File;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.context.ApplicationContext;
import org.springframework.context.ConfigurableApplicationContext;
import org.springframework.context.support.FileSystemXmlApplicationContext;

import application.startup.FileFinder;
import application.startup.SystemProps;

/**
 * システムコンテキスト
 * 
 * @author masanori.matsuoka@netconscious.jp
 */
public class SystemContext {

    static private ConfigurableApplicationContext systemContext = null;
    static private ConfigurableApplicationContext userContext = null;
    static private ServiceComponentLoader componentLoader = null;
    static private Object lockObject = new Object();

    static ApplicationContext getSystemContext() {
        return systemContext;
    }

    /**
     * システムアプリケーションコンテキストの初期化
     * 
     * @return 成功した場合True それ以外はFalseを返す
     */
    static protected boolean openSystemContext() {

        synchronized (lockObject) {
            try {
                if (systemContext == null) {
                    // load the beans of system.context.xml
                    log.debug(System.getProperty(SystemProps.PROPS_SYS_CONTEXT_CONFIG_XML));
                    String configFile = System.getProperty(SystemProps.PROPS_SYS_CONTEXT_CONFIG_XML);
                    if(configFile.startsWith("/")) configFile = "file://" + configFile;
                    systemContext = new FileSystemXmlApplicationContext(configFile);

                    return systemContext != null;
                }
                return false;
            } catch (Exception e) {
                log.error(e.getMessage(), e);
                return false;
            }
        }
    }

    static protected boolean openComponentContext() {
        synchronized (lockObject) {
            try {
                if (systemContext != null && componentLoader == null) {
                    componentLoader = new ServiceComponentLoader();
                    componentLoader.loadServiceComponents(systemContext);
                }
                return true;

            } catch (Exception e) {
                log.error(e.getMessage(), e);
                return false;
            }
        }
    }

    /**
     * ユーザアプリケーションコンテキストの初期化
     * 
     * @return 成功した場合True それ以外はFalseを返す
     */
    static protected boolean openUserContext() {

        synchronized (lockObject) {
            if (systemContext != null && userContext == null) {
                List beanConfigFiles = new ArrayList();

                log.debug(System.getProperty(SystemProps.PROPS_USER_BEANS_PATH));
                String workDirs = System.getProperty(SystemProps.PROPS_USER_BEANS_PATH);
                List directories = FileFinder.splitFilePath(workDirs, ";");
                Iterator iDirectory = directories.iterator();
                while (iDirectory.hasNext()) {
                    File directory = (File) iDirectory.next();
                    String[] fileNames = FileFinder.getFileNames(directory, ".*\\.xml");
                    if (fileNames != null) {
                        for (int curIndex = 0; curIndex < fileNames.length; curIndex++) {
                            beanConfigFiles.add(fileNames[curIndex]);
                        }

                    }
                }

                if (beanConfigFiles.size() > 0) {
                    String[] fileNames = (String[]) beanConfigFiles.toArray(new String[0]);
                    if (fileNames != null) {
                        userContext = new FileSystemXmlApplicationContext(fileNames, systemContext);
                        return userContext != null;
                    }
                    else
                        return false;
                }
                else
                    return false;
            }
            else
                return false;
        }

    }

    static public void refreshComponentContext() {

        synchronized (lockObject) {
            shutdownComponents();
            openComponentContext();
            startComponents(true);

            log.info("ComponentContext were refreshed.");
        }
    }

    static public void refreshUserContext() {

        synchronized (lockObject) {

            if (userContext != null) {
                userContext.refresh();
                log.info("UserContext were refreshed.");
            }
        }
    }

    static public void refreshSystemContext() {

        synchronized (lockObject) {

            if (systemContext != null) {
                systemContext.refresh();
                log.info("SystemContext were refreshed.");
            }
        }
    }

    /**
     * 全てのアプリケーションコンテキストをクローズする。
     * 
     */
    static public void close() {
        closeUserContext();
        closeComponentContext();
        closeSystemContext();
    }

    static public void closeComponentContext(String componentID) {

        synchronized (lockObject) {
            if (componentLoader != null)
                componentLoader.unloadServiceComponent(componentID);
        }
    }

    static public void closeComponentContext() {

        synchronized (lockObject) {
            if (componentLoader != null)
                componentLoader.unloadServiceComponents();
        }
    }

    static public void closeSystemContext() {

        synchronized (lockObject) {
            if (systemContext != null) {
                systemContext.close();
                systemContext = null;
                log.info("SystemContext was closed.");
                return;
            }
        }
    }

    static public void closeUserContext() {

        synchronized (lockObject) {
            if (userContext != null) {
                userContext.close();
                userContext = null;
                log.info("UserContext was closed.");
                return;
            }
        }
    }

    /**
     * アプリケーションコンテキストの取得
     * 
     * @return ApplicationContextオブジェクト
     */
    static public ApplicationContext getContext() {
        ConfigurableApplicationContext context;

        if (userContext != null)
            context = userContext;
        else if (systemContext != null)
            context = systemContext;
        else
            return null;

        return context;
    }

    /**
     * アプリケーションコンテキストからPOJOの取得
     * 
     * @param name
     *            コンテキストからPOJOの取得
     * @return 該当するJavaオブジェクト
     */
    public static Object getBean(String name) {

        ConfigurableApplicationContext context;

        if (userContext != null)
            context = userContext;
        else if (systemContext != null)
            context = systemContext;
        else
            return null;

        return context.getBean(name);
    }

    /**
     * アプリケーションコンテキストからPOJOの取得
     * 
     * @param name
     *            取得するJavaオブジェクトの名前
     * @param beanClass
     *            取得するBeanクラス
     * @return 該当するJavaオブジェクト
     */
    public static Object getBean(String name, Class beanClass) {

        ConfigurableApplicationContext context;

        if (userContext != null)
            context = userContext;
        else if (systemContext != null)
            context = systemContext;
        else
            return null;

        return context.getBean(name, beanClass);
    }

    static public boolean shutdownComponents() {
        synchronized (lockObject) {
            try {
                if (componentLoader != null)
                    return componentLoader.shutdownComponents();
                else
                    return true;

            } catch (Exception e) {
                log.error(e.getMessage(), e);
                return false;
            }
        }
    }

    static public boolean startComponents(boolean onlyAutoStartComponent) {
        synchronized (lockObject) {
            try {
                if (componentLoader != null)
                    return componentLoader.startComponents(onlyAutoStartComponent);
                return true;

            } catch (Exception e) {
                log.error(e.getMessage(), e);
                return false;
            }
        }
    }

    static private Log log = LogFactory.getLog(SystemContext.class);
}