package application.startup;

import java.io.File;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.StringTokenizer;

/**
 * 環境設定を行うための基本クラスであり、システムプロパティ及びクラスローダの初期化を行う。
 * 
 * @author masanori.matsuoka@netconscious.jp
 * 
 */
public class EnvironmentConfigurator {

	static private boolean silentMode;

	static private Object mainClassInstance;

	static public boolean initialize(String systemPropsFile, boolean blLoadClass) {
		System.setProperty(SystemProps.PROPS_SYSTEM_PROPERTIES_XML,
				systemPropsFile);
		return initialize(blLoadClass);
	}

	/**
	 * 環境の初期化を行う。
	 * 
	 * @return 初期化の結果を true 及び false で返す。
	 */
	static public boolean initialize(boolean blLoadClass) {

		try {

			if (SystemProps.getJavaHome() == null) {
				System.out
						.println("[JAVA_HOME] or [system.java.home] is not found.");
				return false;
			}

			Properties props = SystemProps.loadSystemProperties();
			if (props == null)
				return false;

			// Java 1.4対応 shumi
			silentMode = "true".equalsIgnoreCase(System
					.getProperty(SystemProps.PROPS_SYS_SILENT_MODE));
			// silentMode =
			// Boolean.parseBoolean(System.getProperty(SystemProps.PROPS_SYS_SILENT_MODE));

			if (!silentMode
					&& System.getProperties().containsKey(
							SystemProps.PROPS_APPLICATION_DESC)) {
				System.out.println(System
						.getProperty(SystemProps.PROPS_APPLICATION_DESC));
				System.out.println(System
						.getProperty(SystemProps.PROPS_APPLICATION_NAME)
						+ " is starting...");
			}

			// home
			if (!silentMode) {
				System.out.println("[JAVA_HOME] system.java.home="
						+ System.getProperty(SystemProps.PROPS_SYS_JAVA_HOME));
				System.out
						.println("[APPLICATION_HOME] application.home="
								+ System
										.getProperty(SystemProps.PROPS_APPLICATION_HOME));
			}

			// system properties
			if (!silentMode) {
				List propsList = new ArrayList(props.keySet());
				Collections.sort(propsList);
				Iterator itKeys = propsList.iterator();
				while (itKeys.hasNext()) {
					Object key = itKeys.next();
					System.out.println("[EnvironmentConfigurator] "
							+ key.toString() + "=" + props.get(key).toString());
				}
			}

			// check working directories.
			if (!silentMode)
				System.out
						.println("[EnvironmentConfigurator]Checking working directories...");

			boolean prepareWorkDirResult = prepareWorkDirectory();
			if (!silentMode && prepareWorkDirResult)
				System.out
						.println("[EnvironmentConfigurator]The working directories was prepared.");
			if (!prepareWorkDirResult)
				System.out
						.println("Warning - The working directories was not prepared!");

			// load the parameters.
			if (!silentMode)
				System.out
						.println("[EnvironmentConfigurator]Creating the ClassLoader ...");

			if (blLoadClass) {
				ClassLoader cl = getClassLoader();
				if (cl != null)
					Thread.currentThread().setContextClassLoader(cl);
				else {
					System.out
							.println("Error - The ClassLoader initializing failed.");
					return false;
				}
			}

			if (!silentMode)
				System.out
						.println("[EnvironmentConfigurator]The ClassLoader was created and set to ContextClassLoader.");

			return true;

		} catch (Exception e) {
			if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
				e.printStackTrace();
			}
			return false;
		}
	}

	/**
	 * 作業用ディレクトリの生成
	 * 
	 */
	static public boolean prepareWorkDirectory() {

		boolean result;
		result = createDirectory(System
				.getProperty(SystemProps.PROPS_SYS_WORK_PATH));
		result = result
				& createDirectory(System
						.getProperty(SystemProps.PROPS_SYS_WORK_LOG_PATH));
		result = result
				& createDirectory(System
						.getProperty(SystemProps.PROPS_USER_CLASSES_PATH));
		result = result
				& createDirectory(System
						.getProperty(SystemProps.PROPS_USER_LIB_PATH));
		result = result
				& createDirectory(System
						.getProperty(SystemProps.PROPS_USER_BEANS_PATH));
		result = result
				& createDirectory(System
						.getProperty(SystemProps.PROPS_USER_PROPERTIES_PATH));
		return result;
	}

	/**
	 * 指定されたディレクトリの生成
	 */
	static protected boolean createDirectory(String path) {
		if (path != null) {
			File directory = new File(path);
			if (!directory.exists())
				return directory.mkdirs();
			else
				return true;
		}
		return false;
	}

	public static Object startMainClass(String[] args) {

		ClassLoader cl = Thread.currentThread().getContextClassLoader();
		String mainClass = System
				.getProperty(SystemProps.PROPS_APPLICATION_MAIN_CLASS);
		String methodName = System
				.getProperty(SystemProps.PROPS_APPLICATION_START_CLASS_METHOD);
		return EnvironmentConfigurator.invokeMethod(cl, mainClass, methodName,
				args, false);
	}

	public static Object shutdownMainClass() {

		String mainClassName = System
				.getProperty(SystemProps.PROPS_APPLICATION_MAIN_CLASS);
		String methodName = System
				.getProperty(SystemProps.PROPS_APPLICATION_SHUTDOWN_CLASS_METHOD);

		try {
			if (!silentMode)
				System.out.println("[EnvironmentConfigurator]Invoking ["
						+ mainClassName + "." + methodName + "]");

			Class mainClass = Thread.currentThread().getContextClassLoader()
					.loadClass(mainClassName);
			Class[] paramTypes = new Class[] {};

			Method main = mainClass.getMethod(methodName, paramTypes);
			Object[] args = new Object[] {};
			Object result = main.invoke(mainClassInstance, args);
			return result;

		} catch (Exception e) {
			if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
				e.printStackTrace();
			}
			return null;
		}

	}

	/**
	 * 指定したクラスのメソッドをリフレクションを使用して実行する。
	 * 
	 * @param cl
	 *            クラスローダ
	 * @param mainClassname
	 *            クラス名
	 * @param methodName
	 *            メソッド名
	 * @return 実行結果
	 */
	protected static Object invokeMethod(ClassLoader cl, String mainClassName,
			String methodName, String[] params, boolean bStaticMedhod) {

		if (cl == null || mainClassName == null || methodName == null)
			return null;

		Object result = null;
		try {
			if (!silentMode)
				System.out.println("[EnvironmentConfigurator]Invoking ["
						+ mainClassName + "." + methodName + "]");

			Class mainClass = cl.loadClass(mainClassName);
			Class[] paramTypes = new Class[] { String[].class };

			Method main = mainClass.getMethod(methodName, paramTypes);
			Object[] args = new Object[] { params };
			if (!bStaticMedhod) {
				mainClassInstance = mainClass.newInstance();
				result = main.invoke(mainClassInstance, args);
			} else
				result = main.invoke(null, args);
			return result;

		} catch (Exception e) {
			if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
				e.printStackTrace();
			}
			return null;
		}
	}

	/**
	 * 与えられたクラスパスリストからクラスローダを生成する。
	 * 
	 * @param classpathList
	 *            クラスローダを生成するためのクラスパスのFileオブジェクトのリスト
	 * @return クラスローダ
	 */
	public static ClassLoader createClassLoader(List classpathList) {

		ClassLoader parent = Thread.currentThread().getContextClassLoader();
		if (parent == null)
			parent = EnvironmentConfigurator.class.getClassLoader();
		if (parent == null)
			parent = ClassLoader.getSystemClassLoader();

		List url_list = new ArrayList(classpathList.size());
		Iterator itr_classpath = classpathList.iterator();

		try {
			while (itr_classpath.hasNext()) {
				File currentFile = (File) itr_classpath.next();
				url_list.add(currentFile.toURL());
			}
		} catch (MalformedURLException e) {
			if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
				e.printStackTrace();
			}
			return null;
		}

		URL[] urls = (URL[]) url_list.toArray(new URL[0]);
		ClassLoader cl = URLClassLoader.newInstance(urls, parent);
		return cl;
	}

	/**
	 * ファイルパス文字列からクラスパスを生成し、クラスパスリストに追加する。
	 * クラスパスは、ファイルパスのディレクトリとその直下のディレクトリ及び、それらの ディレクトリ内のjarファイル及びzipファイルから生成される。
	 * 
	 * @param classpathList
	 *            追加先のクラスパスリスト
	 * @param filepath
	 *            ファイルパス
	 */
	private static void createClasspathList(List classpathList, String filepath) {

		StringTokenizer token = new StringTokenizer(filepath,
				File.pathSeparator);

		try {
			while (token.hasMoreTokens()) {

				File file = new File(token.nextToken());
				if (file.exists()) {
					File classpath_file = file.getCanonicalFile();

					if (classpath_file.isDirectory()) {

						if (!classpathList.contains(classpath_file)) {
							classpathList.add(classpath_file);
						}

						File[] files = classpath_file.listFiles();
						for (int fileCount = 0; fileCount < files.length; fileCount++) {
							File child_classpath_file = files[fileCount]
									.getCanonicalFile();
							if (child_classpath_file.isDirectory()) {
								createClasspathList(classpathList,
										child_classpath_file.getAbsolutePath());
							} else if (child_classpath_file.getName().matches(
									".*\\.jar")
									|| child_classpath_file.getName().matches(
											".*\\.zip")) {
								if (!classpathList
										.contains(child_classpath_file)) {
									classpathList.add(child_classpath_file);
								}
							}

						}

					} else if (file.isFile()) {
						if (file.getName().matches(".*\\.jar")
								|| file.getName().matches(".*\\.zip")) {
							if (!classpathList.contains(classpath_file)) {
								classpathList.add(classpath_file);
							}
						}

					}
				}
			}
		} catch (Exception e) {
			if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
				e.printStackTrace();
			}
		}

	}

	public static ClassLoader getClassLoader() {

		try {
			List classpath_list = new ArrayList();

			if (System.getProperties().containsKey(
					SystemProps.PROPS_SYS_LIB_PATH))
				createClasspathList(classpath_list, System
						.getProperty(SystemProps.PROPS_SYS_LIB_PATH));

			if (System.getProperties().containsKey(
					SystemProps.PROPS_USER_LIB_PATH))
				createClasspathList(classpath_list, System
						.getProperty(SystemProps.PROPS_USER_LIB_PATH));

			if (System.getProperties().containsKey(
					SystemProps.PROPS_SYS_CLASSES_PATH))
				createClasspathList(classpath_list, System
						.getProperty(SystemProps.PROPS_SYS_CLASSES_PATH));

			ClassLoader cl = createClassLoader(classpath_list);
			return cl;

		} catch (Exception e) {
			if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
				e.printStackTrace();
			}
			return null;
		}

	}

}
