package application.util.monitor;

import java.io.File;
import java.io.FileFilter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

public class DirectoryMonitor implements ChangeMonitor, Runnable {

    private static Log log = LogFactory.getLog(DirectoryMonitor.class);
    public static int MIN_POLLING_INTERVAL = 1000;
    public static int DEFUALT_POLLING_INTERVAL = 5000;
    public static int DEFAULT_THREAD_PRIORITY = Thread.MIN_PRIORITY;
    public static int FILTER_FOR_FILE = 1;
    public static int FILTER_FOR_DIRECTORY = 2;
    public static int FILTER_FOR_ALL = FILTER_FOR_FILE | FILTER_FOR_DIRECTORY;

    private File directory = null;
    private List listeners = new ArrayList();
    private int pollingInterval = DEFUALT_POLLING_INTERVAL;
    private Map latestFileCache = new HashMap();
    private boolean bExecute = true;
    private boolean bIgnoreCurrentFiles = true;
    private FileFilter fileFilter = null;

    public DirectoryMonitor() {
    }

    public void start() {
        if (this.directory == null)
            return;

        Thread monitorThread = new Thread(this);
        monitorThread.setDaemon(true);
        monitorThread.setPriority(DEFAULT_THREAD_PRIORITY);
        monitorThread.start();
        log.info("DirectoryChangeMonitor has started.");
    }

    public void stop() {
        this.bExecute = false;
    }

    public void run() {
        if (bIgnoreCurrentFiles)
            createLatestFileCache();

        while (bExecute) {
            log.info("Checking the modified files in [" + this.directory.getAbsolutePath() + "].");
            findChangedFiles();
            try {
                Thread.sleep(this.pollingInterval);
            } catch (InterruptedException e) {
            }
        }
        log.info("DirectoryChangeMonitor has stopeed.");

    }

    protected void findChangedFiles() {

        if (this.directory == null || this.latestFileCache == null)
            return;

        Map workingFileCache = new HashMap(latestFileCache);

        // check deleted files in the directory
        Iterator iFileCache = workingFileCache.keySet().iterator();
        while (iFileCache.hasNext()) {
            File file = (File) iFileCache.next();
            if (!file.exists()) {
                log.info("[" + file.getName() + "] was deleted.");
                fireChanged(ChangeMonitor.CM_MODIFIED | ChangeMonitor.CM_DELETED | ChangeMonitor.CM_MOVED, file);
                latestFileCache.remove(file);
            }
        }

        // check created files and modified files in the directory.
        File[] files;
        if (this.fileFilter != null)
            files = this.directory.listFiles(this.fileFilter);
        else
            files = this.directory.listFiles();

        for (int index = 0; index < files.length; index++) {
            File file = files[index];
            if (latestFileCache.containsKey(file)) {
                Long cachedLastModified = (Long) latestFileCache.get(file);
                if (cachedLastModified.longValue() != file.lastModified()) {
                    log.info("[" + file.getName() + "] was modified.");
                    fireChanged(ChangeMonitor.CM_MODIFIED, file);
                    latestFileCache.put(file, new Long(file.lastModified()));
                }
            }
            else {
                log.info("[" + file.getName() + "] was created.");
                fireChanged(ChangeMonitor.CM_MODIFIED | ChangeMonitor.CM_CREATED, file);
                latestFileCache.put(file, new Long(file.lastModified()));
            }

        }

    }

    protected void createLatestFileCache() {

        if (this.directory == null)
            return;

        this.latestFileCache.clear();
        File[] files;
        if (this.fileFilter != null)
            files = this.directory.listFiles(this.fileFilter);
        else
            files = this.directory.listFiles();

        for (int index = 0; index < files.length; index++) {
            File file = files[index];
            this.latestFileCache.put(file, new Long(file.lastModified()));
        }
    }

    public void addChangeMonitorListener(ChangeMonitorListener listener) {
        if (!listeners.contains(listener))
            listeners.add(listener);
    }

    public void fireChanged(int eventType, Object object) {

        if (object != null) {
            Iterator iListeners = listeners.iterator();
            while (iListeners.hasNext()) {
                ChangeMonitorListener listener = (ChangeMonitorListener) iListeners.next();
                listener.onChanged(eventType, object);
            }
        }
    }

    public File getDirectory() {
        return directory;
    }

    public void setDirectory(String directoryPath) {
        File directory = new File(directoryPath);
        if (directory != null && directory.exists() && directory.isDirectory())
            this.directory = directory;
        else
            this.directory = null;
    }

    public int getPollingInterval() {
        return pollingInterval;
    }

    public void setPollingInterval(int pollingInterval) {
        this.pollingInterval = (pollingInterval > MIN_POLLING_INTERVAL) ? pollingInterval : MIN_POLLING_INTERVAL;
    }

    public void setFileFilter(String regexPattern, int filterType) {

        Pattern fileFilterPattern;
        try {
            fileFilterPattern = Pattern.compile(regexPattern);
        } catch (Exception e) {
            fileFilterPattern = null;
        }
        ;

        // Create the filter specifing a pattern and a filter type.
        this.fileFilter = new DirectoryChangeMonitorFilter(fileFilterPattern, filterType);

    }

    public void setIgnoreCurrentFiles(boolean bIgnoreCurrentFiles) {
        this.bIgnoreCurrentFiles = bIgnoreCurrentFiles;
    }

}
