package daruma.bus.processor.csv;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.Properties;

import javax.xml.transform.OutputKeys;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xml.serializer.OutputPropertiesFactory;

import application.startup.SystemProps;

import daruma.bus.message.Message;
import daruma.bus.message.StandardPayload;
import daruma.bus.processor.Processor;
import daruma.bus.processor.ProcessorException;
import daruma.bus.processor.xslt.XslTransformer;

public class CSVFileReader implements Processor {

	static private Log log = LogFactory.getLog(CSVFileReader.class);

	// CSVFileAdapter Properties
	static public final String PROPS_CSVX_NAMESPACE = "csvfilereader.csvx.namespace";

	static public final String PROPS_CSVX_NAMESPACE_PREFIX = "csvfilereader.csvx.namespace.prefix";

	static public final String PROPS_INPUT_ENCODING = "csvfilereader.input.encoding";

	static public final String PROPS_CSVX_ENCODING = "csvfilereader.output.encoding";

	static public final String PROPS_CSVX_PRETTYPRINT = "csvfilereader.csvx.prettyprint";

	static public final String PROPS_CSVFILE_COMMENT = "csvfilereader.csvfile.comment";

	static public final String PROPS_CSVFILE_SEPARATOR = "csvfilereader.csvfile.separator";

	static public final String PROPS_CSVFILE_PARENTHESIS = "csvfilereader.csvfile.parenthesis";

	static public final String PROPS_XSLT_COPY_TEMPLATE = "csvfilereader.xslt.copy.template";

	static public final String DEFAULT_CSVX_NAMESPACE = SystemProps.get(
			PROPS_CSVX_NAMESPACE, Const.NS_XMLNS_PRE + ":" + Const.NS_CSVX_PRE + "=\""
					+ Const.NS_CSVX_URI + "\"");

	static public final String DEFAULT_CSVX_NAMESPACE_PREFIX = SystemProps.get(
			PROPS_CSVX_NAMESPACE_PREFIX, Const.NS_CSVX_PRE);

	static public final String DEFAULT_INPUT_ENCODING = SystemProps.get(
			PROPS_INPUT_ENCODING, "UTF-8");

	static public final String DEFAULT_CSVX_ENCODING = SystemProps.get(
			PROPS_CSVX_ENCODING, "UTF-8");

	static public final String DEFAULT_CSVX_PRETTYPRINT = SystemProps.get(
			PROPS_CSVX_PRETTYPRINT, "false");

	static public final String DEFAULT_CSVFILE_COMMENT = SystemProps.get(
			PROPS_CSVFILE_COMMENT, "#");

	static public final String DEFAULT_CSVFILE_SEPARATOR = SystemProps.get(
			PROPS_CSVFILE_SEPARATOR, ",");

	static public final String DEFAULT_CSVFILE_PARENTHESIS = SystemProps.get(
			PROPS_CSVFILE_PARENTHESIS, "\"");

	static public final String DEFAULT_XSLT_COPY_TEMPLATE = SystemProps.get(
			PROPS_XSLT_COPY_TEMPLATE, "");

	private String CSVX_NAMESPACE = DEFAULT_CSVX_NAMESPACE;

	private String CSVX_PREFIX = DEFAULT_CSVX_NAMESPACE_PREFIX;

	private String EL_DATA = CSVX_PREFIX + ":data";

	private String EL_ROW = CSVX_PREFIX + ":row";

	private String EL_COLUMN = CSVX_PREFIX + ":column";

	private String xsltCopyTemplate = DEFAULT_XSLT_COPY_TEMPLATE;

	private char comment = DEFAULT_CSVFILE_COMMENT.charAt(0);

	private char separator = DEFAULT_CSVFILE_SEPARATOR.charAt(0);

	private char parenthesis = DEFAULT_CSVFILE_PARENTHESIS.charAt(0);

	private String inputEncoding = DEFAULT_INPUT_ENCODING;

	private String outputEncoding = DEFAULT_CSVX_ENCODING;

	private boolean prettyPrintMode = Boolean.valueOf(DEFAULT_CSVX_PRETTYPRINT)
			.booleanValue();

	private File workDir = new File(SystemProps.get(
			SystemProps.PROPS_SYS_WORK_PATH, "./work"));

	public void setWorkDir(File directory) {
		this.workDir = directory;
	}

	public File getWorkDir() {
		return this.workDir;
	}

	/**
	 * CSVファイルをcsvxのXML形式にコンバートする。
	 * 
	 * @param in
	 * @param inputCharSet
	 * @param out
	 * @param outputCharSet
	 * @param separator
	 * @param parenthesis
	 * @param hasHeader
	 * @param prettyPrint
	 * @param workDir
	 * @throws CsvMessageException
	 */
	protected void csv2xml(InputStream in, String inputCharSet,
			OutputStream out, String outputCharSet, char separator,
			char parenthesis, char comment, boolean prettyPrint)
			throws ProcessorException {

		if (in == null || out == null) {
			throw new CSVFileReaderException(
					"A source file or a traget file was not found.");
		}

		try {
			BufferedReader bufferedReader = null;
			if (inputCharSet != null) {
				InputStreamReader reader = new InputStreamReader(in,
						inputCharSet);
				bufferedReader = new BufferedReader(reader);
			} else {
				InputStreamReader reader = new InputStreamReader(in);
				bufferedReader = new BufferedReader(reader);
			}

			File tempFile = File.createTempFile(CSVFileReader.class.getName(),
					".xml", this.workDir);
			OutputStream tempOut = new FileOutputStream(tempFile);
			BufferedWriter bufferedWriter = new BufferedWriter(
					new OutputStreamWriter(tempOut, outputEncoding));

			if (bufferedReader != null || bufferedWriter != null) {

				bufferedWriter.write("<?xml version=\"1.0\" encoding=\"");
				bufferedWriter.write(outputEncoding);
				bufferedWriter.write("\"?>");
				bufferedWriter.write("<");
				bufferedWriter.write(EL_DATA);
				bufferedWriter.write(" ");
				bufferedWriter.write(CSVX_NAMESPACE);
				bufferedWriter.write(">");

				String[] headers = null;
				String line = bufferedReader.readLine();
				//BOM対応
				if(line.charAt(0) == '\uFEFF')
				{
				  line = line.substring(1);
				}
				if (line.charAt(0) == comment) {
					line = line.substring(1);
					headers = MispSchemaUtil.getColumns(line, separator, parenthesis);
				}
				int rowCount = 1;
				line = bufferedReader.readLine();

				while (line != null) {
					if ((line.trim().length() > 0)
							&& (line.charAt(0) != comment)) {
						String[] items = MispSchemaUtil.getColumns(line, separator,
								parenthesis);

						bufferedWriter.write("<");
						bufferedWriter.write(EL_ROW);
						bufferedWriter.write(" count=\"");
						bufferedWriter.write(String.valueOf(rowCount));
						bufferedWriter.write("\">");
						for (int index = 0; index < items.length; index++) {
							if (headers != null && (index < headers.length)
									&& (headers[index] != null)) {
								bufferedWriter.write("<");
								bufferedWriter.write(EL_COLUMN);
								bufferedWriter.write(" name=\"");
								bufferedWriter.write(headers[index]);
								bufferedWriter.write("\" pos=\"");
								bufferedWriter.write(String.valueOf(index + 1));
								bufferedWriter.write("\">");
							} else {
								bufferedWriter.write("<");
								bufferedWriter.write(EL_COLUMN);
								bufferedWriter.write(" pos=\"");
								bufferedWriter.write(String.valueOf(index + 1));
								bufferedWriter.write("\">");
							}
							bufferedWriter.write(MispSchemaAccess
									.escapeXmlString(items[index]));
							bufferedWriter.write("</");
							bufferedWriter.write(EL_COLUMN);
							bufferedWriter.write(">");
						}
						bufferedWriter.write("</");
						bufferedWriter.write(EL_ROW);
						bufferedWriter.write(">");
					}
					line = bufferedReader.readLine();

					rowCount = rowCount + 1;
				}

				bufferedWriter.write("</");
				bufferedWriter.write(EL_DATA);
				bufferedWriter.write(">");

				bufferedWriter.flush();
				bufferedWriter.close();
				bufferedReader.close();

				BufferedInputStream bufferIn = new BufferedInputStream(
						new FileInputStream(tempFile));
				BufferedOutputStream bufferOut = new BufferedOutputStream(out);
				if (prettyPrint) {
					XslTransformer trans = new XslTransformer();
					Properties props = OutputPropertiesFactory
							.getDefaultMethodProperties("xml");
					props.setProperty(OutputKeys.INDENT, "yes");
					props.setProperty(OutputKeys.ENCODING, outputCharSet);
					props.setProperty(OutputKeys.OMIT_XML_DECLARATION, "no");
					props.setProperty(OutputKeys.METHOD, "xml");
					trans.setOutputProperties(props);
					trans.setXsltFile(xsltCopyTemplate);
					trans.process(bufferIn, bufferOut);

				} else {
					int len;
					byte[] buf = new byte[8192];
					while ((len = bufferIn.read(buf)) != -1)
						bufferOut.write(buf, 0, len);
				}

				bufferOut.flush();
				bufferIn.close();
				bufferOut.close();
				tempFile.delete();

				return;
			} else {
				throw new CSVFileReaderException(
						"StreamReader or StreamWrtier was not initialized.");
			}

		} catch (Exception e) {
			log.debug(e.getMessage(), e);
			ProcessorException ex = new CSVFileReaderException(e.getMessage());
			ex.setStackTrace(e.getStackTrace());
			throw ex;
		}
	}

	public boolean isPrettyPrintMode() {
		return prettyPrintMode;
	}

	public void setPrettyPrintMode(boolean prettyPrintMode) {
		this.prettyPrintMode = prettyPrintMode;
	}

	public void process(Message message) throws ProcessorException {
		try {

			char separator = this.separator;
			char parenthesis = this.parenthesis;
			char comment = this.comment;
			String inputEncoding = this.inputEncoding;
			String outputEncoding = this.outputEncoding;
			boolean prettyPrintMode = this.prettyPrintMode;
			//String xsltCopyTemplate = this.xsltCopyTemplate;

			// check parameters
			if (message.getProperties().containsKey(PROPS_CSVFILE_SEPARATOR))
				separator = message.getProperties().getProperty(
						PROPS_CSVFILE_SEPARATOR).charAt(0);
			if (message.getProperties().containsKey(PROPS_CSVFILE_PARENTHESIS))
				parenthesis = message.getProperties().getProperty(
						PROPS_CSVFILE_PARENTHESIS).charAt(0);
			if (message.getProperties().containsKey(PROPS_CSVFILE_COMMENT))
				comment = message.getProperties().getProperty(
						PROPS_CSVFILE_COMMENT).charAt(0);
			if (message.getProperties().containsKey(PROPS_INPUT_ENCODING))
				inputEncoding = message.getProperties().getProperty(
						PROPS_INPUT_ENCODING);
			if (message.getProperties().containsKey(PROPS_CSVX_ENCODING))
				outputEncoding = message.getProperties().getProperty(
						PROPS_CSVX_ENCODING);
			if (message.getProperties().containsKey(PROPS_CSVX_PRETTYPRINT))
				prettyPrintMode = Boolean.valueOf(
						message.getProperties().getProperty(
								PROPS_CSVX_PRETTYPRINT)).booleanValue();
			if (message.getProperties().containsKey(PROPS_XSLT_COPY_TEMPLATE))
				xsltCopyTemplate = message.getProperties().getProperty(
						PROPS_XSLT_COPY_TEMPLATE);

			File csvxFile = File.createTempFile(CSVFileReader.class.getName(),
					".xml", this.workDir);
			OutputStream out = new FileOutputStream(csvxFile);
			csv2xml(message.getPayload().getInputStream(), inputEncoding, out,
					outputEncoding, separator, parenthesis, comment,
					prettyPrintMode);

			StandardPayload payload = new StandardPayload();
			payload.attach(csvxFile);
			message.setPayload(payload);

		} catch (Exception e) {
			log.debug(e.getMessage(), e);
			ProcessorException ex = new CSVFileReaderException(e.getMessage());
			ex.fillInStackTrace();
			throw ex;

		}
	}

	public String getInputEncoding() {
		return inputEncoding;
	}

	public void setInputEncoding(String fileEncoding) {
		this.inputEncoding = fileEncoding;
	}

	// public boolean isHasHeader() {
	// return hasHeader;
	// }
	//
	// public void setHasHeader(boolean hasHeader) {
	// this.hasHeader = hasHeader;
	// }

	public char getParenthesis() {
		return parenthesis;
	}

	public void setParenthesis(char parenthesis) {
		this.parenthesis = parenthesis;
	}

	public char getSeparator() {
		return separator;
	}

	public void setSeparator(char separator) {
		this.separator = separator;
	}

	public String getOutputEncoding() {
		return outputEncoding;
	}

	public void setOutputEncoding(String xmlEncoding) {
		this.outputEncoding = xmlEncoding;
	}

	public String getXsltCopyTemplate() {
		return xsltCopyTemplate;
	}

	public void setXsltCopyTemplate(String xsltCopyTemplate) {
		this.xsltCopyTemplate = xsltCopyTemplate;
	}

	public char getComment() {
		return comment;
	}

	public void setComment(char comment) {
		this.comment = comment;
	}
}
