package daruma.bus.processor.csv;

import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import application.startup.SystemProps;

import daruma.bus.message.Message;
import daruma.bus.message.StandardPayload;
import daruma.bus.processor.Processor;
import daruma.bus.processor.ProcessorException;
import daruma.bus.processor.xslt.XslTransformer;

public class CSVFileWriter implements Processor {

    static private Log log = LogFactory.getLog(CSVFileWriter.class);

    static public final String PROPS_OUTPUT_ENCODING = "csvfilewriter.output.encoding";
    static public final String PROPS_OUTPUT_PATH = "csvfilewriter.output.path";
    static public final String PROPS_XSLT_CSVX2CSV_TEMPLATE = "csvfilewriter.xslt.csvx2csv.template";
    static public final String VALUE_OUTPUT_PATH_STDOUT = "csvfilewriter.value.output.path.stdout";

    static public final String DEFAULT_OUTPUT_ENCODING = SystemProps.get(PROPS_OUTPUT_ENCODING, System
            .getProperty(SystemProps.PROPS_SYS_OS_ENCODING));
    static public final String DEFAULT_XSLT_CSVX2CSV_TEMPLATE = SystemProps.get(PROPS_XSLT_CSVX2CSV_TEMPLATE, "");
    static public final String DEFAULT_OUTPUT_PATH = SystemProps.get(PROPS_OUTPUT_PATH, VALUE_OUTPUT_PATH_STDOUT);

    private String csvx2csvTempalte = DEFAULT_XSLT_CSVX2CSV_TEMPLATE;
    private String outputEncoding = DEFAULT_OUTPUT_ENCODING;
    private String outputPath = DEFAULT_OUTPUT_PATH;

    private File workDir = new File(SystemProps.get(SystemProps.PROPS_SYS_WORK_PATH, "./work"));

    /**
     * CSVXからCSVファイルを生成する。
     */
    public void process(Message message) throws ProcessorException {

        try {
            String csvx2csvTempate = this.csvx2csvTempalte;
            String outputEncoding = this.outputEncoding;
            String outputPath = this.outputPath;

            if (message.getProperties().containsKey(PROPS_OUTPUT_ENCODING))
                outputEncoding = message.getProperties().getProperty(PROPS_OUTPUT_ENCODING);
            if (message.getProperties().containsKey(PROPS_XSLT_CSVX2CSV_TEMPLATE))
                csvx2csvTempate = message.getProperties().getProperty(PROPS_XSLT_CSVX2CSV_TEMPLATE);
            if (message.getProperties().containsKey(PROPS_OUTPUT_PATH))
                outputPath = message.getProperties().getProperty(PROPS_OUTPUT_PATH);

            File payloadFile = File.createTempFile(CSVFileWriter.class.getName(), ".csv", this.workDir);
            InputStream in = message.getPayload().getInputStream();
            OutputStream out = new FileOutputStream(payloadFile);
            xml2csv(csvx2csvTempate, in, out, outputEncoding);
            in.close();
            out.flush();
            out.close();
            StandardPayload messageData = new StandardPayload();
            messageData.attach(payloadFile);
            message.setPayload(messageData);

            String internalCharset = System.getProperty(SystemProps.PROPS_SYS_INTERNAL_ENCODING);

            if (outputPath.equals(VALUE_OUTPUT_PATH_STDOUT)) {

                // outout
                InputStream filein = messageData.getInputStream();
                InputStreamReader reader = new InputStreamReader(filein, internalCharset);
                OutputStreamWriter writer = new OutputStreamWriter(System.out, outputEncoding);

                int len;
                char[] buf = new char[1024];
                while ((len = reader.read(buf)) != -1) {
                    writer.write(buf, 0, len);
                }
                writer.flush();
                //writer.close();
                reader.close();
                filein.close();

            }
            else {

                File resultFile = new File(outputPath);
                resultFile.delete();
                if (resultFile.createNewFile()) {
                    messageData.save(resultFile, internalCharset, outputEncoding);
                }
                else {
                    ProcessorException ex = new CSVFileWriterException("An output file was not created.");
                    ex.fillInStackTrace();
                }
            }

        }

        catch (Exception e) {
            log.debug(e.getMessage(), e);
            ProcessorException ex = new CSVFileWriterException(e.getMessage());
            ex.fillInStackTrace();
            throw ex;
        }
    }

    /**
     * csvx形式のXMLをCSVに変換する。
     * 
     * @param in
     * @param out
     * @param outputCharSet
     * @param separator
     * @param parenthesis
     * @param hasHeader
     * @param workDir
     * @throws CsvMessageException
     */
    protected void xml2csv(String xsltTempate, InputStream in, OutputStream out, String outputCharSet) throws ProcessorException {

        try {
            XslTransformer transformer = new XslTransformer();

            transformer.setXsltFile(xsltTempate);
            transformer.setOmitXmlDeclaration("yes");
            transformer.setIndent("no");
            //transformer.setXmlEncoding(outputCharSet);
            transformer.process(in, out);

        } catch (Exception e) {
            ProcessorException ex = new CSVFileWriterException(e.getMessage());
            ex.fillInStackTrace();
            throw ex;
        }
    }

    public void setWorkDir(File directory) {
        this.workDir = directory;
    }

    public File getWorkDir() {
        return this.workDir;
    }

    public String getCsvx2csvTempalte() {
        return csvx2csvTempalte;
    }

    public String getOutputEncoding() {
        return outputEncoding;
    }

    public void setCsvx2csvTempalte(String csvx2csvTempalte) {
        this.csvx2csvTempalte = csvx2csvTempalte;
    }

    public void setOutputEncoding(String fileEncoding) {
        this.outputEncoding = fileEncoding;
    }

    public String getOutputPath() {
        return outputPath;
    }

    public void setOutputPath(String outputPath) {
        this.outputPath = outputPath;
    }

}
