package daruma.bus.processor.csv;

import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import application.startup.SystemProps;
import daruma.bus.message.Message;
import daruma.bus.message.StandardPayload;
import daruma.bus.processor.Processor;
import daruma.bus.processor.ProcessorException;

public class CSVSchemaReader implements Processor {
	static private Log log = LogFactory.getLog(CSVSchemaReader.class);

	static public final String PROPS_CSVSCHEMA_ENCODING = "csvschemareader.csvschema.encoding";

	static public final String PROPS_INPUT_ENCODING = "csvschemareader.csvfile.encoding";

	static public final String PROPS_CSVFILE_COMMENT = "csvschemareader.csvfile.comment";

	static public final String PROPS_CSVFILE_SEPARATOR = "csvschemareader.csvfile.separator";

	static public final String PROPS_CSVFILE_PARENTHESIS = "csvschemareader.csvfile.parenthesis";

	static public final String DEFAULT_CSVSCHEMA_ENCODING = SystemProps.get(
			PROPS_CSVSCHEMA_ENCODING, "UTF-8");

	static public final String DEFAULT_INPUT_ENCODING = SystemProps.get(
			PROPS_INPUT_ENCODING, "UTF-8");

	static public final String DEFAULT_CSVFILE_COMMENT = SystemProps.get(
			PROPS_CSVFILE_COMMENT, String.valueOf(MispSchemaAccess.DFS_DFT_COMMENT));

	static public final String DEFAULT_CSVFILE_SEPARATOR = SystemProps.get(
			PROPS_CSVFILE_SEPARATOR, String.valueOf(MispSchemaAccess.DFS_DFT_SEPARATOR));

	static public final String DEFAULT_CSVFILE_PARENTHESIS = SystemProps.get(
			PROPS_CSVFILE_PARENTHESIS, String.valueOf(MispSchemaAccess.DFS_DFT_PARENTHESIS));

	static public final String PROPS_BUILTIN_ELEMENT_DATATYPE = "mispschema.element.builtindatatype";

	static public final String PROPS_BUILTIN_ATTRIBUTE_DATATYPE = "mispschema.attribute.builtindatatype";

	private String builtinElementDataType = SystemProps.get(
			PROPS_BUILTIN_ELEMENT_DATATYPE, MispSchemaBuiltIn.dftBuiltinElementType);

	private String builtinAttributeDataType = SystemProps.get(
			PROPS_BUILTIN_ATTRIBUTE_DATATYPE, MispSchemaBuiltIn.dftBuiltinAttributeType);;

	
	private char comment = DEFAULT_CSVFILE_COMMENT.charAt(0);

	private char separator = DEFAULT_CSVFILE_SEPARATOR.charAt(0);

	private char parenthesis = DEFAULT_CSVFILE_PARENTHESIS.charAt(0);

	private String inputEncoding = DEFAULT_INPUT_ENCODING;

	private String outputEncoding = DEFAULT_CSVSCHEMA_ENCODING;

	private File workDir = new File(SystemProps.get(
			SystemProps.PROPS_SYS_WORK_PATH, "./work"));

	public CSVSchemaReader()
	{
		MispSchemaBuiltIn.setBuiltinElementType(builtinElementDataType);
		MispSchemaBuiltIn.setBuiltinAttributeType(builtinAttributeDataType);
	}

	public void setWorkDir(File directory) {
		this.workDir = directory;
	}

	public File getWorkDir() {
		return this.workDir;
	}

	/**
	 * CSVファイルをcsvschemaのXML形式にコンバートする。
	 * 
	 * @param in
	 * @param inputCharSet
	 * @param out
	 * @param outputCharSet
	 * @param separator
	 * @param parenthesis
	 * @param comment
	 * @throws CsvMessageException
	 */
	protected void csv2xml(InputStream in, String inputCharSet,
			OutputStream out, String outputCharSet, char separator,
			char parenthesis, char comment) throws ProcessorException {
		if (in == null || out == null) {
			throw new CSVSchemaReaderException(
					"A source file or a traget file was not found.");
		}

		MispSchemaAccess access = new MispSchemaAccess();
		InputStreamReader isr = null;
		if ((inputCharSet != null) && (inputCharSet.length() > 0)) {
			try {
				isr = new InputStreamReader(in, inputCharSet);
			} catch (UnsupportedEncodingException e) {
				throw new CSVSchemaReaderException(e.getMessage());
			}
		} else {
			isr = new InputStreamReader(in);
		}
		MispSchema schema = access.loadSchemaFromCsv(isr, separator,
				parenthesis, comment);
		if (schema != null) {
			String msg = schema.validate(true);
			if (msg.length() <= 0) {
				if(!access.saveSchemaToXml(schema, out, outputCharSet, outputCharSet))
				{
					throw new CSVSchemaReaderException(access.getLastError());
				}
			} else {
				throw new CSVSchemaReaderException(msg);
			}
		} else {
			throw new CSVSchemaReaderException(access.getLastError());
		}
	}


	public void process(Message message) throws ProcessorException {
		try {
			char separator = this.separator;
			char parenthesis = this.parenthesis;
			char comment = this.comment;
			String inputEncoding = this.inputEncoding;
			String outputEncoding = this.outputEncoding;

			// check parameters
			if (message.getProperties().containsKey(PROPS_CSVFILE_SEPARATOR))
				separator = message.getProperties().getProperty(
						PROPS_CSVFILE_SEPARATOR).charAt(0);
			if (message.getProperties().containsKey(PROPS_CSVFILE_PARENTHESIS))
				parenthesis = message.getProperties().getProperty(
						PROPS_CSVFILE_PARENTHESIS).charAt(0);
			if (message.getProperties().containsKey(PROPS_CSVFILE_COMMENT))
				comment = message.getProperties().getProperty(
						PROPS_CSVFILE_COMMENT).charAt(0);
			if (message.getProperties().containsKey(PROPS_INPUT_ENCODING))
				inputEncoding = message.getProperties().getProperty(
						PROPS_INPUT_ENCODING);
			if (message.getProperties().containsKey(PROPS_CSVSCHEMA_ENCODING))
				outputEncoding = message.getProperties().getProperty(
						PROPS_CSVSCHEMA_ENCODING);

			File csvschemaFile = File.createTempFile(CSVFileReader.class
					.getName(), ".xml", this.workDir);
			OutputStream out = new FileOutputStream(csvschemaFile);
			csv2xml(message.getPayload().getInputStream(), inputEncoding, out,
					outputEncoding, separator, parenthesis, comment);

			StandardPayload payload = new StandardPayload();
			payload.attach(csvschemaFile);
			message.setPayload(payload);

		} catch (Exception e) {
			log.debug(e.getMessage(), e);
			ProcessorException ex = new CSVSchemaReaderException(e.getMessage());
			ex.fillInStackTrace();
			throw ex;
		}
	}

	public String getInputEncoding() {
		return inputEncoding;
	}

	public void setInputEncoding(String fileEncoding) {
		this.inputEncoding = fileEncoding;
	}

	public char getParenthesis() {
		return parenthesis;
	}

	public void setParenthesis(char parenthesis) {
		this.parenthesis = parenthesis;
	}

	public char getSeparator() {
		return separator;
	}

	public void setSeparator(char separator) {
		this.separator = separator;
	}

	public String getOutputEncoding() {
		return outputEncoding;
	}

	public void setOutputEncoding(String xmlEncoding) {
		this.outputEncoding = xmlEncoding;
	}

	public char getComment() {
		return comment;
	}

	public void setComment(char comment) {
		this.comment = comment;
	}

}
