package daruma.bus.processor.csv;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.net.URL;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xerces.parsers.DOMParser;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import daruma.bus.processor.csv.MispSchema.MispSchemaComplexType;
import daruma.bus.processor.csv.MispSchema.MispSchemaAttribute;
import daruma.bus.processor.csv.MispSchema.MispSchemaElement;
import daruma.bus.processor.csv.MispSchema.MispSchemaImport;
import daruma.bus.processor.csv.MispSchema.MispSchemaNamespace;

/**
 * DaRuMaスキーマファイルをアクセスするためのクラス
 * 
 * @author shumi
 * 
 */
/**
 * @author shumi
 * 
 */
public class MispSchemaAccess {
	/**
	 * ロガー
	 */
	static private Log log = LogFactory.getLog(MispSchemaAccess.class);

	/**
	 * バージョン情報出力フラグ
	 */
	private boolean dfsOutputVersion = true;

	/**
	 * DFSスキーマの拡張子
	 */
	final public static String DFS_DFT_FILE_EXT = "dfs"; //$NON-NLS-1$

	/**
	 * デフォルトキャラクタセット
	 */
	final public static String DFS_DFT_CHARSET = "Shift_JIS"; //$NON-NLS-1$

	/**
	 * デフォルトCSVコメントキャラクタ
	 */
	final public static char DFS_DFT_COMMENT = '#';

	/**
	 * デフォルトCSVセパレータキャラクタ
	 */
	final public static char DFS_DFT_SEPARATOR = ',';

	/**
	 * デフォルトCSV括弧キャラクタ
	 */
	final public static char DFS_DFT_PARENTHESIS = '\"';

	static private final String DFS_HDR_ID = "ID"; //$NON-NLS-1$

	static private final String DFS_HDR_ANNOTATION = "Annotation"; //$NON-NLS-1$

	static private final String DFS_HDR_NS = "Namespace"; //$NON-NLS-1$

	static private final String DFS_HDR_TNS = "TargetNamespace"; //$NON-NLS-1$

	static private final String DFS_HDR_IMPORT = "Import"; //$NON-NLS-1$

	static private final String DFS_HDR_IMPORT_LOCATION = "location"; //$NON-NLS-1$

	static private final String DFS_HDR_FT = "FeatureType"; //$NON-NLS-1$

	static private final String DFS_HDR_ET = "ElementType"; //$NON-NLS-1$

	static private final String DFS_HDR_ET_EXT = "extension"; //$NON-NLS-1$

	static private final String DFS_HDR_ET_ATTR = "attribute"; //$NON-NLS-1$

	static private final String DFS_HDR_ET_EL = "element"; //$NON-NLS-1$

	/**
	 * DFSスキーマをロードする
	 * 
	 * @param fileName
	 *            ファイル名
	 * @param inputCharSet
	 *            入力文字セット
	 * @param separator
	 *            セパレータキャラクタ
	 * @param parenthesis
	 *            括弧キャラクタ
	 * @param comment
	 *            コメントキャラクタ
	 * @return 成功する場合は該当スキーマ、それ以外はnullを返す
	 */
	public MispSchema loadSchemaFromCsv(String fileName, String inputCharSet,
			char separator, char parenthesis, char comment) {
		try {
			clearError();
			InputStream fis;

			if (fileName.startsWith("http://")) { //$NON-NLS-1$
				URL url = new URL(fileName);
				fis = url.openStream();
			} else {
				File file = new File(fileName);
				fis = new FileInputStream(file);
			}

			InputStreamReader isr = null;
			if ((inputCharSet != null) && (inputCharSet.length() > 0)) {
				isr = new InputStreamReader(fis, inputCharSet);
			} else {
				isr = new InputStreamReader(fis);
			}
			MispSchema schema = loadSchemaFromCsv(isr, separator, parenthesis,
					comment);
			if (schema != null) {
				schema.setFilePath(fileName);
				schema.setDfsCharSet(inputCharSet);
				for (int cnt = 0; cnt < schema.getImportSize(); cnt++) {
					MispSchemaImport im = schema.getImport(cnt);
					loadImportSchema(schema, im, schema.getBaseDir(),
							inputCharSet, false);
				}
			}

			return schema;
		} catch (Exception e) {
			setLastError(-1, e.getMessage());
		}
		return null;
	}

	/**
	 * 
	 * @param isr
	 *            入力ストリームリーダー
	 * @param separator
	 *            セパレータキャラクタ
	 * @param parenthesis
	 *            括弧キャラクタ
	 * @param comment
	 *            コメントキャラクタ
	 * @return 成功する場合は該当スキーマ、それ以外はnullを返す
	 */
	public MispSchema loadSchemaFromCsv(InputStreamReader isr, char separator,
			char parenthesis, char comment) {
		try {
			clearError();
			BufferedReader brd = new BufferedReader(isr);

			MispSchema schema = new MispSchema();

			String line = brd.readLine();
			// BOM対応
			if (line.charAt(0) == '\uFEFF') {
				line = line.substring(1);
			}
			int rowCount = 1;
			String currHeader = ""; //$NON-NLS-1$
			while (line != null) {
				if ((line.length() > 0) && (line.charAt(0) == comment)) {
					/* Comment */
				} else {

					String[] items = MispSchemaUtil.getColumns(line, separator, parenthesis);
					if (items == null) {
						setLastError(rowCount, MispSchemaMessages
								.getString("MispSchemaCsv.16")); //$NON-NLS-1$
						break;
					}
					String header = getColumnVal(items, 0);

					if (header.length() == 0) {
						String subHeader = getColumnVal(items, 1);
						if (subHeader.length() > 0) {
							if (currHeader.equals(DFS_HDR_ET)) {
								MispSchemaComplexType ct = (MispSchemaComplexType) schema
										.getComplexType(schema
												.getComplexTypeSize() - 1);
								if (subHeader.equals(DFS_HDR_ET_EXT)) {
									ct.setExtension(getColumnVal(items, 2));
									if (ct.getExtension().length() <= 0) {
										setLastError(
												rowCount,
												MispSchemaMessages
														.getString("MispSchemaCsv.15")); //$NON-NLS-1$
										break;
									}
									log.debug("ElementType extension:" //$NON-NLS-1$
											+ ct.getExtension());
								} else if (subHeader.equals(DFS_HDR_ET_ATTR)) {
									MispSchemaAttribute attr = ct
											.addAttribute(-1);
									attr.setName(getColumnVal(items, 2));
									attr.setType(getColumnVal(items, 3));
									attr.setIndexedFlg(getColumnVal(items, 4)
											.equals("true")); //$NON-NLS-1$
									if (attr.getName().length() <= 0) {
										setLastError(
												rowCount,
												MispSchemaMessages
														.getString("MispSchemaCsv.23")); //$NON-NLS-1$
										break;
									}
									if (attr.getType().length() <= 0) {
										setLastError(
												rowCount,
												MispSchemaMessages
														.getString("MispSchemaCsv.24")); //$NON-NLS-1$
										break;
									}
									log.debug("ElementType attribute:" //$NON-NLS-1$
											+ attr.getName() + "/" //$NON-NLS-1$
											+ attr.getType());

								} else if (subHeader.equals(DFS_HDR_ET_EL)) {
									MispSchemaElement em = ct.addElement(-1);
									em.setName(getColumnVal(items, 2));
									em.setType(getColumnVal(items, 3));
									em.setIndexedFlg(getColumnVal(items, 4)
											.equals("true")); //$NON-NLS-1$
									String temp = getColumnVal(items, 5);
									if (temp.length() > 0)
										em.setMinOccurs(temp);
									temp = getColumnVal(items, 6);
									if (temp.length() > 0)
										em.setMaxOccurs(temp);
									if (em.getName().length() <= 0) {
										setLastError(
												rowCount,
												MispSchemaMessages
														.getString("MispSchemaCsv.14")); //$NON-NLS-1$
										break;
									}
									if (em.getType().length() <= 0) {
										setLastError(
												rowCount,
												MispSchemaMessages
														.getString("MispSchemaCsv.13")); //$NON-NLS-1$
										break;
									}
									log.debug("ElementType element:" //$NON-NLS-1$
											+ em.getName() + "/" //$NON-NLS-1$
											+ em.getType());

								} else {
									log.warn(MispSchemaMessages
											.getString("MispSchemaCsv.32") //$NON-NLS-1$
											+ rowCount + ":" + subHeader); //$NON-NLS-1$
								}
							} else if (currHeader.equals(DFS_HDR_IMPORT)
									&& subHeader
											.equals(DFS_HDR_IMPORT_LOCATION)) {
								MispSchemaImport im = (MispSchemaImport) schema
										.getImport(schema.getImportSize() - 1);
								im.setLocation(getColumnVal(items, 2));
								if (im.getLocation().length() <= 0) {
									setLastError(rowCount, MispSchemaMessages
											.getString("MispSchemaCsv.34")); //$NON-NLS-1$
									break;
								}
								log.debug("Import location:" //$NON-NLS-1$
										+ im.getLocation());
							} else {
								log
										.warn(MispSchemaMessages
												.getString("MispSchemaCsv.36") + rowCount //$NON-NLS-1$
												+ ":" + subHeader); //$NON-NLS-1$
							}
						}
					} else if (header.equals(DFS_HDR_ID)) {
						schema.setId(getColumnVal(items, 1));
						if (schema.getId().length() <= 0) {
							setLastError(rowCount, MispSchemaMessages
									.getString("MispSchemaCsv.38")); //$NON-NLS-1$
							break;
						}
						log.debug("ID:" + schema.getId()); //$NON-NLS-1$
					} else if (header.equals(DFS_HDR_ANNOTATION)) {
						schema
								.setAnnotation(descapeCrLf(getColumnVal(items,
										1)));
					} else if (header.equals(DFS_HDR_NS)) {
						MispSchemaNamespace ns = schema.addNamespace();
						ns.setPrefix(getColumnVal(items, 1));
						ns.setUri(getColumnVal(items, 2));
						if (ns.getPrefix().length() <= 0) {
							setLastError(rowCount, MispSchemaMessages
									.getString("MispSchemaCsv.40")); //$NON-NLS-1$
							break;
						}
						if (ns.getUri().length() <= 0) {
							setLastError(rowCount, MispSchemaMessages
									.getString("MispSchemaCsv.41")); //$NON-NLS-1$
							break;
						}
						log.debug("Namespace:" + ns.getPrefix() + "/" //$NON-NLS-1$ //$NON-NLS-2$
								+ ns.getUri());

					} else if (header.equals(DFS_HDR_TNS)) {
						schema.setTargetNamespace(getColumnVal(items, 1));
						if (schema.getTargetNamespace().length() <= 0) {
							setLastError(rowCount, MispSchemaMessages
									.getString("MispSchemaCsv.44")); //$NON-NLS-1$
							break;
						}
					} else if (header.equals(DFS_HDR_IMPORT)) {
						MispSchemaImport im = schema.addImport();
						im.setUri(getColumnVal(items, 1));
						if (im.getUri().length() <= 0) {
							setLastError(rowCount, MispSchemaMessages
									.getString("MispSchemaCsv.45")); //$NON-NLS-1$
							break;
						}
						log.debug("Import:" + im.getUri()); //$NON-NLS-1$

					} else if (header.equals(DFS_HDR_FT)) {
						schema.getFeatureElement().setName(
								getColumnVal(items, 1));
						schema.getFeatureElement().setType(
								getColumnVal(items, 2));
						if (schema.getFeatureElement().getName().length() <= 0) {
							setLastError(rowCount, MispSchemaMessages
									.getString("MispSchemaCsv.47")); //$NON-NLS-1$
							break;
						}
						if (schema.getFeatureElement().getType().length() <= 0) {
							setLastError(rowCount, MispSchemaMessages
									.getString("MispSchemaCsv.48")); //$NON-NLS-1$
							break;
						}
						log.debug("FeatureType:" //$NON-NLS-1$
								+ schema.getFeatureElement().getName() + "/" //$NON-NLS-1$
								+ schema.getFeatureElement().getType());
					} else if (header.equals(DFS_HDR_ET)) {
						MispSchemaComplexType ct = schema.addComplexType(null);
						ct.setName(getColumnVal(items, 1));
						if (ct.getName().length() <= 0) {
							setLastError(rowCount, MispSchemaMessages
									.getString("MispSchemaCsv.51")); //$NON-NLS-1$
							break;
						}
						log.debug("ElementType:" + ct.getName()); //$NON-NLS-1$

					} else if (header.charAt(0) == '#') {
						/* Extension purpose */
						header = ""; //$NON-NLS-1$
					} else {
						log.warn(MispSchemaMessages
								.getString("MispSchemaCsv.54") + rowCount + ":" //$NON-NLS-1$ //$NON-NLS-2$
								+ header);
					}

					if (header.length() > 0)
						currHeader = header;
				}
				line = brd.readLine();
				rowCount = rowCount + 1;
			}
			brd.close();
			if (!hasError) {
				convertVersion(schema);
				return schema;
			}
		} catch (Exception e) {
			setLastError(-1, e.getMessage());
		}
		return null;
	}

	/**
	 * 互換性を維持するためのデータタイプ変換テーブル
	 */
	private String[][] tblConvertDataType = {
			{ "string", "" + Const.NS_XSD_PRE + "string", Const.NS_XSD_PRE, //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
					Const.NS_XSD_URI },
			{ "integer", "" + Const.NS_XSD_PRE + "integer", Const.NS_XSD_PRE, //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
					Const.NS_XSD_URI },
			{ "float", "" + Const.NS_XSD_PRE + "float", Const.NS_XSD_PRE, //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$
					Const.NS_XSD_URI },
			{ "TimePositionType", Const.DFS_TYPE_TIMEPOSITIONTYPE, //$NON-NLS-1$
					Const.NS_GML_PRE, Const.NS_GML_URI },
			{ "geometryPropertyType", Const.DFS_TYPE_GEOMETRYPROPERTYTYPE, //$NON-NLS-1$
					Const.NS_GML_PRE, Const.NS_GML_URI }, };

	/**
	 * 互換性を維持するために 古いデータタイプを新しいデータタイプに変換する
	 * 
	 * @param schema
	 *            DaRuMaスキーマ
	 * @param obj
	 *            変換するオブジェクト
	 */
	private void convertDataType(MispSchema schema, Object obj) {
		String type;
		if (obj instanceof MispSchemaAttribute)
			type = ((MispSchemaAttribute) obj).getType();
		else if (obj instanceof MispSchemaElement)
			type = ((MispSchemaElement) obj).getType();
		else
			return;
		for (int cnt = 0; cnt < tblConvertDataType.length; cnt++) {
			if (type.equals(tblConvertDataType[cnt][0])) {
				String newType = tblConvertDataType[cnt][1];
				if (schema.findNamespaceByPrefix(tblConvertDataType[cnt][2]) == null) {
					MispSchemaNamespace ns = schema.addNamespace();
					ns.setPrefix(tblConvertDataType[cnt][2]);
					ns.setUri(tblConvertDataType[cnt][3]);
				}
				if (obj instanceof MispSchemaAttribute)
					((MispSchemaAttribute) obj).setType(newType);
				else if (obj instanceof MispSchemaElement)
					((MispSchemaElement) obj).setType(newType);
			}
		}
	}

	/**
	 * バージョン変換を行います
	 * 
	 * @param schema
	 *            変換するDaRuMaスキーマ
	 */
	private void convertVersion(MispSchema schema) {

		for (int cnti = 0; cnti < schema.getComplexTypeSize(); cnti++) {
			MispSchemaComplexType ct = schema.getComplexType(cnti);
			for (int cntj = 0; cntj < ct.getAttributeSize(); cntj++) {
				MispSchemaAttribute attr = ct.getAttribute(cntj);
				convertDataType(schema, attr);
			}
			for (int cntj = 0; cntj < ct.getElementSize(); cntj++) {
				MispSchemaElement em = ct.getElement(cntj);
				convertDataType(schema, em);
			}
		}
	}

	/**
	 * 指定されたインデックスのカラム値を取得する
	 * 
	 * @param columns
	 *            カラムリスト
	 * @param index
	 *            インデックス
	 * @return カラム値
	 */
	private String getColumnVal(String[] columns, int index) {
		if (columns.length > index)
			return columns[index].trim();
		return ""; //$NON-NLS-1$
	}

	/**
	 * DFSスキーマ形式で出力する
	 * 
	 * @param schema
	 *            出力するスキーマ
	 * @param fileName
	 *            ファイル名
	 * @param outputCharSet
	 *            出力文字セット
	 * @param separator
	 *            セパレータキャラクタ
	 * @param parenthesis
	 *            括弧キャラクタ
	 * @param comment
	 *            コメントキャラクタ
	 * @return 成功する場合はTrue, それ以外はFalseを返す
	 */
	public boolean saveSchemaToCsv(MispSchema schema, String fileName,
			String outputCharSet, char separator, char parenthesis, char comment) {
		try {
			clearError();
			if (!fileName.equals(schema.getFilePath()))
				schema.setFilePathUpdateReference(fileName);
			OutputStream fos = new FileOutputStream(fileName);
			return saveSchemaToCsv(schema, fos, outputCharSet, separator,
					parenthesis, comment);
		} catch (Exception e) {
			setLastError(-1, e.getMessage());
		}
		return false;
	}

	/**
	 * DFSスキーマ形式で出力する
	 * 
	 * @param schema
	 *            出力するスキーマ
	 * @param os
	 *            出力ストリーム
	 * @param outputCharSet
	 *            出力文字セット
	 * @param separator
	 *            セパレータキャラクタ
	 * @param parenthesis
	 *            括弧キャラクタ
	 * @param comment
	 *            コメントキャラクタ
	 * @return 成功する場合はTrue, それ以外はFalseを返す
	 */
	private boolean saveSchemaToCsv(MispSchema schema, OutputStream os,
			String outputCharSet, char separator, char parenthesis, char comment) {
		try {
			clearError();
			OutputStreamWriter osw;
			if ((outputCharSet != null) && (outputCharSet.length() > 0)) {
				osw = new OutputStreamWriter(os, outputCharSet);
			} else {
				osw = new OutputStreamWriter(os);
			}
			return saveSchemaToCsv(schema, osw, separator, parenthesis, comment);
		} catch (Exception e) {
			setLastError(-1, e.getMessage());
		}
		return false;
	}

	/**
	 * CSVライター
	 * 
	 * @author shumi
	 * 
	 */
	private class CsvWriter {
		/**
		 * 出力ストリームライター
		 */
		OutputStreamWriter osw;

		/**
		 * セパレータキャラクタ
		 */
		char separator;

		/**
		 * 括弧キャラクタ
		 */
		char parenthesis;

		/**
		 * コメントキャラクタ
		 */
		char comment;

		/**
		 * 行開始フラグ
		 */
		boolean flgFirst = true;

		/**
		 * CSV文字列を作成する
		 * 
		 * @param val
		 * @param parenthesis
		 * @return
		 */
		private String makeCsvString(String val, char parenthesis) {
			val = val.replaceAll("" + parenthesis, "" + parenthesis //$NON-NLS-1$ //$NON-NLS-2$
					+ parenthesis);
			return parenthesis + val + parenthesis;
		}

		/**
		 * CSVライターを生成する
		 * 
		 * @param osw
		 *            出力ストリームライター
		 * @param outputCharSet
		 *            出力文字セット
		 * @param separator
		 *            セパレータキャラクタ
		 * @param parenthesis
		 *            括弧キャラクタ
		 */
		CsvWriter(OutputStreamWriter osw, char separator, char parenthesis,
				char comment) {
			this.osw = osw;
			this.separator = separator;
			this.parenthesis = parenthesis;
			this.comment = comment;
		}

		/**
		 * カラムを出力する
		 * 
		 * @param val
		 * @throws IOException
		 */
		public void outputColumn(String val) throws IOException {
			osw.write((flgFirst ? "" : ",") + makeCsvString(val, parenthesis)); //$NON-NLS-1$ //$NON-NLS-2$
			flgFirst = false;
		}

		/**
		 * 新しい行をはじめる
		 * 
		 * @throws IOException
		 */
		public void nextLine() throws IOException {
			osw.write(MispSchema.CRLF);
			flgFirst = true;
		}
	}

	/**
	 * DFSスキーマ形式で出力する
	 * 
	 * @param schema
	 *            出力するスキーマ
	 * @param osw
	 *            出力ストリームライター
	 * @param separator
	 *            セパレータキャラクタ
	 * @param parenthesis
	 *            括弧キャラクタ
	 * @param comment
	 *            コメントキャラクタ
	 * @return 成功する場合はTrue, それ以外はFalseを返す
	 */
	private boolean saveSchemaToCsv(MispSchema schema, OutputStreamWriter osw,
			char separator, char parenthesis, char comment) {
		try {
			clearError();

			schema.adjustImportPath();
			CsvWriter writer = new CsvWriter(osw, separator, parenthesis,
					comment);

			/*
			 * バージョンを出力する
			 */
			if (dfsOutputVersion) {
				writer.outputColumn("#DaRuMa"); //$NON-NLS-1$
				writer.outputColumn("format=SchemaCSV"); //$NON-NLS-1$
				writer.outputColumn("version=0.0"); //$NON-NLS-1$
				writer.nextLine();
			}

			/*
			 * ID
			 */
			writer.outputColumn(DFS_HDR_ID);
			writer.outputColumn(schema.getId());
			writer.nextLine();

			/*
			 * Annotation
			 */
			writer.outputColumn(DFS_HDR_ANNOTATION);
			writer.outputColumn(escapeCrLf(schema.getAnnotation()));
			writer.nextLine();

			/*
			 * Namespace
			 */
			boolean hasMisp = false;
			for (int cnt = 0; cnt < schema.getNamespaceSize(); cnt++) {
				MispSchemaNamespace ns = (MispSchemaNamespace) schema
						.getNamespace(cnt);
				writer.outputColumn(DFS_HDR_NS);
				writer.outputColumn(ns.getPrefix());
				writer.outputColumn(ns.getUri());
				writer.nextLine();

				if (ns.getPrefix().equals(Const.NS_MISP_PRE)
						&& ns.getUri().equals(Const.NS_MISP_URI))
					hasMisp = true;
			}
			if (!hasMisp) {
				writer.outputColumn(DFS_HDR_NS);
				writer.outputColumn(Const.NS_MISP_PRE);
				writer.outputColumn(Const.NS_MISP_URI);
				writer.nextLine();
			}

			/*
			 * TargetNamespace
			 */
			writer.outputColumn(DFS_HDR_TNS);
			writer.outputColumn(schema.getTargetNamespace());
			writer.nextLine();

			/*
			 * Import
			 */
			for (int cnt = 0; cnt < schema.getImportSize(); cnt++) {
				MispSchemaImport im = (MispSchemaImport) schema.getImport(cnt);
				writer.outputColumn(DFS_HDR_IMPORT);
				writer.outputColumn(im.getUri());
				if (im.getLocation().length() > 0) {
					writer.nextLine();
					writer.outputColumn(""); //$NON-NLS-1$
					writer.outputColumn(DFS_HDR_IMPORT_LOCATION);
					writer.outputColumn(im.getLocation());
				}
				writer.nextLine();
			}

			/*
			 * FeatureType
			 */
			MispSchemaElement ft = schema.getFeatureElement();
			writer.outputColumn(DFS_HDR_FT);
			writer.outputColumn(ft.getName());
			writer.outputColumn(ft.getType());
			writer.nextLine();

			/*
			 * ComplexType
			 */
			for (int cnt = 0; cnt < schema.getComplexTypeSize(); cnt++) {
				MispSchemaComplexType ct = (MispSchemaComplexType) schema
						.getComplexType(cnt);
				if (ct.getKind() != MispSchema.KIND_INTERNAL)
					continue;
				writer.outputColumn(DFS_HDR_ET);
				writer.outputColumn(ct.getName());
				writer.nextLine();

				if (ct.getExtension().length() > 0) {
					writer.outputColumn(""); //$NON-NLS-1$
					writer.outputColumn(DFS_HDR_ET_EXT);
					writer.outputColumn(ct.getExtension());
					writer.nextLine();
				}

				/*
				 * Attribute
				 */
				for (int cntj = 0; cntj < ct.getAttributeSize(); cntj++) {
					MispSchemaAttribute attr = (MispSchemaAttribute) ct
							.getAttribute(cntj);
					writer.outputColumn(""); //$NON-NLS-1$
					writer.outputColumn(DFS_HDR_ET_ATTR);
					writer.outputColumn(attr.getName());
					writer.outputColumn(attr.getType());
					writer.outputColumn(attr.isIndexedFlg() ? "true" : "false"); //$NON-NLS-1$ //$NON-NLS-2$
					writer.nextLine();
				}

				/*
				 * Element
				 */
				for (int cntj = 0; cntj < ct.getElementSize(); cntj++) {
					MispSchemaElement em = (MispSchemaElement) ct
							.getElement(cntj);
					writer.outputColumn(""); //$NON-NLS-1$
					writer.outputColumn(DFS_HDR_ET_EL);
					writer.outputColumn(em.getName());
					writer.outputColumn(em.getType());
					writer.outputColumn(em.isIndexedFlg() ? "true" : "false"); //$NON-NLS-1$ //$NON-NLS-2$
					writer.outputColumn(em.getMinOccurs());
					writer.outputColumn(em.getMaxOccurs());
					writer.nextLine();
				}
			}

			osw.flush();
			osw.close();
			return true;
		} catch (Exception e) {
			setLastError(-1, e.getMessage());
		}
		return false;
	}

	/**
	 * 改行をエスケープする
	 * 
	 * @param val
	 *            エスケープする文字列
	 * @return エスケープ結果を返す
	 */
	private String escapeCrLf(String val) {
		val = val.replaceAll("\\\\", "\\\\\\\\"); //$NON-NLS-1$ //$NON-NLS-2$
		val = val.replaceAll("\\r", "\\\\r"); //$NON-NLS-1$ //$NON-NLS-2$
		val = val.replaceAll("\\n", "\\\\n"); //$NON-NLS-1$ //$NON-NLS-2$
		return val;
	}

	/**
	 * 改行をデスケープする
	 * 
	 * @param val
	 *            デスケープする文字列
	 * @return デスケープ結果を返す
	 */
	private String descapeCrLf(String val) {
		String temp = ""; //$NON-NLS-1$
		boolean flgEscape = false;
		for (int cnt = 0; cnt < val.length(); cnt++) {
			char chr = val.charAt(cnt);
			if (flgEscape) {
				if (chr == 'r')
					temp += "\r"; //$NON-NLS-1$
				else if (chr == 'n')
					temp += '\n';
				else
					temp += chr;

				flgEscape = false;
			} else {
				flgEscape = (chr == '\\');
				if (!flgEscape)
					temp += chr;
			}
		}
		return temp;
	}

	/**
	 * outputVersionを取得する
	 * 
	 * @return Returns the outputVersion.
	 */
	public boolean isDfsOutputVersion() {
		return dfsOutputVersion;
	}

	/**
	 * outputVersionをセットする
	 * 
	 * @param outputVersion
	 *            The outputVersion to set.
	 */
	public void setDfsOutputVersion(boolean outputVersion) {
		this.dfsOutputVersion = outputVersion;
	}

	/**
	 * XMLスキーマの拡張子
	 */
	final public static String XSD_DFT_FILE_EXT = "xsd"; //$NON-NLS-1$

	static public final String XSD_TAG_SCHEMA = "schema"; //$NON-NLS-1$

	static public final String XSD_FULL_TAG_SCHEMA = Const.NS_XSD_PRE + ":" //$NON-NLS-1$
			+ XSD_TAG_SCHEMA;

	static public final String XSD_ATTR_SCHEMA_ID = "id"; //$NON-NLS-1$

	static public final String XSD_FULL_ATTR_SCHEMA_ID = Const.NS_MISP_PRE
			+ ":" //$NON-NLS-1$
			+ XSD_ATTR_SCHEMA_ID;

	static public final String XSD_ATTR_SCHEMA_TNS = "targetNamespace"; //$NON-NLS-1$

	static public final String XSD_TAG_IMPORT = "import"; //$NON-NLS-1$

	static public final String XSD_FULL_TAG_IMPORT_FULL = Const.NS_XSD_PRE
			+ ":" //$NON-NLS-1$
			+ XSD_TAG_IMPORT;

	static public final String XSD_TAG_INCLUDE = "include"; //$NON-NLS-1$

	static public final String XSD_ATTR_INCLUDE_LOCATION = "schemaLocation"; //$NON-NLS-1$

	static public final String XSD_ATTR_IMPORT_NAMESPACE = "namespace"; //$NON-NLS-1$

	static public final String XSD_ATTR_IMPORT_LOCATION = "schemaLocation"; //$NON-NLS-1$

	static public final String XSD_TAG_ELEMENT = "element"; //$NON-NLS-1$

	static public final String XSD_FULL_TAG_ELEMENT = Const.NS_XSD_PRE
			+ ":" + XSD_TAG_ELEMENT; //$NON-NLS-1$

	static public final String XSD_ATTR_ELEMENT_NAME = "name"; //$NON-NLS-1$

	static public final String XSD_ATTR_ELEMENT_TYPE = "type"; //$NON-NLS-1$

	static public final String XSD_ATTR_ELEMENT_REF = "ref"; //$NON-NLS-1$

	static public final String XSD_ATTR_ELEMENT_MIN_OCCURS = "minOccurs"; //$NON-NLS-1$

	static public final String XSD_ATTR_ELEMENT_MAX_OCCURS = "maxOccurs"; //$NON-NLS-1$

	static public final String XSD_ATTR_ELEMENT_INDEXED = "indexed"; //$NON-NLS-1$


	static public final String XSD_FULL_ATTR_ELEMENT_INDEXED = Const.NS_MISP_PRE
			+ ":" //$NON-NLS-1$
			+ XSD_ATTR_ELEMENT_INDEXED;
	
	static public final String XSD_ATTR_ELEMENT_DFSTYPE = "dfsType"; //$NON-NLS-1$

	static public final String XSD_FULL_ATTR_ELEMENT_DFSTYPE = Const.NS_MISP_PRE
			+ ":" //$NON-NLS-1$
			+ XSD_ATTR_ELEMENT_DFSTYPE;
	
	static public final String XSD_TAG_ATTR = "attribute"; //$NON-NLS-1$

	static public final String XSD_FULL_TAG_ATTR = Const.NS_XSD_PRE
			+ ":" + XSD_TAG_ATTR; //$NON-NLS-1$

	static public final String XSD_ATTR_ATTR_NAME = "name"; //$NON-NLS-1$

	static public final String XSD_ATTR_ATTR_TYPE = "type"; //$NON-NLS-1$

	static public final String XSD_ATTR_ATTR_INDEXED = Const.NS_MISP_PRE
			+ ":indexed"; //$NON-NLS-1$

	static public final String XSD_ATTR_ATTR_REF = "ref"; //$NON-NLS-1$

	static public final String XSD_TAG_COMPLEXTYPE = "complexType"; //$NON-NLS-1$

	static public final String XSD_FULL_TAG_COMPLEXTYPE = Const.NS_XSD_PRE
			+ ":" + XSD_TAG_COMPLEXTYPE; //$NON-NLS-1$

	static public final String XSD_ATTR_COMPLEXTYPE_NAME = "name"; //$NON-NLS-1$

	static public final String XSD_TAG_COMPLEXCONTENT = "complexContent"; //$NON-NLS-1$

	static public final String XSD_FULL_TAG_COMPLEX_CONTENT = Const.NS_XSD_PRE
			+ ":" //$NON-NLS-1$
			+ XSD_TAG_COMPLEXCONTENT;

	static public final String XSD_TAG_EXTENSION = "extension"; //$NON-NLS-1$

	static public final String XSD_FULL_TAG_EXTENSION = Const.NS_XSD_PRE + ":" //$NON-NLS-1$
			+ XSD_TAG_EXTENSION;

	static public final String XSD_ATTR_EXTENSION_BASE = "base"; //$NON-NLS-1$

	static public final String XSD_TAG_SEQUENCE = "sequence"; //$NON-NLS-1$

	static public final String XSD_FULL_TAG_SEQUENCE = Const.NS_XSD_PRE + ":" //$NON-NLS-1$
			+ XSD_TAG_SEQUENCE;

	static public final String XSD_ATTR_SUBSTITUTION_GROUP = "substitutionGroup"; //$NON-NLS-1$

	static public final String XSD_ATTRVAL_SUBSTITUTION_GROUP = Const.NS_GML_PRE
			+ ":_Feature"; //$NON-NLS-1$

	static public final String XSD_TAG_ANNOTATION = "annotation"; //$NON-NLS-1$

	static public final String XSD_FULL_TAG_ANNOTATION = Const.NS_XSD_PRE + ":" //$NON-NLS-1$
			+ XSD_TAG_ANNOTATION;

	static public final String XSD_TAG_DOCUMENTATION = "documentation"; //$NON-NLS-1$

	static public final String XSD_FULL_TAG_DOCUMENTATION = Const.NS_XSD_PRE
			+ ":" //$NON-NLS-1$
			+ XSD_TAG_DOCUMENTATION;

	/**
	 * XMLスキーマ形式で保存する
	 * 
	 * @param schema
	 *            保存するスキーマ
	 * @param fileName
	 *            ファイル名
	 * @param outputCharSet
	 *            出力文字セット
	 * @param xmlEncoding
	 *            XML文字セット
	 * @return 成功する場合はTrue, それ以外はFalseを返す
	 */
	public boolean saveSchemaToXml(MispSchema schema, String fileName,
			String outputCharSet, String xmlEncoding) {
		try {
			if (!fileName.equals(schema.getFilePath()))
				schema.setFilePathUpdateReference(fileName);
			FileOutputStream fos = new FileOutputStream(fileName);
			return saveSchemaToXml(schema, fos, outputCharSet, xmlEncoding);
		} catch (Exception e) {
			setLastError(e.getMessage());
		}
		return false;
	}

	/**
	 * XMLスキーマ形式で保存する
	 * 
	 * @param schema
	 *            保存するスキーマ
	 * @param os
	 *            出力ストリーム
	 * @param outputCharSet
	 *            出力文字セット
	 * @param xmlEncoding
	 *            XML文字セット
	 * @return 成功する場合はTrue, それ以外はFalseを返す
	 */
	public boolean saveSchemaToXml(MispSchema schema, OutputStream os,
			String outputCharSet, String xmlEncoding) {
		try {
			OutputStreamWriter osw;
			if ((outputCharSet != null) && (outputCharSet.length() > 0)) {
				osw = new OutputStreamWriter(os, outputCharSet);
			} else {
				osw = new OutputStreamWriter(os);
			}
			return saveSchemaToXml(schema, osw, xmlEncoding);
		} catch (Exception e) {
			setLastError(e.getMessage());
		}
		return false;
	}

	/**
	 * XMLスキーマ形式で保存する
	 * 
	 * @param schema
	 *            保存するスキーマ
	 * @param osw
	 *            出力ストリームライター
	 * @param xmlEncoding
	 *            XML文字セット
	 * @return 成功する場合はTrue, それ以外はFalseを返す
	 */
	private boolean saveSchemaToXml(MispSchema schema, OutputStreamWriter osw,
			String xmlEncoding) {
		try {
			schema.adjustImportPath();

			String prefix = ""; //$NON-NLS-1$
			// Output xml header
			osw.write("<?xml version=\"1.0\" encoding=\"" + xmlEncoding //$NON-NLS-1$
					+ "\"?>" + MispSchema.CRLF); //$NON-NLS-1$

			outputXml(schema, osw, prefix);

			osw.flush();
			osw.close();
			return true;
		} catch (Exception e) {
			if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
				e.printStackTrace();
			}
			setLastError(e.getMessage());
		}
		return false;
	}

	/**
	 * XMLスキーマを出力する
	 * 
	 * @param schema
	 *            保存するスキーマ
	 * @param osw
	 *            出力ストリームライター
	 * @param prefix
	 *            XMLスキーマ接頭辞
	 */
	public static void outputXml(MispSchema schema, OutputStreamWriter osw,
			String prefix) throws IOException {
		// Open xsd:schema
		osw.write(prefix + "<" + XSD_FULL_TAG_SCHEMA + MispSchema.CRLF); //$NON-NLS-1$
		osw.write(prefix + "  " + Const.NS_XMLNS_PRE + "=\"" //$NON-NLS-1$ //$NON-NLS-2$
				+ escapeXmlString(schema.getTargetNamespace()) + "\"" //$NON-NLS-1$
				+ MispSchema.CRLF);

		boolean hasMisp = false;
		boolean hasXsd = false;

		/*
		 * 名前空間
		 */
		for (int cnt = 0; cnt < schema.getNamespaceSize(); cnt++) {
			MispSchemaNamespace ns = (MispSchemaNamespace) schema
					.getNamespace(cnt);
			osw.write(prefix + "  " + Const.NS_XMLNS_PRE + ":" //$NON-NLS-1$ //$NON-NLS-2$
					+ escapeXmlString(ns.getPrefix()) + "=\"" //$NON-NLS-1$
					+ escapeXmlString(ns.getUri()) + "\"" + MispSchema.CRLF); //$NON-NLS-1$
			if (ns.getPrefix().equals(Const.NS_MISP_PRE)
					&& ns.getUri().equals(Const.NS_MISP_URI))
				hasMisp = true;
			if (ns.getPrefix().equals(Const.NS_XSD_PRE)
					&& ns.getUri().equals(Const.NS_XSD_URI))
				hasXsd = true;
		}

		/*
		 * 必須名前空間(misp, xsd)
		 * 
		 */
		if (!hasMisp) {
			osw.write(prefix + "  " + Const.NS_XMLNS_PRE + ":" //$NON-NLS-1$ //$NON-NLS-2$
					+ Const.NS_MISP_PRE + "=\"" + Const.NS_MISP_URI + "\"" //$NON-NLS-1$ //$NON-NLS-2$
					+ MispSchema.CRLF);
		}
		if (!hasXsd) {
			osw.write(prefix + "  " + Const.NS_XMLNS_PRE + ":" //$NON-NLS-1$ //$NON-NLS-2$
					+ Const.NS_XSD_PRE + "=\"" + Const.NS_XSD_URI + "\"" //$NON-NLS-1$ //$NON-NLS-2$
					+ MispSchema.CRLF);
		}
		// osw.write(prefix + " " + Const.NS_XMLNS_PRE + ":" + EL_XLINK_PRE +
		// "=\""
		// + EL_XLINK_URI + "\"" + MispSchema.CRLF);

		/*
		 * スキーマID, ターゲット名前空間
		 */
		osw.write(prefix + "  " + XSD_FULL_ATTR_SCHEMA_ID + "=\"" //$NON-NLS-1$ //$NON-NLS-2$
				+ escapeXmlString(schema.getId()) + "\"" + MispSchema.CRLF); //$NON-NLS-1$
		osw.write(prefix + "  " + XSD_ATTR_SCHEMA_TNS + "=\"" //$NON-NLS-1$ //$NON-NLS-2$
				+ escapeXmlString(schema.getTargetNamespace()) + "\">" //$NON-NLS-1$
				+ MispSchema.CRLF);

		osw.write(MispSchema.CRLF);

		/*
		 * Output annotation
		 */
		osw.write(prefix + " <" + XSD_FULL_TAG_ANNOTATION + ">"); //$NON-NLS-1$ //$NON-NLS-2$
		osw.write(prefix + " <" + XSD_FULL_TAG_DOCUMENTATION + ">"); //$NON-NLS-1$ //$NON-NLS-2$
		osw.write(escapeXmlString(schema.getAnnotation()));
		osw.write(prefix + " </" + XSD_FULL_TAG_DOCUMENTATION + ">"); //$NON-NLS-1$ //$NON-NLS-2$
		osw.write(prefix + " </" + XSD_FULL_TAG_ANNOTATION + ">"); //$NON-NLS-1$ //$NON-NLS-2$

		/*
		 * Output import
		 */
		for (int cnt = 0; cnt < schema.getImportSize(); cnt++) {
			MispSchemaImport im = (MispSchemaImport) schema.getImport(cnt);
			osw.write(prefix
					+ "  <" + XSD_FULL_TAG_IMPORT_FULL + MispSchema.CRLF); //$NON-NLS-1$
			osw.write(prefix + "  " + XSD_ATTR_IMPORT_NAMESPACE + "=\"" //$NON-NLS-1$ //$NON-NLS-2$
					+ escapeXmlString(im.getUri()) + "\"" + MispSchema.CRLF); //$NON-NLS-1$
			osw.write(prefix + "  " + XSD_ATTR_IMPORT_LOCATION + "=\"" //$NON-NLS-1$ //$NON-NLS-2$
					+ escapeXmlString(im.getLocation()) + "\"/>" //$NON-NLS-1$
					+ MispSchema.CRLF);
		}
		osw.write(MispSchema.CRLF);

		/*
		 * Output featureType
		 */
		MispSchemaElement featureElement = schema.getFeatureElement();
		osw.write(prefix + "  <" + XSD_FULL_TAG_ELEMENT); //$NON-NLS-1$
		osw.write(" " + XSD_ATTR_ELEMENT_NAME + "=\"" //$NON-NLS-1$ //$NON-NLS-2$
				+ escapeXmlString(featureElement.getName()) + "\""); //$NON-NLS-1$
		osw.write(" " + XSD_ATTR_ELEMENT_TYPE + "=\"" //$NON-NLS-1$ //$NON-NLS-2$
				+ escapeXmlString(featureElement.getType()) + "\""); //$NON-NLS-1$
		osw
				.write(" " + XSD_ATTR_SUBSTITUTION_GROUP + "=\"" + XSD_ATTRVAL_SUBSTITUTION_GROUP + "\""); //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$

		osw.write("/>" + MispSchema.CRLF); //$NON-NLS-1$
		osw.write(MispSchema.CRLF);

		/*
		 * Output elementType
		 */
		for (int cnt = 0; cnt < schema.getComplexTypeSize(); cnt++) {
			MispSchemaComplexType ct = (MispSchemaComplexType) schema
					.getComplexType(cnt);
			if (ct.getKind() != MispSchema.KIND_INTERNAL)
				continue;

			boolean hasExt = ct.getExtension().length() > 0;

			// <complexType>
			osw.write(prefix + "  <" + XSD_FULL_TAG_COMPLEXTYPE); //$NON-NLS-1$
			osw
					.write("  " + XSD_ATTR_COMPLEXTYPE_NAME + "=\"" + escapeXmlString(ct.getName()) //$NON-NLS-1$ //$NON-NLS-2$
							+ "\""); //$NON-NLS-1$
			osw.write(">" + MispSchema.CRLF); //$NON-NLS-1$

			if (hasExt) {
				// <complexContent>
				osw.write(prefix + "    <" + XSD_FULL_TAG_COMPLEX_CONTENT + ">" //$NON-NLS-1$ //$NON-NLS-2$
						+ MispSchema.CRLF);
				// <extension>
				osw.write(prefix + "      <" + XSD_FULL_TAG_EXTENSION + " " //$NON-NLS-1$ //$NON-NLS-2$
						+ XSD_ATTR_EXTENSION_BASE + "=\"" //$NON-NLS-1$
						+ escapeXmlString(ct.getExtension()) + "\"" + ">" //$NON-NLS-1$ //$NON-NLS-2$
						+ MispSchema.CRLF);
			}

			// <sequence>
			osw.write(prefix + "        <" + XSD_FULL_TAG_SEQUENCE + ">" //$NON-NLS-1$ //$NON-NLS-2$
					+ MispSchema.CRLF);

			// Elements
			for (int cntj = 0; cntj < ct.getElementSize(); cntj++) {
				MispSchemaElement em = (MispSchemaElement) ct.getElement(cntj);
				osw.write(prefix + "          <" + XSD_FULL_TAG_ELEMENT); //$NON-NLS-1$
				osw.write(" " + XSD_ATTR_ELEMENT_NAME + "=\"" //$NON-NLS-1$ //$NON-NLS-2$
						+ escapeXmlString(em.getName()) + "\""); //$NON-NLS-1$
				osw.write(" " + XSD_ATTR_ELEMENT_TYPE + "=\"" //$NON-NLS-1$ //$NON-NLS-2$
						+ escapeXmlString(em.getXsdType()) + "\""); //$NON-NLS-1$
				osw.write(" " + XSD_ATTR_ELEMENT_MIN_OCCURS + "=\"" //$NON-NLS-1$ //$NON-NLS-2$
						+ escapeXmlString(em.getMinOccurs()) + "\""); //$NON-NLS-1$
				osw.write(" " + XSD_ATTR_ELEMENT_MAX_OCCURS + "=\"" //$NON-NLS-1$ //$NON-NLS-2$
						+ escapeXmlString(em.getMaxOccurs()) + "\""); //$NON-NLS-1$
				osw.write(" " + XSD_FULL_ATTR_ELEMENT_INDEXED + "=\"" //$NON-NLS-1$ //$NON-NLS-2$
						+ escapeXmlString(em.isIndexedFlg() ? "true" : "false") //$NON-NLS-1$ //$NON-NLS-2$
						+ "\""); //$NON-NLS-1$
				if(!em.getType().equals(em.getXsdType()))
				{
					osw.write(" " + XSD_FULL_ATTR_ELEMENT_DFSTYPE + "=\"" //$NON-NLS-1$ //$NON-NLS-2$
							+ escapeXmlString(em.getType()) //$NON-NLS-1$ //$NON-NLS-2$
							+ "\""); //$NON-NLS-1$
				}
				osw.write("/>" + MispSchema.CRLF); //$NON-NLS-1$
			}

			// <sequence>
			osw.write(prefix + "        </" + XSD_FULL_TAG_SEQUENCE + ">" //$NON-NLS-1$ //$NON-NLS-2$
					+ MispSchema.CRLF);

			// Attributes
			for (int cntj = 0; cntj < ct.getAttributeSize(); cntj++) {
				MispSchemaAttribute attr = (MispSchemaAttribute) ct
						.getAttribute(cntj);
				osw.write(prefix + "        <" + XSD_FULL_TAG_ATTR); //$NON-NLS-1$
				osw.write(" " + XSD_ATTR_ATTR_NAME + "=\"" //$NON-NLS-1$ //$NON-NLS-2$
						+ escapeXmlString(attr.getName()) + "\""); //$NON-NLS-1$
				osw.write(" " + XSD_ATTR_ATTR_TYPE + "=\"" //$NON-NLS-1$ //$NON-NLS-2$
						+ escapeXmlString(attr.getType()) + "\""); //$NON-NLS-1$
				osw.write(" " //$NON-NLS-1$
						+ XSD_ATTR_ATTR_INDEXED + "=\"" //$NON-NLS-1$
						+ escapeXmlString(attr.isIndexedFlg() ? "true" //$NON-NLS-1$
								: "false") + "\""); //$NON-NLS-1$ //$NON-NLS-2$
				osw.write("/>" + MispSchema.CRLF); //$NON-NLS-1$

			}

			// <extension>
			if (hasExt) {
				osw.write(prefix + "      </" + XSD_FULL_TAG_EXTENSION + ">" //$NON-NLS-1$ //$NON-NLS-2$
						+ MispSchema.CRLF);

				// <complexContent>
				osw.write(prefix
						+ "    </" + XSD_FULL_TAG_COMPLEX_CONTENT + ">" //$NON-NLS-1$ //$NON-NLS-2$
						+ MispSchema.CRLF);
			}

			// <complexType>
			osw
					.write(prefix
							+ "  </" + XSD_FULL_TAG_COMPLEXTYPE + ">" + MispSchema.CRLF); //$NON-NLS-1$ //$NON-NLS-2$
			osw.write(MispSchema.CRLF);
		}

		// Close xsd:schema
		osw.write(prefix + "</" + XSD_FULL_TAG_SCHEMA + ">" + MispSchema.CRLF); //$NON-NLS-1$ //$NON-NLS-2$
	}

	/**
	 * XML文字列をエスケープする
	 * 
	 * @param src
	 *            ソース文字列
	 * @return エスケープされた文字列
	 */
	public static String escapeXmlString(String src) {
		src = src.replaceAll("&", "&amp;"); //$NON-NLS-1$ //$NON-NLS-2$
		src = src.replaceAll("<", "&lt;"); //$NON-NLS-1$ //$NON-NLS-2$
		src = src.replaceAll(">", "&gt;"); //$NON-NLS-1$ //$NON-NLS-2$
		src = src.replaceAll("\"", "&quot;"); //$NON-NLS-1$ //$NON-NLS-2$
		src = src.replaceAll("'", "&apos;"); //$NON-NLS-1$ //$NON-NLS-2$
		return src;
	}

	/**
	 * XMLスキーマを読み込む
	 * 
	 * @param fileName
	 *            ファイル名
	 * @param dfsCharset
	 *            デフォルトキャラクタセット
	 * @return 成功する場合は該当スキーマ、それ以外はnullを返す
	 */
	public MispSchema loadSchemaFromXml(String fileName, String dfsCharset) {
		try {
			MispSchema schema = new MispSchema();
			schema.setFilePath(fileName);
			schema.setDfsCharSet(dfsCharset);

			InputStream fis;
			if (fileName.startsWith("http://")) { //$NON-NLS-1$
				URL url = new URL(fileName);
				fis = url.openStream();
			} else {
				File file = new File(fileName);
				fis = new FileInputStream(file);
			}

			DOMParser parser = new DOMParser();
			parser.setEntityResolver(new EntityResolver() {
				public InputSource resolveEntity(java.lang.String publicId,
						java.lang.String systemId) throws SAXException,
						java.io.IOException {
					if (Const.DTD_URI.equals(systemId)) {
						return new InputSource(new ByteArrayInputStream(
								"<?xml version='1.0' encoding='UTF-8'?>" //$NON-NLS-1$
								.getBytes()));
					}
					return null;
				}
			});

			parser.parse(new InputSource(fis));
			Document doc = parser.getDocument();
			Element root = doc.getDocumentElement();
			if (root.getNamespaceURI().equals(Const.NS_XSD_URI)
					&& root.getLocalName().equals(XSD_TAG_SCHEMA)) {
				parseSchema(schema, root);
			}

			return schema;
		} catch (Exception e) {
			log.error("Fail to open xsd file:" + fileName); //$NON-NLS-1$
		}
		return null;
	}

	/**
	 * XMLスキーマをパースする
	 * 
	 * @param schema
	 *            パース結果を保存するDaRuMaスキーマ
	 * @param root
	 *            スキーマのルートエレメント
	 */
	public void parseSchema(MispSchema schema, Element root) {
		String temp;
		int index = 0;

		/**
		 * Schema ID
		 */
		temp = root.getAttributeNS(Const.NS_MISP_URI, XSD_ATTR_SCHEMA_ID);
		schema.setId((temp == null) ? "" : temp); //$NON-NLS-1$

		/**
		 * TargetNamespace
		 */
		temp = root.getAttribute(XSD_ATTR_SCHEMA_TNS);
		schema.setTargetNamespace((temp == null) ? "" : temp); //$NON-NLS-1$

		/**
		 * Annotation
		 */
		NodeList nodeList = root.getElementsByTagNameNS(Const.NS_MISP_URI,
				XSD_TAG_ANNOTATION);
		index = 0;
		for (int cnt = 0; cnt < nodeList.getLength(); cnt++) {
			Element em = (Element) nodeList.item(cnt);
			if (em.getParentNode() == root) {
				nodeList = em.getElementsByTagNameNS(Const.NS_MISP_URI,
						XSD_TAG_DOCUMENTATION);
				if (nodeList.getLength() > 0) {
					em = (Element) nodeList.item(0);
					schema.setAnnotation(em.getNodeValue());
				}
				index++;
				break;
			}
		}

		/**
		 * NamespaceList
		 */
		NamedNodeMap map = root.getAttributes();
		String xmlNsPre = Const.NS_XMLNS_PRE + ":"; //$NON-NLS-1$
		for (int cnt = 0; cnt < map.getLength(); cnt++) {
			Node node = map.item(cnt);
			String attrName = node.getNodeName();
			if (attrName.startsWith(xmlNsPre)) {
				String prefix = attrName.substring(xmlNsPre.length());
				String uri = node.getNodeValue();
				MispSchemaNamespace ns = schema.addNamespace();
				ns.setPrefix(prefix);
				ns.setUri(uri);
			}
		}

		/**
		 * ImportList
		 */
		nodeList = root
				.getElementsByTagNameNS(Const.NS_XSD_URI, XSD_TAG_IMPORT);
		index = 0;
		for (int cnt = 0; cnt < nodeList.getLength(); cnt++) {
			Element em = (Element) nodeList.item(cnt);
			if (em.getParentNode() == root) {
				MispSchemaImport im = schema.addImport();
				temp = em.getAttribute(XSD_ATTR_IMPORT_NAMESPACE);
				im.setUri((temp == null) ? "" : temp); //$NON-NLS-1$
				temp = em.getAttribute(XSD_ATTR_IMPORT_LOCATION);
				im.setLocation((temp == null) ? "" : temp); //$NON-NLS-1$
				index++;
			}

		}

		/**
		 * IncludeList
		 */
		nodeList = root.getElementsByTagNameNS(Const.NS_XSD_URI,
				XSD_TAG_INCLUDE);
		index = 0;
		for (int cnt = 0; cnt < nodeList.getLength(); cnt++) {
			Element em = (Element) nodeList.item(cnt);
			if (em.getParentNode() == root) {
				MispSchemaImport im = schema.addImport();
				im.setIncludeFlg(true);
				im.setUri(""); //$NON-NLS-1$
				temp = em.getAttribute(XSD_ATTR_INCLUDE_LOCATION);
				im.setLocation((temp == null) ? "" : temp); //$NON-NLS-1$
				index++;
			}
		}

		/**
		 * RootElement(FeatureType)
		 */
		nodeList = root.getElementsByTagNameNS(Const.NS_XSD_URI,
				XSD_TAG_ELEMENT);
		index = 0;
		for (int cnt = 0; cnt < nodeList.getLength(); cnt++) {
			Element em = (Element) nodeList.item(cnt);
			if (em.getParentNode() == root) {
				MispSchemaElement et;
				if (index == 0) {
					et = schema.getFeatureElement();
				} else {
					et = schema.addRootElement();
				}
				temp = em.getAttribute(XSD_ATTR_ELEMENT_NAME);
				et.setName((temp == null) ? "" : temp); //$NON-NLS-1$
				temp = em.getAttribute(XSD_ATTR_ELEMENT_TYPE);
				et.setType((temp == null) ? "" : temp); //$NON-NLS-1$
				index++;
			}
		}

		/**
		 * RootAttribute
		 */
		nodeList = root.getElementsByTagNameNS(Const.NS_XSD_URI, XSD_TAG_ATTR);
		index = 0;
		for (int cnt = 0; cnt < nodeList.getLength(); cnt++) {
			Element em = (Element) nodeList.item(cnt);
			if (em.getParentNode() == root) {
				MispSchemaAttribute attr;
				attr = schema.addRootAttribute();
				temp = em.getAttribute(XSD_ATTR_ATTR_NAME);
				attr.setName((temp == null) ? "" : temp); //$NON-NLS-1$
				temp = em.getAttribute(XSD_ATTR_ATTR_TYPE);
				attr.setType((temp == null) ? "" : temp); //$NON-NLS-1$
				index++;
			}
		}

		/**
		 * ComplexType
		 */
		nodeList = root.getElementsByTagNameNS(Const.NS_XSD_URI,
				XSD_TAG_COMPLEXTYPE);
		index = 0;
		for (int cnt = 0; cnt < nodeList.getLength(); cnt++) {
			Element em = (Element) nodeList.item(cnt);
			if (em.getParentNode() == root) {
				MispSchemaComplexType ct = schema.addComplexType(null);
				parseComplexType(ct, em);
				index++;
			}
		}

		/**
		 * load import
		 */
		for (int cnt = 0; cnt < schema.getImportSize(); cnt++) {
			MispSchemaImport im = schema.getImport(cnt);
			loadImportSchema(schema, im, schema.getBaseDir(), schema
					.getDfsCharSet(), false);
		}

		return;
	}

	/**
	 * 複合型をパースする
	 * 
	 * @param ct
	 *            パース結果を保存する複合型オブジェクト
	 * @param root
	 *            複合型のルートエレメント
	 */
	public static void parseComplexType(MispSchemaComplexType ct, Element root) {
		String temp = root.getAttribute(XSD_ATTR_COMPLEXTYPE_NAME);
		ct.setName((temp == null) ? "" : temp); //$NON-NLS-1$

		NodeList nodeList = root.getElementsByTagNameNS(Const.NS_XSD_URI,
				XSD_TAG_COMPLEXCONTENT);
		if (nodeList.getLength() > 0) {
			Element ctElement = (Element) nodeList.item(0);
			nodeList = ctElement.getElementsByTagNameNS(Const.NS_XSD_URI,
					XSD_TAG_EXTENSION);
			if (nodeList.getLength() > 0) {
				ctElement = (Element) nodeList.item(0);
				temp = ctElement.getAttribute(XSD_ATTR_EXTENSION_BASE);
				ct.setExtension((temp == null) ? "" : temp); //$NON-NLS-1$
			}
			parseComplexTypeDetail(ct, ctElement);
		} else {
			parseComplexTypeDetail(ct, root);
		}
	}

	/**
	 * 複合型の詳細をパースする
	 * 
	 * @param ct
	 *            パース結果を保存する複合型オブジェクト
	 * @param root
	 *            複合型のルートエレメント
	 */
	private static void parseComplexTypeDetail(MispSchemaComplexType ct,
			Element root) {
		String temp;
		NodeList nodeList = root.getElementsByTagNameNS(Const.NS_XSD_URI,
				XSD_TAG_SEQUENCE);
		if (nodeList.getLength() > 0) {
			nodeList = ((Element) nodeList.item(0)).getElementsByTagNameNS(
					Const.NS_XSD_URI, XSD_TAG_ELEMENT);
			for (int cnt = 0; cnt < nodeList.getLength(); cnt++) {
				Element em = (Element) nodeList.item(cnt);
				MispSchemaElement et = ct.addElement(-1);
				temp = em.getAttribute(XSD_ATTR_ELEMENT_REF);
				if ((temp != null) && (temp.length() > 0)) {
					et.setName(temp);
					et.setFlgRef(true);
				} else {
					temp = em.getAttribute(XSD_ATTR_ELEMENT_NAME);
					et.setName((temp == null) ? "" : temp); //$NON-NLS-1$
				}
				temp = em.getAttributeNS(Const.NS_MISP_URI,
						XSD_ATTR_ELEMENT_DFSTYPE);
				if((temp == null) || (temp.length() == 0))
					temp = em.getAttribute(XSD_ATTR_ELEMENT_TYPE);
				et.setType((temp == null) ? "" : temp); //$NON-NLS-1$
				temp = em.getAttributeNS(Const.NS_MISP_URI,
						XSD_ATTR_ELEMENT_INDEXED);
				if (temp != null)
					et.setIndexedFlg(temp.equals("true")); //$NON-NLS-1$
				temp = em.getAttribute(XSD_ATTR_ELEMENT_MIN_OCCURS);
				if (temp != null)
					et.setMinOccurs(temp);
				temp = em.getAttribute(XSD_ATTR_ELEMENT_MAX_OCCURS);
				if (temp != null)
					et.setMaxOccurs(temp);
			}
		}
		nodeList = root.getElementsByTagNameNS(Const.NS_XSD_URI, XSD_TAG_ATTR);
		for (int cnt = 0; cnt < nodeList.getLength(); cnt++) {
			Element em = (Element) nodeList.item(cnt);
			MispSchemaAttribute attr = ct.addAttribute(-1);
			temp = em.getAttribute(XSD_ATTR_ATTR_REF);
			if ((temp != null) && (temp.length() > 0)) {
				attr.setName(temp);
				attr.setFlgRef(true);
			} else {
				temp = em.getAttribute(XSD_ATTR_ATTR_NAME);
				attr.setName((temp == null) ? "" : temp); //$NON-NLS-1$
			}
			temp = em.getAttribute(XSD_ATTR_ATTR_TYPE);
			attr.setType((temp == null) ? "" : temp); //$NON-NLS-1$
			temp = em.getAttributeNS(Const.NS_MISP_URI,
					XSD_ATTR_ATTR_INDEXED);
			if (temp != null)
				attr.setIndexedFlg(temp.equals("true")); //$NON-NLS-1$

		}
	}

	/**
	 * エラーフラグ
	 */
	private boolean hasError = false;

	/**
	 * エラー行番号
	 */
	private int errorLine = -1;

	/**
	 * エラーメッセージ
	 */
	private String errorMsg = ""; //$NON-NLS-1$

	/**
	 * 最近のエラーを取得する
	 * 
	 * @return 最近のエラー
	 */
	protected String getLastError() {
		if (hasError) {
			if (errorLine > 0)
				return ("Error at line " + errorLine + ":" + errorMsg); //$NON-NLS-1$ //$NON-NLS-2$
			else
				return ("Error:" + errorMsg); //$NON-NLS-1$
		}
		return ""; //$NON-NLS-1$
	}

	/**
	 * エラーをクリアする
	 * 
	 */
	protected void clearError() {
		hasError = false;
		errorLine = -1;
		errorMsg = ""; //$NON-NLS-1$
	}

	/**
	 * エラーを設定する
	 * 
	 * @param line
	 *            行番号
	 * @param msg
	 *            メッセージ
	 */
	protected void setLastError(int line, String msg) {
		hasError = true;
		errorLine = line;
		errorMsg = msg;
		log.error(getLastError());
	}

	/**
	 * エラーを設定する
	 * 
	 * @param line
	 *            行番号
	 * @param msg
	 *            メッセージ
	 */
	protected void setLastError(String msg) {
		hasError = true;
		errorLine = -1;
		errorMsg = msg;
		log.error(getLastError());
	}

	/**
	 * スキーマにあるインポートファイルをロードする
	 * 
	 * @param schema
	 *            DaRuMaスキーマ
	 * @param im
	 *            インポートオブジェクト
	 * @param baseDir
	 *            ベースフォルダ
	 * @param csvCharset
	 *            キャラクタセット
	 * @param loadHttp
	 *            インターネット上にあるファイルのロードフラグ
	 * @return 成功する場合はtrue それ以外はfalseを返す
	 */
	public boolean loadImportSchema(MispSchema schema, MispSchemaImport im,
			String baseDir, String csvCharset, boolean loadHttp) {
		boolean rc = false;
		String location = im.getLocation();

		for (int cnt = schema.getComplexTypeSize(); cnt-- > 0;) {
			MispSchemaComplexType ct = schema.getComplexType(cnt);
			if (ct.getIm() == im) {
				schema.delComplexType(cnt);
			}
		}

		try {
			if (location.startsWith("http://")) //$NON-NLS-1$
			{
				if (!loadHttp) {
					location = null;
					rc = true;
				}
			} else {
				if (!MispSchema.isAbsolutePath(location)) {
					location = MispSchema.makeFullPath(baseDir, location);
				}
			}

			if (location != null) {
				location = location.replaceAll("\\\\", "/"); //$NON-NLS-1$ //$NON-NLS-2$
				log.info("load import schema:" + location); //$NON-NLS-1$

				MispSchema importSchema = null;
				if (location.endsWith("." + MispSchemaAccess.DFS_DFT_FILE_EXT)) { //$NON-NLS-1$
					importSchema = loadSchemaFromCsv(location, csvCharset,
							MispSchemaAccess.DFS_DFT_SEPARATOR,
							MispSchemaAccess.DFS_DFT_PARENTHESIS,
							MispSchemaAccess.DFS_DFT_COMMENT);
				} else if (location.endsWith("." //$NON-NLS-1$
						+ MispSchemaAccess.XSD_DFT_FILE_EXT)) {
					importSchema = loadSchemaFromXml(location, csvCharset);
				}
				if (importSchema != null) {
					im.setUri(importSchema.getTargetNamespace());

					MispSchemaNamespace tns = null;
					if (!im.isIncludeFlg()) {
						// Import元のターゲット名前空間を設定する
						tns = schema.findNamespaceByUri(importSchema
								.getTargetNamespace());
						if (tns == null) {
							tns = schema.addNamespaceAutoPrefix(importSchema
									.getTargetNamespace());
						}
					}

					// Import元の名前空間をインポート
					for (int cnt = 0; cnt < importSchema.getNamespaceSize(); cnt++) {
						MispSchemaNamespace importNs = importSchema
								.getNamespace(cnt);
						MispSchemaNamespace currNs = schema
								.findNamespaceByUri(importNs.getUri());
						if (currNs == null) {
							if (schema.findNamespaceByPrefix(importNs
									.getPrefix()) == null) {
								currNs = schema.addNamespace();
								importNs.copyTo(currNs);
							} else {
								currNs = schema.addNamespaceAutoPrefix(importNs
										.getUri());
							}
						}
						if (!importNs.getPrefix().equals(currNs.getPrefix())) {
							MispSchemaNamespace temp2 = importSchema
									.findNamespaceByPrefix(currNs.getPrefix());
							if (temp2 != null) {
								temp2.setPrefixReference(importSchema
										.getAutoPrefix());
							}
							importNs.setPrefixReference(currNs.getPrefix());
						}
					}

					// Import元のルート要素をインポート
					for (int cnt = 0; cnt < importSchema.getRootElementSize(); cnt++) {
						MispSchemaElement et = importSchema.getRootElement(cnt);
						MispSchemaElement newEt = schema.addRootElement();
						et.copyToWithoutRef(newEt);
					}

					// Import元のルート属性をインポート
					for (int cnt = 0; cnt < importSchema.getRootAttributeSize(); cnt++) {
						MispSchemaAttribute attr = importSchema
								.getRootAttribute(cnt);
						MispSchemaAttribute newAttr = schema.addRootAttribute();
						attr.copyToWithoutRef(newAttr);
					}

					// Import元の複合型をインポート
					for (int cnt = 0; cnt < importSchema.getComplexTypeSize(); cnt++) {
						MispSchemaComplexType ct = importSchema
								.getComplexType(cnt);

						// 存在しるかどうかをチェック
						Object[] oldCts = schema.findComplexTypesByName(ct
								.getName(), ct.getFilePath());
						if (oldCts.length > 0)
							continue;

						MispSchemaComplexType newCt = schema.addComplexType(im);
						ct.copyToWithoutRef(newCt);
						if (ct.getKind() == MispSchema.KIND_INTERNAL)
							newCt.setFilePath(importSchema.getFilePath());
						else
							newCt.setFilePath(ct.getFilePath());

						if (newCt.getKind() == MispSchema.KIND_IMPORT) {
							newCt.setPrefixAll(tns.getPrefix());
						}
					}
					schema.sortComplexType();
					im.setLoadResult("");
					rc = true;
				} else {
					im.setLoadResult(MispSchemaMessages
							.getString("MispSchemaCsv.17")); //$NON-NLS-1$
				}
			}

		} catch (Exception e) {
			log.error("Fail to load import file:" + location); //$NON-NLS-1$
		}
		return rc;
	}

}
