package daruma.bus.processor.csv;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.List;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xerces.parsers.DOMParser;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import daruma.bus.processor.csv.MispSchema.MispSchemaComplexType;
import daruma.bus.processor.csv.MispSchema.MispSchemaElement;
import daruma.bus.processor.csv.MispSchema.MispSchemaNamespace;
import daruma.bus.processor.csv.MispSchemaCx.MispSchemaCxColumn;
import daruma.bus.processor.csv.MispSchemaCx.MispSchemaCxColumn.XpathDetail;
import daruma.bus.processor.csv.MispSchemaCx.MispSchemaCxColumn.XmlTagDef;
import daruma.bus.processor.csv.MispSchemaUtil.MispSchemaXpath;

/**
 * DARUMAスキーマのCsv2Xmlデータをアクセスするためのクラス
 * 
 * @author shumi
 * 
 */
public class MispSchemaCxAccess {
	/**
	 * Csv2Xmlデータファイルの拡張子
	 */
	final public static String DFT_FILE_EXT = "xml";

	static private final String EL_CONFIG = "Config";

	static private final String EL_CONFIG_FULL = Const.NS_CX_PRE + ":"
			+ EL_CONFIG;

	static private final String EL_XML = "xml";

	static private final String EL_XML_FULL = Const.NS_CX_PRE + ":" + EL_XML;

	static private final String EL_CSV = "csv";

	static private final String EL_CSV_FULL = Const.NS_CX_PRE + ":" + EL_CSV;

	static private final String EL_COLUMN_LIST = "columnList";

	static private final String EL_COLUMN_LIST_FULL = Const.NS_CX_PRE + ":"
			+ EL_COLUMN_LIST;

	static private final String EL_COLUMN = "column";

	static private final String EL_COLUMN_FULL = Const.NS_CX_PRE + ":"
			+ EL_COLUMN;

	static private final String EL_COLUMN_ATTR_NAME = "name";

	static private final String EL_COLUMN_ATTR_XPATH = "xpath";

	/**
	 * ロガー
	 */
	static private Log log = LogFactory.getLog(MispSchemaCxAccess.class);

	/**
	 * MispSchemaAccess
	 */
	MispSchemaAccess schemaAccess = new MispSchemaAccess();

	/**
	 * スキーマおよびCSV項目をXML形式で保存する
	 * 
	 * @param cx
	 *            保存するスキーマおよびCSV項目
	 * @param fileName
	 *            ファイル名
	 * @param outputCharSet
	 *            出力文字セット
	 * @return 成功する場合はTrue, それ以外はFalseを返す
	 */
	public boolean saveSchemaCxToXml(MispSchemaCx cx, String fileName,
			String outputCharSet) {
		return saveSchemaCxToXml(cx, fileName, outputCharSet, outputCharSet);
	}

	/**
	 * スキーマおよびCSV項目をXML形式で保存する
	 * 
	 * @param cx
	 *            保存するスキーマおよびCSV項目
	 * @param fileName
	 *            ファイル名
	 * @param outputCharSet
	 *            出力文字セット
	 * @param xmlEncoding
	 *            XML文字セット
	 * @return 成功する場合はTrue, それ以外はFalseを返す
	 */
	public boolean saveSchemaCxToXml(MispSchemaCx cx, String fileName,
			String outputCharSet, String xmlEncoding) {
		try {
			FileOutputStream fos = new FileOutputStream(fileName);
			return saveSchemaCxToXml(cx, fos, outputCharSet, xmlEncoding);
		} catch (Exception e) {
			setLastError(e.getMessage());
		}
		return false;
	}

	/**
	 * スキーマおよびCSV項目をXML形式で保存する
	 * 
	 * @param cx
	 *            保存するスキーマおよびCSV項目
	 * @param os
	 *            出力ストリーム
	 * @param outputCharSet
	 *            出力文字セット
	 * @return 成功する場合はTrue, それ以外はFalseを返す
	 */
	public boolean saveSchemaCxToXml(MispSchemaCx cx, OutputStream os,
			String outputCharSet) {
		return saveSchemaCxToXml(cx, os, outputCharSet, outputCharSet);
	}

	/**
	 * スキーマおよびCSV項目をXML形式で保存する
	 * 
	 * @param cx
	 *            保存するスキーマおよびCSV項目
	 * @param os
	 *            出力ストリーム
	 * @param outputCharSet
	 *            出力文字セット
	 * @param xmlEncoding
	 *            XML文字セット
	 * @return 成功する場合はTrue, それ以外はFalseを返す
	 */
	public boolean saveSchemaCxToXml(MispSchemaCx cx, OutputStream os,
			String outputCharSet, String xmlEncoding) {
		try {
			OutputStreamWriter osw;
			if ((outputCharSet != null) && (outputCharSet.length() > 0)) {
				osw = new OutputStreamWriter(os, outputCharSet);
			} else {
				osw = new OutputStreamWriter(os);
			}
			return saveSchemaCxToXml(cx, osw, xmlEncoding);
		} catch (Exception e) {
			setLastError(e.getMessage());
		}
		return false;
	}

	/**
	 * スキーマおよびCSV項目をXML形式で保存する
	 * 
	 * @param cx
	 *            保存するスキーマおよびCSV項目
	 * @param osw
	 *            出力ストリームライター
	 * @return 成功する場合はTrue, それ以外はFalseを返す
	 */
	public boolean saveSchemaCxToXml(MispSchemaCx cx, OutputStreamWriter osw) {
		return saveSchemaCxToXml(cx, osw, osw.getEncoding());
	}

	/**
	 * スキーマおよびCSV項目をXML形式で保存する
	 * 
	 * @param cx
	 *            保存するスキーマおよびCSV項目
	 * @param osw
	 *            出力ストリームライター
	 * @param xmlEncoding
	 *            XML文字セット
	 * @return 成功する場合はTrue, それ以外はFalseを返す
	 */
	public boolean saveSchemaCxToXml(MispSchemaCx cx, OutputStreamWriter osw,
			String xmlEncoding) {
		try {
			// Output xml header
			osw.write("<?xml version=\"1.0\" encoding=\"" + xmlEncoding
					+ "\"?>" + MispSchema.CRLF);

			String prefix = "";
			// cx:config
			osw.write(prefix + "<" + EL_CONFIG_FULL + MispSchema.CRLF);
			osw.write(prefix + "  " + Const.NS_XMLNS_PRE + ":"
					+ Const.NS_CX_PRE + "=\"" + Const.NS_CX_URI + "\">"
					+ MispSchema.CRLF);

			// cx:xml
			osw.write(prefix + "  <" + EL_XML_FULL + ">" + MispSchema.CRLF);
			MispSchemaAccess.outputXml(cx.getSchema(), osw, "    ");
			osw.write(prefix + "  </" + EL_XML_FULL + ">" + MispSchema.CRLF);

			// cx:csv
			osw.write(prefix + "  <" + EL_CSV_FULL);
			osw
					.write("  element=\""
							+ cx.getSchema().getFeatureElement().getName()
							+ "\" cs=\",\" rs=\"\\n\" quoteChar='\"' commentPrefix=\"#\""
							+ ">" + MispSchema.CRLF);
			osw.write(prefix + "    <" + EL_COLUMN_LIST_FULL + ">"
					+ MispSchema.CRLF);
			for (int cnt = 0; cnt < cx.getCxColumnSize(); cnt++) {
				MispSchemaCxColumn cxColumn = cx.getCxColumn(cnt);
				osw.write(prefix + "      <" + EL_COLUMN_FULL);
				osw.write(" " + EL_COLUMN_ATTR_XPATH + "=\""
						+ cxColumn.getXpath() + "\"");
				osw.write(" " + EL_COLUMN_ATTR_NAME + "=\""
						+ cxColumn.getName() + "\"");
				osw.write("/>" + MispSchema.CRLF);
			}
			osw.write(prefix + "    </" + EL_COLUMN_LIST_FULL + ">"
					+ MispSchema.CRLF);
			osw.write(prefix + "  </" + EL_CSV_FULL + ">" + MispSchema.CRLF);

			osw.write(prefix + "</" + EL_CONFIG_FULL + ">" + MispSchema.CRLF);

			osw.flush();
			osw.close();
			return true;
		} catch (Exception e) {
			if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
				e.printStackTrace();
			}
			setLastError(e.getMessage());
		}
		return false;
	}

	/**
	 * エラーフラグ
	 */
	private boolean hasError = false;

	/**
	 * エラー行番号
	 */
	private int errorLine = -1;

	/**
	 * エラーメッセージ
	 */
	private String errorMsg = "";

	/**
	 * 最近のエラーを取得する
	 * 
	 * @return 最近のエラー
	 */
	protected String getLastError() {
		if (hasError) {
			if (errorLine > 0)
				return ("Error at line " + errorLine + ":" + errorMsg);
			else
				return ("Error:" + errorMsg);
		}
		return "";
	}

	/**
	 * エラーを設定する
	 * 
	 * @param msg
	 *            設定するエラー
	 */
	protected void setLastError(String msg) {
		hasError = true;
		errorMsg = msg;
		log.error(getLastError());
	}

	/**
	 * XML文字列をエスケープする
	 * 
	 * @param src
	 *            ソース文字列
	 * @return エスケープされた文字列
	 */
	public static String escapeXmlString(String src) {
		src = src.replaceAll("&", "&amp;");
		src = src.replaceAll("<", "&lt;");
		src = src.replaceAll(">", "&gt;");
		src = src.replaceAll("\"", "&quot;");
		src = src.replaceAll("'", "&apos;");
		return src;
	}

	/**
	 * XMLファイルからDARUMAスキーマおよびCSV項目をロードする
	 * 
	 * @param fileName
	 *            ファイル名
	 * @return DARUMAスキーマおよびCSV項目
	 */
	public MispSchemaCx loadSchemaCxFromXml(String fileName, String dfsCharset) {
		MispSchemaCx cx = new MispSchemaCx();
		try {
			InputStream fis = new FileInputStream(fileName);
			DOMParser parser = new DOMParser();
			parser.setEntityResolver(new EntityResolver() {
				public InputSource resolveEntity(java.lang.String publicId,
						java.lang.String systemId) throws SAXException,
						java.io.IOException {
					if (Const.DTD_URI.equals(systemId)) {
						return new InputSource(new ByteArrayInputStream(
								"<?xml version='1.0' encoding='UTF-8'?>"
										.getBytes()));
					}
					return null;
				}
			});

			cx.getSchema().setFilePath(fileName);
			cx.getSchema().setDfsCharSet(dfsCharset);
			parser.parse(new InputSource(fis));
			Document doc = parser.getDocument();
			Element root = doc.getDocumentElement();
			if (root.getNamespaceURI().equals(Const.NS_CX_URI)
					&& root.getLocalName().equals(EL_CONFIG)) {
				NodeList nodeList = root.getElementsByTagNameNS(
						Const.NS_CX_URI, EL_XML);
				if (nodeList.getLength() > 0) {
					Element element = (Element) nodeList.item(0);
					nodeList = element.getElementsByTagNameNS(Const.NS_XSD_URI,
							MispSchemaAccess.XSD_TAG_SCHEMA);
					if (nodeList.getLength() > 0) {
						schemaAccess.parseSchema(cx.getSchema(),
								(Element) nodeList.item(0));
					}
				}

				nodeList = root.getElementsByTagNameNS(Const.NS_CX_URI, EL_CSV);
				if (nodeList.getLength() > 0) {
					Element element = (Element) nodeList.item(0);
					nodeList = element.getElementsByTagNameNS(Const.NS_CX_URI,
							EL_COLUMN_LIST);
					if (nodeList.getLength() > 0) {
						parserCxColumn(cx, (Element) nodeList.item(0));
					}
				}

			}
			return cx;
		} catch (Exception e) {
			log.error("loadSchemaCxFromXml fail" + e.getMessage());
			if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
				e.printStackTrace();
			}
		}

		return null;
	}

	/**
	 * CSV項目をパースする
	 * 
	 * @param cx
	 *            CSV項目を保存するオブジェクト
	 * @param root
	 *            XMLルート要素
	 */
	private void parserCxColumn(MispSchemaCx cx, Element root) {
		NodeList nodeList = root.getElementsByTagNameNS(Const.NS_CX_URI,
				EL_COLUMN);
		for (int cnt = 0; cnt < nodeList.getLength(); cnt++) {
			Element em = (Element) nodeList.item(cnt);
			MispSchemaCxColumn column = cx.addCxColumn(-1);
			column.setName(em.getAttribute(EL_COLUMN_ATTR_NAME));
			column.setXpath(em.getAttribute(EL_COLUMN_ATTR_XPATH));
		}
	}

	/**
	 * CSVデータをXmlデータへ変換するためのXSLTファイルを生成する
	 * 
	 * @param cx
	 *            Csv2Xmlデータ
	 * @param fileName
	 *            ファイル名
	 * @param outputCharSet
	 *            出力キャラクタセット
	 * @param xmlEncoding
	 *            XMLエンコード
	 * @return 成功した場合はtrue それ以外はfalseを返す
	 */
	public boolean createXsltCsv2Xml(MispSchemaCx cx, String fileName,
			String outputCharSet, String xmlEncoding) {
		boolean rc = false;
		try {
			// Step1
			String tmpFile = fileName + ".tmp";
			FileOutputStream os = new FileOutputStream(tmpFile);
			OutputStreamWriter osw;
			if ((outputCharSet != null) && (outputCharSet.length() > 0)) {
				osw = new OutputStreamWriter(os, outputCharSet);
			} else {
				osw = new OutputStreamWriter(os);
			}
			if (createXsltCsv2Xml(cx, osw, xmlEncoding)) {
				// Step2
				FileInputStream is = new FileInputStream(tmpFile);
				os = new FileOutputStream(fileName);
				SortXmlTag sorter = new SortXmlTag(Const.NS_XSL_URI,
						"template", "_");
				sorter.transform(is, os);
				rc = true;
			}
			File tmp = new File(tmpFile);
			tmp.delete();
		} catch (Exception e) {
			setLastError(e.getMessage());
		}
		return rc;
	}

	/**
	 * CSVデータをXmlデータへ変換するためのXSLTファイルを生成する
	 * 
	 * @param cx
	 *            Csv2Xmlデータ
	 * @param osw
	 *            出力ストリームライター
	 * @param xmlEncoding
	 *            XMLエンコード
	 * @return 成功した場合はtrue それ以外はfalseを返す
	 */
	private boolean createXsltCsv2Xml(MispSchemaCx cx, OutputStreamWriter osw,
			String xmlEncoding) {
		boolean rc = false;

		try {
			MispSchema schema = cx.getSchema();

			/**
			 * Export extenstion member
			 */
			List cxColumnList = cx.getSortedCxColumnList();
			if (cxColumnList != null) {
				// Output xml header
				osw.write("<?xml version=\"1.0\" encoding=\"" + xmlEncoding
						+ "\"?>" + MispSchema.CRLF);

				// Output namespace
				osw.write("<" + Const.NS_XSL_PRE
						+ ":stylesheet version=\"2.0\"");

				osw.write(" " + Const.NS_XMLNS_PRE + ":" + Const.NS_XSL_PRE
						+ "=\"" + Const.NS_XSL_URI + "\"");
				osw.write(" " + Const.NS_XMLNS_PRE + ":" + Const.NS_CSVX_PRE
						+ "=\"" + Const.NS_CSVX_URI + "\"");
				boolean hasMisp = false;
				for (int cnt = 0; cnt < schema.getNamespaceSize(); cnt++) {
					MispSchemaNamespace ns = schema.getNamespace(cnt);
					osw.write(" " + Const.NS_XMLNS_PRE + ":" + ns.getPrefix()
							+ "=\"" + ns.getUri() + "\"");
					if (ns.getPrefix().equals(Const.NS_MISP_PRE)
							&& ns.getUri().equals(Const.NS_MISP_URI))
						hasMisp = true;
				}
				if (!hasMisp)
					osw.write(" " + Const.NS_XMLNS_PRE + ":"
							+ Const.NS_MISP_PRE + "=\"" + Const.NS_MISP_URI
							+ "\"");
				osw.write(">" + MispSchema.CRLF);

				// Output XML output
				osw.write("<" + Const.NS_XSL_PRE
						+ ":output method=\"xml\" version=\"1.0\""
						+ " encoding=\"" + xmlEncoding + "\" indent=\"yes\"/>"
						+ MispSchema.CRLF);

				// Output Template
				osw.write("<" + Const.NS_XSL_PRE + ":template match=\"/\">"
						+ MispSchema.CRLF);
				osw
						.write("<" + Const.NS_MISP_PRE + ":Transaction "
								+ Const.NS_XMLNS_PRE + ":" + Const.NS_MISP_PRE
								+ "=\"" + Const.NS_MISP_URI + "\" "
								+ Const.NS_XMLNS_PRE + ":" + Const.NS_TNS_PRE
								+ "=\"" + schema.getTargetNamespace() + "\">"
								+ MispSchema.CRLF);
				osw.write("<" + Const.NS_MISP_PRE + ":Insert>"
						+ MispSchema.CRLF);
				osw.write("<" + Const.NS_XSL_PRE + ":apply-templates select=\""
						+ Const.NS_CSVX_PRE + ":data\"/>" + MispSchema.CRLF);
				osw.write("</" + Const.NS_MISP_PRE + ":Insert>"
						+ MispSchema.CRLF);
				osw.write("</" + Const.NS_MISP_PRE + ":Transaction>"
						+ MispSchema.CRLF);
				osw.write("</" + Const.NS_XSL_PRE + ":template>"
						+ MispSchema.CRLF);

				// Output csvs:data template
				osw.write("<" + Const.NS_XSL_PRE + ":template match=\""
						+ Const.NS_CSVX_PRE + ":data\">" + MispSchema.CRLF);
				osw.write("<" + Const.NS_XSL_PRE + ":apply-templates select=\""
						+ Const.NS_CSVX_PRE + ":row\"/>" + MispSchema.CRLF);
				osw.write("</" + Const.NS_XSL_PRE + ":template>"
						+ MispSchema.CRLF);

				// Output csvs:data template
				osw.write("<" + Const.NS_XSL_PRE + ":template match=\""
						+ Const.NS_CSVX_PRE + ":row\">" + MispSchema.CRLF);
				osw.write("<" + schema.getFeatureElement().getName() + " "
						+ Const.NS_XMLNS_PRE + "=\""
						+ schema.getTargetNamespace() + "\">");
				outputCxColumnList(cx, osw, cxColumnList);

				osw.write("</" + schema.getFeatureElement().getName() + ">");
				osw.write("</" + Const.NS_XSL_PRE + ":template>"
						+ MispSchema.CRLF);

				osw.write("</" + Const.NS_XSL_PRE + ":stylesheet>");
				osw.flush();
				osw.close();
				rc = true;
			}

		} catch (Exception e) {
			if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
				e.printStackTrace();
			}
			setLastError(e.getMessage());
		}
		return rc;
	}

	/**
	 * CSV項目をXmlデータへ変換するためのXSLTファイルを生成する
	 * 
	 * @param cx
	 *            Csv2Xmlデータ
	 * @param osw
	 *            出力ストリームライター
	 * @param cxColumnList
	 *            CSV項目リスト
	 */
	private void outputCxColumnList(MispSchemaCx cx, OutputStreamWriter osw,
			List cxColumnList) {

		try {
			for (int currOutputIndex = 0; currOutputIndex < cxColumnList.size(); currOutputIndex++) {
				MispSchemaCxColumn col = (MispSchemaCxColumn) cxColumnList
						.get(currOutputIndex);
				XpathDetail def = col.parserXpath();
				String columnName = "" + Const.NS_CSVX_PRE + ":column[@name='"
						+ col.getName() + "']";

				for (int cnti = 0; cnti < def.getTagDefList().size(); cnti++) {
					XmlTagDef tagDef = (XmlTagDef) def.getTagDefList()
							.get(cnti);
					if (tagDef.getIndex() > 0) {
						for (int cntj = 1; cntj < tagDef.getIndex(); cntj++) {
							osw.write("<" + tagDef.getName() + "_" + cntj
									+ "/>");
						}
					}
					osw.write("<" + tagDef.getName() + "_" + tagDef.getIndex()
							+ ">");
				}

				if (def.isBlAttr()) {
					osw.write("<" + Const.NS_XSL_PRE + ":attribute_" + currOutputIndex + " name=\""
							+ def.getName() + "\">" + MispSchema.CRLF);
					osw.write("<" + Const.NS_XSL_PRE + ":value-of select=\""
							+ Const.NS_CSVX_PRE + ":column[@name='"
							+ col.getName() + "']\"/>" + MispSchema.CRLF);
					osw.write("</" + Const.NS_XSL_PRE + ":attribute_" + currOutputIndex + ">"
							+ MispSchema.CRLF);
				} else {
					if (def.getIndex() > 0) {
						for (int cntj = 1; cntj < def.getIndex(); cntj++) {
							osw.write("<" + def.getName() + "_" + cntj + "/>");
						}
					}

					Object ref = col.getXpathRef();
					String dfsType = "";
					if ((ref != null) && (ref instanceof MispSchemaElement)) {
						dfsType = ((MispSchemaElement) ref).getType();
					}

					osw.write("<" + def.getName() + "_" + def.getIndex() + " "
							+ MispSchemaAccess.XSD_FULL_ATTR_ELEMENT_DFSTYPE
							+ "=\"" + dfsType + "\">");

					osw.write("<" + Const.NS_XSL_PRE + ":value-of select=\""
							+ columnName + "\"/>" + MispSchema.CRLF);

					osw
							.write("</" + def.getName() + "_" + def.getIndex()
									+ ">");
				}

				for (int cnti = def.getTagDefList().size(); cnti-- > 0;) {
					XmlTagDef tagDef = (XmlTagDef) def.getTagDefList()
							.get(cnti);
					osw.write("</" + tagDef.getName() + "_" + tagDef.getIndex()
							+ ">");
				}
				osw.write("\r\n");
			}
		} catch (Exception e) {

		}
	}

	/**
	 * NULLタグを削除するためのXSLTファイルを生成する
	 * 
	 * @param cx
	 *            Csv2Xmlデータ
	 * @param fileName
	 *            ファイル名
	 * @param outputCharSet
	 *            出力キャラクタセット
	 * @param xmlEncoding
	 *            XMLエンコード
	 * @return 成功した場合はtrue それ以外はfalseを返す
	 */
	public boolean createXsltDeleteNullTag(MispSchemaCx cx, String fileName,
			String outputCharSet, String xmlEncoding) {
		try {
			FileOutputStream os = new FileOutputStream(fileName);
			OutputStreamWriter osw;
			if ((outputCharSet != null) && (outputCharSet.length() > 0)) {
				osw = new OutputStreamWriter(os, outputCharSet);
			} else {
				osw = new OutputStreamWriter(os);
			}
			createXsltDeleteNullTag(cx, osw, xmlEncoding);

		} catch (Exception e) {
			setLastError(e.getMessage());
		}
		return false;
	}

	/**
	 * NULLタグを削除するためのXSLTファイルを生成する
	 * 
	 * @param cx
	 *            Csv2Xmlデータ
	 * @param osw
	 *            出力ストリームライター
	 * @param xmlEncoding
	 *            XMLエンコード
	 * @return 成功した場合はtrue それ以外はfalseを返す
	 */
	public boolean createXsltDeleteNullTag(MispSchemaCx cx,
			OutputStreamWriter osw, String xmlEncoding) {
		boolean rc = false;

		try {
			MispSchema schema = cx.getSchema();

			// Output xml header
			osw.write("<?xml version=\"1.0\" encoding=\"" + xmlEncoding
					+ "\"?>" + MispSchema.CRLF);

			// Output namespace
			osw.write("<" + Const.NS_XSL_PRE + ":stylesheet version=\"1.0\"");

			osw.write(" " + Const.NS_XMLNS_PRE + ":" + Const.NS_XSL_PRE + "=\""
					+ Const.NS_XSL_URI + "\"");
			osw.write(" " + Const.NS_XMLNS_PRE + ":" + Const.NS_TNS_PRE + "=\""
					+ schema.getTargetNamespace() + "\"");
			boolean hasMisp = false;
			for (int cnt = 0; cnt < schema.getNamespaceSize(); cnt++) {
				MispSchemaNamespace ns = schema.getNamespace(cnt);
				osw.write(" " + Const.NS_XMLNS_PRE + ":" + ns.getPrefix()
						+ "=\"" + ns.getUri() + "\"");
				if (ns.getPrefix().equals(Const.NS_MISP_PRE)
						&& ns.getUri().equals(Const.NS_MISP_URI))
					hasMisp = true;
			}
			if (!hasMisp) {
				osw.write(" " + Const.NS_XMLNS_PRE + ":" + Const.NS_MISP_PRE
						+ "=\"" + Const.NS_MISP_URI + "\"");
			}
			osw.write(">" + MispSchema.CRLF);

			// Output XML output
			osw.write("<" + Const.NS_XSL_PRE
					+ ":output method=\"xml\" version=\"1.0\"" + " encoding=\""
					+ xmlEncoding + "\" indent=\"yes\"/>" + MispSchema.CRLF);

			// Output template
			List ctXpathList = MispSchemaUtil.getDeleteNullXpathList(schema,
					Const.NS_TNS_PRE);
			for (int cnt = 0; cnt < ctXpathList.size(); cnt++) {
				List ctChecmList = null;
				MispSchemaXpath xpath = (MispSchemaXpath) ctXpathList.get(cnt);
				MispSchemaElement em = (MispSchemaElement) xpath.getRef();
				switch (schema.getElementTypeGrp(em.getType())) {
				case MispSchema.GRP_COMPLEX_INTERNAL:
				case MispSchema.GRP_COMPLEX_EXTERNAL:
					MispSchemaComplexType childEt = schema
							.findComplexTypeByName(em.getType());
					if (childEt != null)
						ctChecmList = MispSchemaUtil.getXPathList(childEt,
								Const.NS_TNS_PRE);
					break;
				default:
					ctChecmList = new ArrayList();
					ctChecmList.add(null);
					break;

				}
				String tagName;
				if (ctChecmList != null) {
					osw.write("<" + Const.NS_XSL_PRE + ":template match=\"/"
							+ Const.NS_MISP_PRE + ":Transaction/"
							+ Const.NS_MISP_PRE + ":Insert/" + ""
							+ Const.NS_TNS_PRE + ":"
							+ schema.getFeatureElement().getName() + "/"
							+ xpath.getXpath() + "\">" + MispSchema.CRLF);
					osw.write("<" + Const.NS_XSL_PRE + ":if test=\"");
					for (int cntj = 0; cntj < ctChecmList.size(); cntj++) {
						Object child = ctChecmList.get(cntj);
						String name;
						if (child instanceof MispSchemaXpath)
							name = ((MispSchemaXpath) child).getXpath();
						else
							name = ".";
						if (cntj > 0)
							osw.write(" or ");
						osw.write("(string-length(" + name + ") > 0)");
					}
					osw.write("\">" + MispSchema.CRLF);

					if (em.getNamePrefix().length() > 0)
						tagName = em.getName();
					else
						tagName = Const.NS_TNS_PRE + ":" + em.getName();

					osw.write("<" + tagName + "><" + Const.NS_XSL_PRE
							+ ":apply-templates select=\"@*|node()\"/></"
							+ tagName + ">" + MispSchema.CRLF);

					osw.write("</" + Const.NS_XSL_PRE + ":if>"
							+ MispSchema.CRLF);

					osw.write("</" + Const.NS_XSL_PRE + ":template>"
							+ MispSchema.CRLF);

				}
			}

			// Output Default Template
			osw.write("<" + Const.NS_XSL_PRE + ":template match=\"@*|node()\">"
					+ MispSchema.CRLF);
			osw.write("<" + Const.NS_XSL_PRE + ":copy>" + MispSchema.CRLF);
			osw.write("<" + Const.NS_XSL_PRE
					+ ":apply-templates select=\"@*|node()\"/>"
					+ MispSchema.CRLF);
			osw.write("</" + Const.NS_XSL_PRE + ":copy>" + MispSchema.CRLF);
			osw.write("</" + Const.NS_XSL_PRE + ":template>" + MispSchema.CRLF);

			osw.write("</" + Const.NS_XSL_PRE + ":stylesheet>");
			osw.flush();
			osw.close();

		} catch (Exception e) {
			if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
				e.printStackTrace();
			}
			setLastError(e.getMessage());
		}

		return rc;
	}

	/**
	 * XMLデータをCSVデータへ変換するためのXSLTファイルを生成する
	 * 
	 * @param cx
	 *            Csv2Xmlデータ
	 * @param fileName
	 *            ファイル名
	 * @param outputCharSet
	 *            出力キャラクタセット
	 * @param xmlEncoding
	 *            XMLエンコード
	 * @return 成功した場合はtrue それ以外はfalseを返す
	 */
	public boolean createXsltXml2Csv(MispSchemaCx cx, String fileName,
			String outputCharSet, String xmlEncoding) {
		try {
			FileOutputStream os = new FileOutputStream(fileName);
			OutputStreamWriter osw;
			if ((outputCharSet != null) && (outputCharSet.length() > 0)) {
				osw = new OutputStreamWriter(os, outputCharSet);
			} else {
				osw = new OutputStreamWriter(os);
			}
			createXsltXml2Csv(cx, osw, xmlEncoding);

		} catch (Exception e) {
			setLastError(e.getMessage());
		}
		return false;
	}

	/**
	 * XMLデータをCSVデータへ変換するためのXSLTファイルを生成する
	 * 
	 * @param cx
	 *            Csv2Xmlデータ
	 * @param osw
	 *            出力ストリームライター
	 * @param xmlEncoding
	 *            XMLエンコード
	 * @return 成功した場合はtrue それ以外はfalseを返す
	 */
	private boolean createXsltXml2Csv(MispSchemaCx cx, OutputStreamWriter osw,
			String xmlEncoding) {
		boolean rc = false;

		try {
			MispSchema schema = cx.getSchema();

			// Output xml header
			osw.write("<?xml version=\"1.0\" encoding=\"" + xmlEncoding
					+ "\"?>" + MispSchema.CRLF);

			// Output namespace
			osw.write("<" + Const.NS_XSL_PRE + ":stylesheet version=\"2.0\"");
			osw.write(" " + Const.NS_XMLNS_PRE + ":" + Const.NS_XSL_PRE + "=\""
					+ Const.NS_XSL_URI + "\"");
			osw.write(" " + Const.NS_XMLNS_PRE + ":" + Const.NS_CSVX_PRE
					+ "=\"" + Const.NS_CSVX_URI + "\"");
			osw.write(" " + Const.NS_XMLNS_PRE + ":" + Const.NS_TNS_PRE + "=\""
					+ schema.getTargetNamespace() + "\"");

			boolean hasMisp = false;
			boolean hasGml = false;
			for (int cnt = 0; cnt < schema.getNamespaceSize(); cnt++) {
				MispSchemaNamespace ns = schema.getNamespace(cnt);
				osw.write(" " + Const.NS_XMLNS_PRE + ":" + ns.getPrefix()
						+ "=\"" + ns.getUri() + "\"");
				if (ns.getPrefix().equals(Const.NS_MISP_PRE)
						&& ns.getUri().equals(Const.NS_MISP_URI))
					hasMisp = true;
				if (ns.getPrefix().equals(Const.NS_GML_PRE)
						&& ns.getUri().equals(Const.NS_GML_URI))
					hasGml = true;
			}
			if (!hasMisp)
				osw.write(" " + Const.NS_XMLNS_PRE + ":" + Const.NS_MISP_PRE
						+ "=\"" + Const.NS_MISP_URI + "\"");
			if (!hasGml)
				osw.write(" " + Const.NS_XMLNS_PRE + ":" + Const.NS_GML_PRE
						+ "=\"" + Const.NS_GML_URI + "\"");

			osw.write(">" + MispSchema.CRLF);

			osw.write("<" + Const.NS_XSL_PRE + ":template match=\""
					+ Const.NS_MISP_PRE + ":GetFeatureResponse\">"
					+ MispSchema.CRLF);
			osw.write("<" + Const.NS_XSL_PRE + ":apply-templates select=\""
					+ Const.NS_MISP_PRE + ":FeatureCollection\"/>"
					+ MispSchema.CRLF);
			osw.write("</" + Const.NS_XSL_PRE + ":template>" + MispSchema.CRLF);

			osw.write("<" + Const.NS_XSL_PRE + ":template match=\""
					+ Const.NS_MISP_PRE + ":ElapsedTime\"/>" + MispSchema.CRLF);

			osw.write("<" + Const.NS_XSL_PRE + ":template match=\""
					+ Const.NS_MISP_PRE + ":FeatureCollection\">"
					+ MispSchema.CRLF);
			osw.write("<" + Const.NS_CSVX_PRE + ":data>" + MispSchema.CRLF);
			osw
					.write("<" + Const.NS_XSL_PRE
							+ ":apply-templates select=\"" + Const.NS_GML_PRE
							+ ":featureMember\"/>" + MispSchema.CRLF);
			osw.write("</" + Const.NS_CSVX_PRE + ":data>" + MispSchema.CRLF);
			osw.write("</" + Const.NS_XSL_PRE + ":template>" + MispSchema.CRLF);

			osw.write("<" + Const.NS_XSL_PRE + ":template match=\""
					+ Const.NS_GML_PRE + ":featureMember\">" + MispSchema.CRLF);
			osw.write("<" + Const.NS_XSL_PRE + ":variable name=\"rowcount\">"
					+ MispSchema.CRLF);
			osw.write("<" + Const.NS_XSL_PRE + ":number/>" + MispSchema.CRLF);
			osw.write("</" + Const.NS_XSL_PRE + ":variable>" + MispSchema.CRLF);
			osw.write("<" + Const.NS_CSVX_PRE + ":row count=\"{$rowcount}\">"
					+ MispSchema.CRLF);
			osw.write("<" + Const.NS_XSL_PRE + ":apply-templates select=\""
					+ Const.NS_TNS_PRE + ":"
					+ schema.getFeatureElement().getName() + "\"/>"
					+ MispSchema.CRLF);
			osw.write("</" + Const.NS_CSVX_PRE + ":row>" + MispSchema.CRLF);
			osw.write("</" + Const.NS_XSL_PRE + ":template>" + MispSchema.CRLF);

			osw.write("<" + Const.NS_XSL_PRE + ":template match=\""
					+ Const.NS_TNS_PRE + ":"
					+ schema.getFeatureElement().getName() + "\">"
					+ MispSchema.CRLF);
			for (int cnt = 0; cnt < cx.getCxColumnSize(); cnt++) {
				MispSchemaCxColumn col = cx.getCxColumn(cnt);
				osw.write("<" + Const.NS_CSVX_PRE + ":column pos=\""
						+ (cnt + 1) + "\" name=\"" + col.getName() + "\">"
						+ MispSchema.CRLF);
				osw.write("<" + Const.NS_XSL_PRE + ":value-of select=\""
						+ col.getFullXpath(Const.NS_TNS_PRE) + "\"/>"
						+ MispSchema.CRLF);
				osw.write("</" + Const.NS_CSVX_PRE + ":column>"
						+ MispSchema.CRLF);
			}
			osw.write("</" + Const.NS_XSL_PRE + ":template>" + MispSchema.CRLF);

			osw.write("</" + Const.NS_XSL_PRE + ":stylesheet>"
					+ MispSchema.CRLF);
			osw.flush();
			osw.close();

		} catch (Exception e) {
			if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
				e.printStackTrace();
			}
			setLastError(e.getMessage());
		}
		return rc;
	}

}
