package daruma.bus.processor.misp;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.apache.commons.codec.binary.Base64;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.XMLSerializer;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLFilter;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.AttributesImpl;
import org.xml.sax.helpers.XMLFilterImpl;

import application.startup.SystemProps;

import daruma.bus.message.Message;
import daruma.bus.message.StandardPayload;
import daruma.bus.processor.Processor;
import daruma.bus.processor.ProcessorException;
import daruma.bus.processor.csv.Const;
import daruma.bus.processor.csv.MispSchema;
import daruma.bus.processor.csv.MispSchemaAccess;
import daruma.bus.processor.csv.MispSchemaCx;
import daruma.bus.processor.csv.MispSchemaUtil;
import daruma.bus.processor.csv.MispSchemaUtil.MispSchemaXpath;

public class BuiltinFilter implements Processor {
	/**
	 * ログ
	 */
	static private Log log = LogFactory.getLog(BuiltinFilter.class);

	/**
	 * CSV->XMLモード（S式をXMLへ変換)
	 */
	final public static int MODE_CSV2XML = 0;

	/**
	 * XML->CSVモード（XMLをS式へ変換)
	 */
	final public static int MODE_XML2CSV = 1;

	/**
	 * GML名前空間
	 */

	final private static String EXP_SYMBOL_POINT = "POINT";

	final private static String EXP_SYMBOL_LINESTRING = "LINESTRING";

	final private static String EXP_SYMBOL_POLYGON = "POLYGON";

	final private char EXP_COORD_SEP = ' ';

	final private char EXP_POINT_SEP = ',';

	final private char EXP_LINEAR_RING_SEP = ',';

	final private static String XML_TAG_POINT = "Point";

	final private static String XML_TAG_MULTI_POINT = "MultiPoint";

	final private static String XML_TAG_POINT_MEMBER = "pointMember";

	final private static String XML_TAG_POINT_COORD = "coordinates";

	final private static String XML_TAG_POINT_COORD2 = "coord";

	final private static String XML_TAG_LINESTRING = "LineString";

	final private static String XML_TAG_LINESTRING_COORD = "coordinates";

	final private static String XML_TAG_MULTI_LINESTRING = "MultiLineString";

	final private static String XML_TAG_LINESTRING_MEMBER = "lineStringMember";

	final private static String XML_TAG_POLYGON = "Polygon";

	final private static String XML_TAG_POLYGON_OUTER_BOUNDARY = "outerBoundaryIs";

	final private static String XML_TAG_POLYGON_INNER_BOUNDARY = "innerBoundaryIs";

	final private static String XML_TAG_POLYGON_LINEAR_RING = "LinearRing";

	final private static String XML_TAG_POLYGON_COORD = "coordinates";

	final private static String XML_TAG_MULTI_POLYGON = "MultiPolygon";

	final private static String XML_TAG_POLYGON_MEMBER = "polygonMember";

	final private char XML_COORD_SEP = ',';

	final private char XML_POINT_SEP = ' ';

	private int filterMode;

	private String attatchDir = "";

	private MispSchemaCx schemaCx = null;

	private File workDir = new File(SystemProps.get(
			SystemProps.PROPS_SYS_WORK_PATH, "./work"));

	public BuiltinFilter(int mode) {
		this.filterMode = mode;
	}

	/**
	 * attatchDirを取得する
	 * 
	 * @return Returns the attatchDir.
	 */
	public String getAttatchDir() {
		return attatchDir;
	}

	/**
	 * attatchDirをセットする
	 * 
	 * @param attatchDir
	 *            The attatchDir to set.
	 */
	public void setAttatchDir(String attatchDir) {
		this.attatchDir = attatchDir;
	}

	/**
	 * schemaCxを取得する
	 * 
	 * @return Returns the schemaCx.
	 */
	public MispSchemaCx getSchemaCx() {
		return schemaCx;
	}

	/**
	 * schemaCxをセットする
	 * 
	 * @param schemaCx
	 *            The schemaCx to set.
	 */
	public void setSchemaCx(MispSchemaCx schemaCx) {
		this.schemaCx = schemaCx;
	}

	public void process(Message message) throws ProcessorException {
		if (message == null) {
			ProcessorException e = new BuiltinFilterException(
					"A message is null.");
			e.fillInStackTrace();
			throw e;
		}

		try {
			File workFile = File.createTempFile(BuiltinFilter.class.getName(),
					".xml", workDir);
			OutputStream out = new BufferedOutputStream(new FileOutputStream(
					workFile));
			InputStream in = new BufferedInputStream(message.getPayload()
					.getInputStream());
			transform(in, out);

			in.close();
			out.flush();
			out.close();

			StandardPayload payload = new StandardPayload();
			payload.attach(workFile);
			message.setPayload(payload);

		} catch (Exception e) {
			BuiltinFilterException ex;
			log.error(e.getMessage(), e);
			ex = new BuiltinFilterException(e.getMessage());
			ex.fillInStackTrace();
			throw ex;
		}
	}

	private List attatchFileElementList = new ArrayList();

	protected void transform(InputStream in, OutputStream out)
			throws ProcessorException {
		try {
			ContentHandler contentHandler = (ContentHandler) getSerializer(out);
			XMLReader parser = getXMLReader();
			XMLFilter filter;
			if (filterMode == MODE_XML2CSV)
				filter = new BuiltinFilterXml2Csv(parser);
			else
				filter = new BuiltinFilterCsv2Xml(parser);
			makeAttatchFileElementList();
			filter.setContentHandler(contentHandler);
			filter.parse(new InputSource(in));
		} catch (Exception e) {
			log.error(e.getMessage(), e);
			ProcessorException exception = new BuiltinFilterException(e
					.getMessage());
			exception.setStackTrace(e.getStackTrace());
			throw exception;
		}

	}

	private void makeAttatchFileElementList() {
		attatchFileElementList.clear();
		if (schemaCx != null) {
			MispSchema schema = schemaCx.getSchema();
			List xpathList = MispSchemaUtil.getAttatchFileXpathList(schema, "");
			String header = MispSchemaUtil.makeTagTreeText(Const.NS_SOAP_URI,
					"Envelope")
					+ Const.INTERNAL_TAG_SEP
					+ MispSchemaUtil.makeTagTreeText(Const.NS_SOAP_URI, "Body")
					+ Const.INTERNAL_TAG_SEP
					+ MispSchemaUtil.makeTagTreeText(Const.NS_MISP_URI,
							"GetFeatureResponse")
					+ Const.INTERNAL_TAG_SEP
					+ MispSchemaUtil.makeTagTreeText(Const.NS_MISP_URI,
							"FeatureCollection")
					+ Const.INTERNAL_TAG_SEP
					+ MispSchemaUtil.makeTagTreeText(Const.NS_GML_URI,
							"featureMember")
					+ Const.INTERNAL_TAG_SEP
					+ MispSchemaUtil.makeTagTreeText(schema
							.getTargetNamespace(), schema.getFeatureElement()
							.getName()) + Const.INTERNAL_TAG_SEP;

			for (int cnt = 0; cnt < xpathList.size(); cnt++) {
				MispSchemaXpath xpath = (MispSchemaXpath) xpathList.get(cnt);
				String tagTreeText = MispSchemaUtil.convertXpathToTagTreeText(
						schema, xpath.getXpath());
				attatchFileElementList.add(header + tagTreeText);
			}
		}
	}

	private static String trimCrLf(String val) {
		val = val.replaceAll("\r", "");
		val = val.replaceAll("\n", "");
		return val;
	}

	class BuiltinFilterCsv2Xml extends XMLFilterImpl {
		private int MODE_NONE = 0;

		private int MODE_SAME_NS_PREFIX = 1;

		private int MODE_SAME_PREFIX = 2;

		public BuiltinFilterCsv2Xml(XMLReader parser) {
			super(parser);
		}

		private String convertCoordCsv2Xml(String coord) {
			return coord.replace(EXP_COORD_SEP, XML_COORD_SEP);
		}

		private String convertCoordinatesCsv2Xml(String coordinates) {
			String[] coords = coordinates.split("" + EXP_POINT_SEP);
			coordinates = "";
			for (int cnt = 0; cnt < coords.length; cnt++) {
				if (cnt > 0)
					coordinates += XML_POINT_SEP;
				coordinates += convertCoordCsv2Xml(coords[cnt].trim());
			}
			return coordinates;
		}

		int currPrefixMode = MODE_NONE;

		String currPrefixName = "";

		public void startPrefixMapping(String prefix, String namespace) {
			if (currPrefixMode == MODE_NONE) {
				if (namespace.equals(Const.NS_GML_URI)) {
					if (prefix.equals(Const.NS_GML_PRE)) {
						currPrefixMode = MODE_SAME_NS_PREFIX;
					} else {
						currPrefixMode = MODE_SAME_PREFIX;
						currPrefixName = prefix;
					}
				}
			}
		}

		public void endPrefixMapping(String prefix) {
			if (currPrefixMode == MODE_SAME_NS_PREFIX) {
				if (prefix.equals(Const.NS_GML_PRE))
					currPrefixMode = MODE_NONE;
			} else if (currPrefixMode == MODE_SAME_PREFIX) {
				if (prefix.equals(currPrefixName))
					currPrefixMode = MODE_NONE;
			}
		}

		String currTagVal = "";

		String currTrimedTagVal = "";

		String currDfsType = "";

		Attributes makeAttributes(Attributes atts) {
			AttributesImpl newAtts = new AttributesImpl(atts);
			int index = newAtts.getIndex(Const.NS_MISP_URI,
					MispSchemaAccess.XSD_ATTR_ELEMENT_DFSTYPE);
			if (index >= 0)
				newAtts.removeAttribute(index);
			return newAtts;
		}

		private List currTagTree = new ArrayList();

		public void startElement(String namespaceURI, String localName,
				String qName, Attributes atts) throws SAXException {
			ContentHandler handler = this.getContentHandler();
			if (handler == null)
				return;

			currTagTree.add(MispSchemaUtil.makeTagTreeText(namespaceURI,
					localName));

			currTagVal = "";
			currTrimedTagVal = "";
			currDfsType = atts.getValue(Const.NS_MISP_URI,
					MispSchemaAccess.XSD_ATTR_ELEMENT_DFSTYPE);
			if (currDfsType == null)
				currDfsType = "";
			Attributes newAtts = makeAttributes(atts);

			if (qName.equals(Const.NS_TNS_PRE + ":" + localName))
				handler.startElement(namespaceURI, localName, localName,
						newAtts);
			else
				handler.startElement(namespaceURI, localName, qName, newAtts);
		}

		public void characters(char[] ch, int start, int length)
				throws SAXException {
			String text = new String(ch, start, length);
			currTagVal += text;
			currTrimedTagVal += trimCrLf(text);
		}

		public void endElement(String namespaceURI, String localName,
				String qName) throws SAXException {
			boolean match = false;

			ContentHandler handler = this.getContentHandler();
			if (handler == null)
				return;

			if (currDfsType.equals(Const.DFS_TYPE_ATTATCHEDFILE)) {
				match = true;
				currTrimedTagVal = currTrimedTagVal.trim();
				if (currTrimedTagVal.length() > 0) {
					log.debug("Load attatch file from " + currTrimedTagVal);
					encodeAttatchFile(handler, currTrimedTagVal);
				}
			} else if ((currDfsType.length() == 0)
					|| currDfsType.equals(Const.DFS_TYPE_GEOMETRYPROPERTYTYPE)) {
				Pattern pattern;
				Matcher matcher;
				currTrimedTagVal = currTrimedTagVal.trim();
				pattern = Pattern.compile("^" + EXP_SYMBOL_POINT
						+ "\\((.*)\\)$", Pattern.CASE_INSENSITIVE);
				matcher = pattern.matcher(currTrimedTagVal);
				if (matcher.find()) {
					match = true;
					handlePoint(handler, "(" + matcher.group(1) + ")");
				}

				pattern = Pattern.compile("^" + EXP_SYMBOL_LINESTRING
						+ "\\((.*)\\)$", Pattern.CASE_INSENSITIVE);
				matcher = pattern.matcher(currTrimedTagVal);
				if (matcher.find()) {
					match = true;
					handleLineString(handler, "(" + matcher.group(1) + ")");
				}

				pattern = Pattern.compile("^" + EXP_SYMBOL_POLYGON
						+ "\\((.*)\\)$", Pattern.CASE_INSENSITIVE);
				matcher = pattern.matcher(currTrimedTagVal);
				if (matcher.find()) {
					match = true;
					handlePolygon(handler, "(" + matcher.group(1) + ")");
				}
			}

			if (!match) {
				if (currTagVal.length() > 0) {
					char[] ch = currTagVal.toCharArray();
					handler.characters(ch, 0, currTagVal.length());
				}
			}

			if (qName.equals(Const.NS_TNS_PRE + ":" + localName))
				handler.endElement(namespaceURI, localName, localName);
			else
				handler.endElement(namespaceURI, localName, qName);

			if (currTagTree.size() > 0)
				currTagTree.remove(currTagTree.size() - 1);
			currTagVal = "";
			currTrimedTagVal = "";
		}

		private int encodeAttatchFile(ContentHandler handler, String file) {
			int rc = -1;
			FileInputStream fis = null;
			if (!MispSchema.isAbsolutePath(file)) {
				file = MispSchema.makeFullPath(attatchDir, file);
			}
			try {
				fis = new FileInputStream(file);
				int ENCODE_LEN = 3;
				byte[] readBuf = new byte[ENCODE_LEN * 1024];
				byte[] remainBuf = new byte[ENCODE_LEN];
				int currReadLen;
				int currRemainLen = 0;
				int cntTotalReadLen = 0;
				int cntTotalWriteLen = 0;
				while (true) {
					currReadLen = fis.read(readBuf);
					if (currReadLen < 0)
						currReadLen = 0;
					cntTotalReadLen += currReadLen;

					byte[] tempBuf;
					byte[] base64Buf;
					if ((currRemainLen > 0)
							|| (readBuf.length != currReadLen)
							|| (((currRemainLen + currReadLen)) % ENCODE_LEN != 0)) {
						int currEncodeLen = currRemainLen + currReadLen;
						if (currReadLen > 0)
							currEncodeLen = (currEncodeLen / ENCODE_LEN)
									* ENCODE_LEN;
						tempBuf = new byte[currEncodeLen];

						int currLen1 = (currRemainLen > currEncodeLen) ? currEncodeLen
								: currRemainLen;
						int currLen2 = currEncodeLen - currLen1;
						/**
						 * Encodeバッファを作成
						 */
						for (int cnti = 0; cnti < currLen1; cnti++)
							tempBuf[cnti] = remainBuf[cnti];
						for (int cnti = 0; cnti < currLen2; cnti++)
							tempBuf[currLen1 + cnti] = readBuf[cnti];

						/**
						 * 余ったバイトを保存
						 */
						currLen1 = currRemainLen - currLen1;
						currRemainLen = (currRemainLen + currReadLen)
								- currEncodeLen;
						for (int cntj = currLen1; cntj < currRemainLen; cntj++) {
							remainBuf[cntj] = readBuf[currLen2 + cntj
									- currLen1];
						}

					} else {
						tempBuf = readBuf;
					}
					if (tempBuf.length > 0) {
						base64Buf = Base64.encodeBase64(tempBuf);
						char[] chrData = new String(base64Buf).toCharArray();
						handler.characters(chrData, 0, chrData.length);
						cntTotalWriteLen += chrData.length;
					}

					if (currReadLen <= 0)
						break;
				}
				if (cntTotalReadLen == 0) {
					char[] chrData = { '=', '=', '=', '=' };
					handler.characters(chrData, 0, chrData.length);
					cntTotalWriteLen += chrData.length;
				}
				rc = cntTotalWriteLen;
			} catch (FileNotFoundException e) {
				log.error("Fail to open file:" + file);
			} catch (Exception e) {
				log.error("Fail to encode attatch file:" + e.getMessage());
			}
			if (fis != null)
				try {
					fis.close();
				} catch (IOException e) {
					if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
						e.printStackTrace();
					}
				}
			return rc;
		}

		private void handlePoint(ContentHandler handler, String data)
				throws SAXException {
			log.debug("Match EXP SYMBOL POINT:" + data);

			String[] datas = data.toUpperCase().split(EXP_SYMBOL_POINT);
			if (datas.length > 1) {
				handler.startElement(Const.NS_GML_URI, XML_TAG_MULTI_POINT, "",
						new AttributesImpl());
				handler.startElement(Const.NS_GML_URI, XML_TAG_POINT_MEMBER,
						"", new AttributesImpl());
			}
			for (int cnt = 0; cnt < datas.length; cnt++) {
				data = datas[cnt].trim();
				Pattern pattern = Pattern.compile("^\\((.*)\\)$",
						Pattern.CASE_INSENSITIVE);
				Matcher matcher = pattern.matcher(data);
				if (!matcher.find())
					continue;
				data = matcher.group(1);

				if (currPrefixMode == MODE_NONE)
					handler.startPrefixMapping(Const.NS_GML_PRE,
							Const.NS_GML_URI);
				handler.startElement(Const.NS_GML_URI, XML_TAG_POINT, "",
						new AttributesImpl());
				handler.startElement(Const.NS_GML_URI, XML_TAG_POINT_COORD, "",
						new AttributesImpl());
				data = convertCoordCsv2Xml(data);
				char[] chrData = data.toCharArray();
				handler.characters(chrData, 0, chrData.length);
				handler.endElement(Const.NS_GML_URI, XML_TAG_POINT_COORD, "");
				handler.endElement(Const.NS_GML_URI, XML_TAG_POINT, "");
				if (currPrefixMode == MODE_NONE)
					handler.endPrefixMapping(Const.NS_GML_PRE);
			}
			if (datas.length > 1) {
				handler.endElement(Const.NS_GML_URI, XML_TAG_POINT_MEMBER, "");
				handler.endElement(Const.NS_GML_URI, XML_TAG_MULTI_POINT, "");
			}
		}

		private void handleLineString(ContentHandler handler, String data)
				throws SAXException {
			log.debug("Match EXP SYMBOL LINESTRING:" + data);

			String[] datas = data.toUpperCase().split(EXP_SYMBOL_LINESTRING);
			if (datas.length > 1) {
				handler.startElement(Const.NS_GML_URI,
						XML_TAG_MULTI_LINESTRING, "", new AttributesImpl());
				handler.startElement(Const.NS_GML_URI,
						XML_TAG_LINESTRING_MEMBER, "", new AttributesImpl());
			}
			for (int cnt = 0; cnt < datas.length; cnt++) {
				data = datas[cnt].trim();
				Pattern pattern = Pattern.compile("^\\((.*)\\)$",
						Pattern.CASE_INSENSITIVE);
				Matcher matcher = pattern.matcher(data);
				if (!matcher.find())
					continue;
				data = matcher.group(1);

				if (currPrefixMode == MODE_NONE)
					handler.startPrefixMapping(Const.NS_GML_PRE,
							Const.NS_GML_URI);
				handler.startElement(Const.NS_GML_URI, XML_TAG_LINESTRING, "",
						new AttributesImpl());
				handler.startElement(Const.NS_GML_URI,
						XML_TAG_LINESTRING_COORD, "", new AttributesImpl());
				data = convertCoordinatesCsv2Xml(data);
				char[] chrData = data.toCharArray();
				handler.characters(chrData, 0, chrData.length);
				handler.endElement(Const.NS_GML_URI, XML_TAG_LINESTRING_COORD,
						"");
				handler.endElement(Const.NS_GML_URI, XML_TAG_LINESTRING, "");
				if (currPrefixMode == MODE_NONE)
					handler.endPrefixMapping(Const.NS_GML_PRE);
			}
			if (datas.length > 1) {
				handler.endElement(Const.NS_GML_URI, XML_TAG_LINESTRING_MEMBER,
						"");
				handler.endElement(Const.NS_GML_URI, XML_TAG_MULTI_LINESTRING,
						"");

			}
		}

		private void handlePolygon(ContentHandler handler, String data)
				throws SAXException {
			log.debug("Match EXP SYMBOL POLYGON:" + data);

			String[] datas = data.toUpperCase().split(EXP_SYMBOL_POLYGON);
			if (datas.length > 1) {
				handler.startElement(Const.NS_GML_URI, XML_TAG_MULTI_POLYGON,
						"", new AttributesImpl());
				handler.startElement(Const.NS_GML_URI, XML_TAG_POLYGON_MEMBER,
						"", new AttributesImpl());
			}
			for (int cnt = 0; cnt < datas.length; cnt++) {
				data = datas[cnt].trim();
				Pattern pattern = Pattern.compile("^\\((.*)\\)$",
						Pattern.CASE_INSENSITIVE);
				Matcher matcher = pattern.matcher(data);
				if (!matcher.find())
					continue;
				data = matcher.group(1);

				if (currPrefixMode == MODE_NONE)
					handler.startPrefixMapping(Const.NS_GML_PRE,
							Const.NS_GML_URI);
				handler.startElement(Const.NS_GML_URI, XML_TAG_POLYGON, "",
						new AttributesImpl());
				pattern = Pattern.compile("\\)" + EXP_LINEAR_RING_SEP + "*",
						Pattern.CASE_INSENSITIVE);
				String[] boundarys = pattern.split(data);
				for (int cntj = 0; cntj < boundarys.length; cntj++) {
					data = boundarys[cntj].trim();
					if (data.length() < 1)
						continue;
					data = data.substring(1);
					data = convertCoordinatesCsv2Xml(data);
					String boundary = (cntj == 0) ? XML_TAG_POLYGON_OUTER_BOUNDARY
							: XML_TAG_POLYGON_INNER_BOUNDARY;
					handler.startElement(Const.NS_GML_URI, boundary, "",
							new AttributesImpl());
					handler.startElement(Const.NS_GML_URI,
							XML_TAG_POLYGON_LINEAR_RING, "",
							new AttributesImpl());
					handler.startElement(Const.NS_GML_URI,
							XML_TAG_POLYGON_COORD, "", new AttributesImpl());
					char[] chrData = data.toCharArray();
					handler.characters(chrData, 0, chrData.length);
					handler.endElement(Const.NS_GML_URI, XML_TAG_POLYGON_COORD,
							"");
					handler.endElement(Const.NS_GML_URI,
							XML_TAG_POLYGON_LINEAR_RING, "");
					handler.endElement(Const.NS_GML_URI, boundary, "");
				}
				handler.endElement(Const.NS_GML_URI, XML_TAG_POLYGON, "");
				if (currPrefixMode == MODE_NONE)
					handler.endPrefixMapping(Const.NS_GML_PRE);
			}
			if (datas.length > 1) {
				handler
						.endElement(Const.NS_GML_URI, XML_TAG_POLYGON_MEMBER,
								"");
				handler.endElement(Const.NS_GML_URI, XML_TAG_MULTI_POLYGON, "");
			}
		}
	}

	class BuiltinFilterXml2Csv extends XMLFilterImpl {
		public BuiltinFilterXml2Csv(XMLReader parser) {
			super(parser);
		}

		private String convertCoordXml2Csv(String coord) {
			return coord.replace(XML_COORD_SEP, EXP_COORD_SEP);
		}

		private String convertCoordinatesXml2Csv(String coordinates) {
			String[] coords = coordinates.split("" + XML_POINT_SEP);
			coordinates = "";
			for (int cnt = 0; cnt < coords.length; cnt++) {
				if (cnt > 0)
					coordinates += "" + EXP_POINT_SEP;
				coordinates += convertCoordXml2Csv(coords[cnt].trim());
			}
			return coordinates;
		}

		String currTagVal = "";

		String currTrimedTagVal = "";

		private int currTagLevel = 0;

		private String currExpValue = "";

		private String[] tblTag = new String[8];

		private List currTagTree = new ArrayList();

		private File tmpFile;

		private FileWriter tmpFos;

		private String currTagTreeText = "";

		private long cntExpNum = 0;

		public void startElement(String namespaceURI, String localName,
				String qName, Attributes atts) throws SAXException {

			ContentHandler handler = this.getContentHandler();
			if (handler == null)
				return;

			currTagTree.add(namespaceURI + Const.INTERNAL_NAMESPACE_SEP
					+ localName);
			currTagTreeText = MispSchemaUtil.makeTagTreeText(currTagTree);
			if (attatchFileElementList.contains(currTagTreeText)) {
				try {
					tmpFile = File.createTempFile(
							BuiltinFilter.class.getName(), ".tmp", workDir);
					tmpFos = new FileWriter(tmpFile);
				} catch (Exception e) {
					log.error("createTempFile fail:" + e.getMessage());
				}
			}

			if (currTagLevel > 0) {
				if (currTagLevel < tblTag.length)
					tblTag[currTagLevel] = (namespaceURI
							.equals(Const.NS_GML_URI)) ? localName : "";
				currTagLevel++;
			} else {
				currExpValue = "";
				if (namespaceURI.equals(Const.NS_GML_URI)) {
					if (localName.equals(XML_TAG_POINT)) {
						log.debug("Match XML TAG POINT");
						currExpValue = EXP_SYMBOL_POINT + "(";
					} else if (localName.equals(XML_TAG_LINESTRING)) {
						log.debug("Match XML TAG LINESTRING");
						currExpValue = EXP_SYMBOL_LINESTRING + "(";
					} else if (localName.equals(XML_TAG_POLYGON)) {
						log.debug("Match XML TAG POLYGON");
						currExpValue = EXP_SYMBOL_POLYGON + "(";
					}
				}
				if (currExpValue.length() > 0) {
					tblTag[0] = localName;
					currTagLevel = 1;
					cntExpNum = 0;
				} else {
					if (namespaceURI.equals(Const.NS_GML_URI)
							&& (localName.equals(XML_TAG_MULTI_POINT)
									|| localName.equals(XML_TAG_POINT_MEMBER)
									|| localName
											.equals(XML_TAG_MULTI_LINESTRING)
									|| localName
											.equals(XML_TAG_LINESTRING_MEMBER)
									|| localName.equals(XML_TAG_MULTI_POLYGON) || localName
									.equals(XML_TAG_POLYGON_MEMBER))) {
					} else
						handler.startElement(namespaceURI, localName, qName,
								atts);
				}
			}
			currTagVal = "";
			currTrimedTagVal = "";
		}

		public void characters(char[] ch, int start, int length)
				throws SAXException {
			if (attatchFileElementList.contains(currTagTreeText)) {
				try {
					tmpFos.write(ch, start, length);
				} catch (IOException e) {
					log.error("FileWriter write fail:" + e.getMessage());
				}
			} else {
				String text = new String(ch, start, length);
				currTagVal += text;
				currTrimedTagVal += trimCrLf(text);
			}
		}

		public void endElement(String namespaceURI, String localName,
				String qName) throws SAXException {
			ContentHandler handler = this.getContentHandler();
			if (handler == null)
				return;
			if (attatchFileElementList.contains(currTagTreeText)) {
				File tempDir = new File((attatchDir.length() == 0) ? "./"
						: attatchDir);
				tempDir.mkdirs();
				try {
					tmpFos.flush();
					tmpFos.close();
					File attatchFile = File.createTempFile("attach", ".bin",
							tempDir);
					int len = decodeAttatchFile(tmpFile, attatchFile);
					if (len >= 0) {
						log.debug("Save attatch file to "
								+ attatchFile.getAbsolutePath());
						char[] ch = attatchFile.getName().toCharArray();
						handler.characters(ch, 0, ch.length);
					} else {
						attatchFile.delete();
					}
					tmpFile.delete();
				} catch (Exception e) {
					log.error("createTempFile fail:" + e.getMessage());
				}
				handler.endElement(namespaceURI, localName, qName);
			} else {
				if (currTagLevel > 1) {
					if (currTrimedTagVal.length() > 0) {
						if (tblTag[0].equals(XML_TAG_POINT)
								&& (tblTag[1].equals(XML_TAG_POINT_COORD) || tblTag[1]
										.equals(XML_TAG_POINT_COORD2))) {
							currExpValue = EXP_SYMBOL_POINT + "("
									+ convertCoordXml2Csv(currTrimedTagVal);
							cntExpNum++;
						} else if (tblTag[0].equals(XML_TAG_LINESTRING)
								&& tblTag[1].equals(XML_TAG_LINESTRING_COORD)) {
							currExpValue = EXP_SYMBOL_LINESTRING
									+ "("
									+ convertCoordinatesXml2Csv(currTrimedTagVal);
							cntExpNum++;
						} else if (tblTag[0].equals(XML_TAG_POLYGON)
								&& (tblTag[1]
										.equals(XML_TAG_POLYGON_OUTER_BOUNDARY) || tblTag[1]
										.equals(XML_TAG_POLYGON_INNER_BOUNDARY))
								&& tblTag[2]
										.equals(XML_TAG_POLYGON_LINEAR_RING)
								&& tblTag[3].equals(XML_TAG_POLYGON_COORD)) {
							if (cntExpNum > 0)
								currExpValue += EXP_LINEAR_RING_SEP;
							currExpValue += "("
									+ convertCoordinatesXml2Csv(currTrimedTagVal)
									+ ")";
							cntExpNum++;
						}
					}
				} else {
					if (currTagVal.length() > 0) {
						char[] ch = currTagVal.toCharArray();
						handler.characters(ch, 0, currTagVal.length());
					}
				}

				if (currTagLevel > 0) {
					currTagLevel--;
					if (currTagLevel < tblTag.length)
						tblTag[currTagLevel] = "";
					if (currTagLevel == 0) {
						currExpValue += ")";
						char[] chrData = currExpValue.toCharArray();
						handler.characters(chrData, 0, chrData.length);
					}
				} else {
					if (namespaceURI.equals(Const.NS_GML_URI)
							&& (localName.equals(XML_TAG_MULTI_POINT)
									|| localName.equals(XML_TAG_POINT_MEMBER)
									|| localName
											.equals(XML_TAG_MULTI_LINESTRING)
									|| localName
											.equals(XML_TAG_LINESTRING_MEMBER)
									|| localName.equals(XML_TAG_MULTI_POLYGON) || localName
									.equals(XML_TAG_POLYGON_MEMBER))) {
					} else
						handler.endElement(namespaceURI, localName, qName);
				}
			}

			if (currTagTree.size() > 0) {
				currTagTree.remove(currTagTree.size() - 1);
				currTagTreeText = MispSchemaUtil.makeTagTreeText(currTagTree);
			}
			currTagVal = "";
			currTrimedTagVal = "";
		}

		private int decodeAttatchFile(File tmpFile, File attatchFile) {
			int rc = -1;
			FileInputStream fis = null;
			FileOutputStream fos = null;
			try {
				fis = new FileInputStream(tmpFile);
				fos = new FileOutputStream(attatchFile);
				int ENCODE_LEN = 4;
				byte[] readBuf = new byte[ENCODE_LEN * 1024];
				byte[] remainBuf = new byte[ENCODE_LEN];
				int currReadLen;
				int currRemainLen = 0;
				int cntTotalReadLen = 0;
				int cntTotalWriteLen = 0;
				while (true) {
					currReadLen = fis.read(readBuf);
					if (currReadLen < 0)
						currReadLen = 0;
					cntTotalReadLen += currReadLen;

					byte[] tempBuf;
					byte[] base64Buf;
					if ((currRemainLen > 0)
							|| (readBuf.length != currReadLen)
							|| (((currRemainLen + currReadLen)) % ENCODE_LEN != 0)) {
						int currEncodeLen = currRemainLen + currReadLen;
						if (currReadLen > 0)
							currEncodeLen = (currEncodeLen / ENCODE_LEN)
									* ENCODE_LEN;
						tempBuf = new byte[currEncodeLen];

						int currLen1 = (currRemainLen > currEncodeLen) ? currEncodeLen
								: currRemainLen;
						int currLen2 = currEncodeLen - currLen1;
						/**
						 * Decodeバッファを作成
						 */
						for (int cnti = 0; cnti < currLen1; cnti++)
							tempBuf[cnti] = remainBuf[cnti];
						for (int cnti = 0; cnti < currLen2; cnti++)
							tempBuf[currLen1 + cnti] = readBuf[cnti];

						/**
						 * 余ったバイトを保存
						 */
						currLen1 = currRemainLen - currLen1;
						currRemainLen = (currRemainLen + currReadLen)
								- currEncodeLen;
						for (int cntj = currLen1; cntj < currRemainLen; cntj++) {
							remainBuf[cntj] = readBuf[currLen2 + cntj
									- currLen1];
						}

					} else {
						tempBuf = readBuf;
					}
					if (tempBuf.length > 0) {
						base64Buf = Base64.decodeBase64(tempBuf);
						fos.write(base64Buf);
						cntTotalWriteLen += base64Buf.length;
					}
					if (currReadLen <= 0)
						break;
				}
				rc = (cntTotalWriteLen == 0) ? ((cntTotalReadLen > 0) ? 0 : -1)
						: cntTotalWriteLen;
			} catch (FileNotFoundException e) {
				log.error("Fail to open file:" + tmpFile.getAbsolutePath()
						+ "/" + attatchFile.getAbsolutePath());
			} catch (Exception e) {
				if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
					e.printStackTrace();
				}
				log.error("Fail to decode attatch file:" + e.getMessage());
			}
			if (fis != null)
				try {
					fis.close();
				} catch (IOException e) {
					if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
						e.printStackTrace();
					}
				}
			if (fos != null)
				try {
					fos.flush();
					fos.close();
				} catch (IOException e) {
					if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
						e.printStackTrace();
					}
				}
			return rc;
		}
	}

	private XMLReader getXMLReader() throws ParserConfigurationException,
			SAXException {

		SAXParserFactory factory = SAXParserFactory.newInstance();
		factory.setFeature(Const.SAX_NS_URI, true);
		SAXParser parser = factory.newSAXParser();
		XMLReader xmlReader = parser.getXMLReader();
		xmlReader.setFeature(Const.SAX_NS_URI, true);
		return xmlReader;
	}

	private XMLSerializer getSerializer(OutputStream out) {
		OutputFormat formatter = new OutputFormat();
		formatter.setPreserveSpace(true);
		XMLSerializer serializer = new XMLSerializer(out, formatter);
		return serializer;
	}
}
