package daruma.bus.processor.misp;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;

import application.startup.SystemProps;

import daruma.bus.message.Message;
import daruma.bus.message.StandardPayload;
import daruma.bus.processor.Processor;
import daruma.bus.processor.ProcessorException;
import daruma.bus.processor.csv.Const;
import daruma.bus.processor.csv.MispSchema;
import daruma.bus.processor.csv.MispSchemaAccess;
import daruma.bus.processor.csv.MispSchemaUtil;
import daruma.bus.processor.csv.MispSchema.*;

public class FeatureTypeFilter implements Processor {

	static public final String PROPS_OUTPUT_ENCODING = "featuretypefilter.output.encoding";

	static public final String PROPS_SEPARATOR = "featuretypefilter.separator";

	static public final String DEFAULT_OUTPUT_ENCODING = SystemProps.get(
			PROPS_OUTPUT_ENCODING, System
					.getProperty(SystemProps.PROPS_SYS_OS_ENCODING));

	static public final String DEFAULT_SEPARATOR = SystemProps.get(
			PROPS_SEPARATOR, ":");

	private String outputEncoding = DEFAULT_OUTPUT_ENCODING;

	private String separator = DEFAULT_SEPARATOR;

	/**
	 * ログ
	 */
	static private Log log = LogFactory.getLog(FeatureTypeFilter.class);

	/**
	 * ターゲット名前空間
	 */
	private String filterTargetNameSpace = Const.NS_XSD_URI;

	/**
	 * ターゲットタグ名
	 */
	private String filterTargetTagName = "schema";

	/**
	 * ターゲットレベル
	 */
	private int filterTargetLevel = 0;

	private File workDir = new File(SystemProps.get(
			SystemProps.PROPS_SYS_WORK_PATH, "./work"));

	public FeatureTypeFilter() {
	}

	public void process(Message message) throws ProcessorException {
		if (message == null) {
			ProcessorException e = new FeatureTypeFilterException(
					"A message is null.");
			e.fillInStackTrace();
			throw e;
		}

		String outputEncoding = this.outputEncoding;

		if (message.getProperties().containsKey(PROPS_OUTPUT_ENCODING))
			outputEncoding = message.getProperties().getProperty(
					PROPS_OUTPUT_ENCODING);
		if (message.getProperties().containsKey(PROPS_SEPARATOR))
			this.separator = message.getProperties().getProperty(
					PROPS_SEPARATOR);

		try {
			File workFile = File.createTempFile(FeatureTypeFilter.class
					.getName(), ".xml", workDir);
			OutputStreamWriter out = new OutputStreamWriter(
					new FileOutputStream(workFile), outputEncoding);
			InputStream in = new BufferedInputStream(message.getPayload()
					.getInputStream());
			transform(in, out);
			in.close();
			out.flush();
			out.close();

			StandardPayload payload = new StandardPayload();
			payload.attach(workFile);
			message.setPayload(payload);

		} catch (Exception e) {
			log.error(e.getMessage(), e);
			FeatureTypeFilterException ex = new FeatureTypeFilterException(e
					.getMessage());
			ex.fillInStackTrace();
			throw ex;
		}
	}

	protected void transform(InputStream in, OutputStreamWriter out)
			throws ProcessorException {
		try {
			XMLReader parser = getXMLReader();
			GenericHandler handler = new GenericHandler(out);
			handler.setSeparator(this.separator);
			parser.setContentHandler(handler);
			parser.parse(new InputSource(in));
		} catch (Exception e) {
			if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
				e.printStackTrace();
			}
			log.debug(e.getMessage(), e);
			ProcessorException exception = new FeatureTypeFilterException(e
					.getMessage());
			exception.setStackTrace(e.getStackTrace());
			throw exception;
		}
	}

	class GenericHandler extends DefaultHandler {
		OutputStreamWriter out;

		MispSchema schema = new MispSchema();

		public GenericHandler(OutputStreamWriter out) {
			super();
			this.out = out;
		}

		private int currTotalTagLevel = -1;

		private int currTotalTagIndex = 0;

		private int currTargetTagLevel = -1;

		private int currTargetTagIndex = 0;

		private int currLevel = -1;

		private int currIndex = 0;

		private String separator = FeatureTypeFilter.DEFAULT_SEPARATOR;

		public void startElement(String namespaceURI, String localName,
				String qName, Attributes atts) throws SAXException {
			currTotalTagLevel++;

			if (currTargetTagLevel < 0) {
				if ((filterTargetNameSpace.equals("") || namespaceURI
						.equals(filterTargetNameSpace))
						&& (filterTargetTagName.equals("") || localName
								.equals(filterTargetTagName))) {
					currTargetTagLevel = 0;
					currTargetTagIndex = 0;
				}
			} else {
				currTargetTagLevel++;
			}

			if (currTargetTagLevel >= filterTargetLevel) {
				if (currTargetTagLevel == filterTargetLevel) {
					currLevel = 0;
					currIndex = 0;
				} else
					currLevel++;
				startElement2(namespaceURI, localName, qName, atts);
			} else {
				// outputStr("<" + qName + ">");
			}
		}

		public void characters(char[] ch, int start, int length)
				throws SAXException {
			currTotalTagIndex++;
			if (currTargetTagLevel >= 0) {
				currTargetTagIndex++;
				if (currTargetTagLevel >= filterTargetLevel) {
					currIndex++;
					characters2(ch, start, length);
				}
			} else {
				// outputChars(ch, start, length);
			}
		}

		public void endElement(String namespaceURI, String localName,
				String qName) throws SAXException {
			currTotalTagLevel--;

			if (currTargetTagLevel >= filterTargetLevel) {
				endElement2(namespaceURI, localName, qName);
				if (currLevel >= 0)
					currLevel--;
			} else {
				// outputStr("</" + qName + ">");
			}
			if (currTargetTagLevel >= 0)
				currTargetTagLevel--;
		}

		public void startPrefixMapping(String prefix, String uri) {
			if (uri.equals(Const.NS_XSD_URI) || uri.equals(Const.NS_XLINK_URI)) {

			} else {
				if (currLevel == XMLNS_LEVEL) {
					if (prefix.length() > 0) {
						MispSchemaNamespace ns = schema.addNamespace();
						ns.setPrefix(prefix);
						ns.setUri(uri);
						log.debug("Namespace:" + prefix + "/" + uri);
					}
				}
			}
		}

		private List currQNameTree = new ArrayList();

		private String currTagVal = "";

		private String getAttrVal(Attributes atts, String name) {
			String val = atts.getValue(name);
			if (val == null)
				val = "";
			return val;
		}

		private boolean inComplexType = false;

		private final int XMLNS_LEVEL = 0;

		private final int SCHEMA_LEVEL = 0;

		private final int FEATURETYPE_LEVEL = 1;

		private final int COMPLEXTYPE_LEVEL = 1;

		private final int IMPORT_LEVEL = 1;

		private final int EXTENSION_LEVEL = 3;

		private void startElement2(String namespaceURI, String localName,
				String qName, Attributes atts) throws SAXException {
			if (currLevel == 0) {
				currQNameTree.clear();
				inComplexType = false;
			}
			if (namespaceURI.equals(Const.NS_XSD_URI)) {
				if ((currLevel == SCHEMA_LEVEL)
						&& (localName.equals(MispSchemaAccess.XSD_TAG_SCHEMA))) {
					schema.setId(getAttrVal(atts,
							MispSchemaAccess.XSD_ATTR_SCHEMA_ID));
					schema.setTargetNamespace(getAttrVal(atts,
							MispSchemaAccess.XSD_ATTR_SCHEMA_TNS));
					log.debug("ID:" + schema.getId());
					log.debug("TargetNamespace" + schema.getTargetNamespace());
				} else if ((currLevel == EXTENSION_LEVEL)
						&& localName.equals(MispSchemaAccess.XSD_TAG_EXTENSION)) {
					if (inComplexType) {
						MispSchemaComplexType ct = (MispSchemaComplexType) schema
								.getComplexType(schema.getComplexTypeSize() - 1);
						ct.setExtension(getAttrVal(atts,
								MispSchemaAccess.XSD_ATTR_EXTENSION_BASE));
						log.debug("ElementType extension:" + ct.getExtension());
					}
				} else if ((currLevel == IMPORT_LEVEL)
						&& localName.equals(MispSchemaAccess.XSD_TAG_IMPORT)) {
					MispSchemaImport im = schema.addImport();
					im.setUri(getAttrVal(atts,
							MispSchemaAccess.XSD_ATTR_IMPORT_NAMESPACE));
					im.setLocation(getAttrVal(atts,
							MispSchemaAccess.XSD_ATTR_IMPORT_LOCATION));
					log.debug("Import:" + im.getUri() + "/" + im.getLocation());
				} else if ((currLevel == COMPLEXTYPE_LEVEL)
						&& localName
								.equals(MispSchemaAccess.XSD_TAG_COMPLEXTYPE)) {
					MispSchemaComplexType ct = schema.addComplexType(null);
					ct.setName(getAttrVal(atts,
							MispSchemaAccess.XSD_ATTR_COMPLEXTYPE_NAME));
					log.debug("ElementType:" + ct.getName());
					inComplexType = true;
				} else if (localName.equals(MispSchemaAccess.XSD_TAG_ELEMENT)) {
					if (currLevel == FEATURETYPE_LEVEL) {
						schema
								.getFeatureElement()
								.setName(
										getAttrVal(
												atts,
												MispSchemaAccess.XSD_ATTR_ELEMENT_NAME));
						schema
								.getFeatureElement()
								.setType(
										getAttrVal(
												atts,
												MispSchemaAccess.XSD_ATTR_ELEMENT_TYPE));
						log.debug("FeatureType:"
								+ schema.getFeatureElement().getName() + "/"
								+ schema.getFeatureElement().getType());
					} else if (inComplexType) {
						MispSchemaComplexType ct = (MispSchemaComplexType) schema
								.getComplexType(schema.getComplexTypeSize() - 1);
						MispSchemaElement el = ct.addElement(-1);
						el.setName(getAttrVal(atts,
								MispSchemaAccess.XSD_ATTR_ELEMENT_NAME));
						el.setType(getAttrVal(atts,
								MispSchemaAccess.XSD_ATTR_ELEMENT_TYPE));
						log.debug("ElementType element:" + el.getName() + "/"
								+ el.getType());
					}
				} else if (localName.equals(MispSchemaAccess.XSD_TAG_ATTR)) {
					if (inComplexType) {
						MispSchemaComplexType ct = (MispSchemaComplexType) schema
								.getComplexType(schema.getComplexTypeSize() - 1);
						MispSchemaAttribute attr = ct.addAttribute(-1);
						attr.setName(getAttrVal(atts,
								MispSchemaAccess.XSD_ATTR_ATTR_NAME));
						attr.setType(getAttrVal(atts,
								MispSchemaAccess.XSD_ATTR_ATTR_TYPE));

						log.debug("ElementType attribute:" + attr.getName()
								+ "/" + attr.getType());

					}
				}
			}
			currQNameTree.add(qName);
			currTagVal = "";
		}

		private void characters2(char[] ch, int start, int length)
				throws SAXException {
			currTagVal += new String(ch, start, length);
		}

		private void endElement2(String namespaceURI, String localName,
				String qName) throws SAXException {
			currTagVal = "";

			if (namespaceURI.equals(Const.NS_XSD_URI)) {
				if ((currLevel == COMPLEXTYPE_LEVEL)
						&& (localName
								.equals(MispSchemaAccess.XSD_TAG_COMPLEXTYPE))) {
					inComplexType = false;
				}
			}

			if (currLevel == SCHEMA_LEVEL) {
				MispSchemaUtil.dumpFeatureType(schema, out, this.separator);
			}

			if (currQNameTree.size() > 0)
				currQNameTree.remove(currQNameTree.size() - 1);

		}

		public String getSeparator() {
			return separator;
		}

		public void setSeparator(String separator) {
			this.separator = separator;
		}
	}

	private XMLReader getXMLReader() throws ParserConfigurationException,
			SAXException {
		SAXParserFactory factory = SAXParserFactory.newInstance();
		factory.setFeature(Const.SAX_NS_URI, true);
		SAXParser parser = factory.newSAXParser();
		XMLReader xmlReader = parser.getXMLReader();
		xmlReader.setFeature(Const.SAX_NS_URI, true);
		return xmlReader;
	}

	public String getOutputEncoding() {
		return outputEncoding;
	}

	public String getSeparator() {
		return separator;
	}

	public void setOutputEncoding(String outputEncoding) {
		this.outputEncoding = outputEncoding;
	}

	public void setSeparator(String separator) {
		this.separator = separator;
	}
}
