package daruma.bus.processor.misp;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.XMLSerializer;
import org.xml.sax.Attributes;
import org.xml.sax.ContentHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLFilter;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLFilterImpl;

import application.startup.SystemProps;

import daruma.bus.message.Message;
import daruma.bus.message.StandardPayload;
import daruma.bus.processor.Processor;
import daruma.bus.processor.ProcessorException;
import daruma.bus.processor.csv.Const;

/**
 * ターゲット要素の子要素をフィルタする
 * 
 * @author shumi
 * 
 */
public class GetSubXmlFilter implements Processor {
	/**
	 * ログ
	 */
	static private Log log = LogFactory.getLog(GetSubXmlFilter.class);

	/**
	 * ターゲット名前空間
	 */
	private String targetNameSpace = "";

	/**
	 * ターゲットタグ名
	 */
	private String targetTagName = "";

	/**
	 * 
	 */
	private boolean includeTargetTag = false;

	private File workDir = new File(SystemProps.get(
			SystemProps.PROPS_SYS_WORK_PATH, "./work"));

	public GetSubXmlFilter(String nameSpace, String tagName,
			boolean includeTargetTag) {
		this.targetNameSpace = nameSpace;
		this.targetTagName = tagName;
		this.includeTargetTag = includeTargetTag;
	}

	public void process(Message message) throws ProcessorException {
		if (message == null) {
			ProcessorException e = new BuiltinFilterException(
					"A message is null.");
			e.fillInStackTrace();
			throw e;
		}

		try {
			File workFile = File.createTempFile(
					GetSubXmlFilter.class.getName(), ".xml", workDir);
			OutputStream out = new BufferedOutputStream(new FileOutputStream(
					workFile));
			InputStream in = new BufferedInputStream(message.getPayload()
					.getInputStream());
			transform(in, out);
			in.close();
			out.flush();
			out.close();

			StandardPayload payload = new StandardPayload();
			payload.attach(workFile);
			message.setPayload(payload);

		} catch (Exception e) {
			log.error(e.getMessage(), e);
			BuiltinFilterException ex = new BuiltinFilterException(e
					.getMessage());
			ex.fillInStackTrace();
			throw ex;
		}
	}

	protected void transform(InputStream in, OutputStream out)
			throws ProcessorException {
		try {
			ContentHandler contentHandler = (ContentHandler) getSerializer(out);
			XMLReader parser = getXMLReader();
			XMLFilter filter;
			filter = new GetSubXmlFilterImpl(parser);
			filter.setContentHandler(contentHandler);
			filter.parse(new InputSource(in));
		} catch (Exception e) {
			if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
				e.printStackTrace();
			}
			log.error(e.getMessage(), e);
			ProcessorException exception = new BuiltinFilterException(e
					.getMessage());
			exception.setStackTrace(e.getStackTrace());
			throw exception;
		}

	}

	class GetSubXmlFilterImpl extends XMLFilterImpl {
		public GetSubXmlFilterImpl(XMLReader parser) {
			super(parser);
		}

		private int currTagLevel = -1;

		public void startElement(String namespaceURI, String localName,
				String qName, Attributes atts) throws SAXException {

			ContentHandler handler = this.getContentHandler();
			if (handler == null)
				return;

			if (currTagLevel >= 0) {
				handler.startElement(namespaceURI, localName, qName, atts);
				currTagLevel++;
			} else {
				if ((targetNameSpace.equals("") || namespaceURI
						.equals(targetNameSpace))
						&& (targetTagName.equals("") || localName
								.equals(targetTagName))) {
					log.debug("Match TARGET XML TAG " + targetTagName);
					currTagLevel = 0;
					if (includeTargetTag)
						handler.startElement(namespaceURI, localName, qName,
								atts);
				}
			}
		}

		public void characters(char[] ch, int start, int length)
				throws SAXException {
			ContentHandler handler = this.getContentHandler();
			if (handler == null)
				return;

			if (currTagLevel > 0) {
				handler.characters(ch, start, length);
			}
		}

		public void endElement(String namespaceURI, String localName,
				String qName) throws SAXException {
			ContentHandler handler = this.getContentHandler();
			if (handler == null)
				return;

			if (currTagLevel > 0) {
				handler.endElement(namespaceURI, localName, qName);
				currTagLevel--;
			} else if (currTagLevel == 0) {
				if (includeTargetTag)
					handler.endElement(namespaceURI, localName, qName);
				currTagLevel--;
			}

		}
	}

	private XMLReader getXMLReader() throws ParserConfigurationException,
			SAXException {

		SAXParserFactory factory = SAXParserFactory.newInstance();
		factory.setFeature(Const.SAX_NS_URI, true);
		SAXParser parser = factory.newSAXParser();
		XMLReader xmlReader = parser.getXMLReader();
		xmlReader.setFeature(Const.SAX_NS_URI, true);
		return xmlReader;
	}

	private XMLSerializer getSerializer(OutputStream out) {
		OutputFormat formatter = new OutputFormat();
		formatter.setPreserveSpace(true);
		XMLSerializer serializer = new XMLSerializer(out, formatter);
		return serializer;
	}

}
