package daruma.bus.processor.misp;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.channels.SocketChannel;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import application.startup.SystemProps;

import daruma.bus.message.StandardPayload;
import daruma.bus.processor.Processor;
import daruma.bus.processor.ProcessorException;
import daruma.bus.processor.xslt.XslTransformer;

public class MispSoapClient implements Processor {

	static private Log log = LogFactory.getLog(MispSoapClient.class);

	static public final String PROPS_SERVER_HOST = "mispsoapclient.server.host";

	static public final String PROPS_SERVER_PORT = "mispsoapclient.server.port";

	static public final String PROPS_XSLT_CREATE_SOAP_TEMPLATE = "mispsoapclient.xslt.create.soap.template";

	static public final String PROPS_MESSAGE_TYPE = "mispsoapclient.message.type";

	static public final String PROPS_SOAP_RESPONSE_SUFFIX = "mispsoapclient.soap.response.suffix";

	static public final String DEFAULT_SERVER_HOST = SystemProps.get(
			PROPS_SERVER_HOST, "localhost");

	static public final String DEFAULT_SERVER_PORT = SystemProps.get(
			PROPS_SERVER_PORT, "5050");

	static public final String DEFAULT_XSLT_CREATE_SOAP_TEMPLATE = SystemProps
			.get(PROPS_XSLT_CREATE_SOAP_TEMPLATE, "");

	static public final String DEFAULT_SOAP_RESPONSE_SUFFIX = SystemProps.get(
			PROPS_SOAP_RESPONSE_SUFFIX, "</SOAP-ENV:Envelope>");

	// static public final String MSGTYPE_SOAP_DIRECT =
	// "mispsoapclient.message.soap.direct";
	// static public final String MSGTYPE_GET_FEATURE =
	// "mispsoapclient.message.get.feature";
	// static public final String MSGTYPE_TRANSACTION_INSERT =
	// "mispsoapclient.message.transaction.insert";
	// static public final String MSGTYPE_TRANSACTION_UPDATE =
	// "mispsoapclient.message.transaction.update";
	// static public final String MSGTYPE_TRANSACTION_DELETE =
	// "mispsoapclient.message.transaction.delete";

	private String host = DEFAULT_SERVER_HOST;

	private int portNumber = Integer.valueOf(DEFAULT_SERVER_PORT).intValue();

	private String xsltCreateSoapTemplate = DEFAULT_XSLT_CREATE_SOAP_TEMPLATE;

	private File workDir = new File(SystemProps.get(
			SystemProps.PROPS_SYS_WORK_PATH, "./work"));

	private String soapResponseSuffix = DEFAULT_SOAP_RESPONSE_SUFFIX;

	private boolean addSoapHeader = true;

	protected File xml2soapRequest(InputStream in, String xsltTemplate)
			throws ProcessorException {

		try {
			File file = File.createTempFile(MispSoapClient.class.getName(),
					".xml", this.workDir);
			BufferedInputStream bufferIn = new BufferedInputStream(in);
			BufferedOutputStream bufferOut = new BufferedOutputStream(
					new FileOutputStream(file));
			if (addSoapHeader) {
				XslTransformer transformer = new XslTransformer();
				transformer.setXsltFile(xsltTemplate);
				transformer.process(bufferIn, bufferOut);
			} else {
				int len;
				byte[] buf = new byte[8192];
				while ((len = bufferIn.read(buf)) != -1) {
					bufferOut.write(buf, 0, len);
				}
			}
			bufferOut.flush();
			bufferOut.close();
			bufferIn.close();
			return file;
		} catch (Exception e) {
			log.debug(e.getMessage(), e);
			ProcessorException ex = new MispSoapException(e.getMessage());
			ex.fillInStackTrace();
			throw ex;
		}

	}

	public void process(daruma.bus.message.Message message)
			throws ProcessorException {

		try {

			String xsltCreateSoapTemplate = this.xsltCreateSoapTemplate;
			String host = this.host;
			int portNumber = this.portNumber;
			String soapResponseSuffix = this.soapResponseSuffix;

			if (message.getProperties().containsKey(
					PROPS_XSLT_CREATE_SOAP_TEMPLATE))
				xsltCreateSoapTemplate = message.getProperties().getProperty(
						PROPS_XSLT_CREATE_SOAP_TEMPLATE);
			if (message.getProperties().containsKey(PROPS_SERVER_HOST))
				host = message.getProperties().getProperty(PROPS_SERVER_HOST);
			if (message.getProperties().containsKey(PROPS_SERVER_PORT))
				portNumber = Integer.valueOf(
						message.getProperties().getProperty(PROPS_SERVER_PORT))
						.intValue();
			if (message.getProperties().containsKey(PROPS_SOAP_RESPONSE_SUFFIX))
				soapResponseSuffix = message.getProperties().getProperty(
						PROPS_SOAP_RESPONSE_SUFFIX);

			File requestFile = xml2soapRequest(message.getPayload()
					.getInputStream(), xsltCreateSoapTemplate);

			if (requestFile != null) {
				File responseFile = File.createTempFile(MispSoapClient.class
						.getName(), ".xml", this.workDir);

				FileInputStream requestIn = new FileInputStream(requestFile);
				FileOutputStream responseOut = new FileOutputStream(
						responseFile);
				sendRequest(host, portNumber, requestIn, responseOut,
						soapResponseSuffix);

				requestIn.close();
				responseOut.close();
				if (!message.isVerboseMode())
					requestFile.delete();

				StandardPayload payload = new StandardPayload();
				payload.attach(responseFile);
				message.setPayload(payload);

			} else {
				ProcessorException e = new MispSoapException(
						"MessageType is unknown.");
				e.fillInStackTrace();
				throw e;
			}

		} catch (Exception e) {
			log.debug(e.getMessage(), e);
			ProcessorException ex = new MispSoapException(e.getMessage());
			ex.fillInStackTrace();
			throw ex;
		}
	}

	static protected void sendRequest(String host, int portNumber,
			FileInputStream request, FileOutputStream response,
			String responseEndElement) throws ProcessorException {

		try {
			SocketAddress address = new InetSocketAddress(host, portNumber);
			SocketChannel serverSc = SocketChannel.open(address);
			serverSc.configureBlocking(true);

			ByteBuffer buffer = ByteBuffer.allocate(8192);

			FileChannel requestFc = request.getChannel();
			FileChannel responseFc = response.getChannel();

			while ((requestFc.read(buffer)) != -1) {
				buffer.flip();
				serverSc.write(buffer);
				buffer.clear();
			}
			requestFc.close();
			buffer.clear();

			while ((serverSc.read(buffer)) > 0) {
				buffer.flip();
				responseFc.write(buffer);
				buffer.clear();
			}

			responseFc.close();
			serverSc.finishConnect();
			serverSc.close();

		} catch (Exception e) {
			log.debug(e.getMessage(), e);
			ProcessorException ex = new MispSoapException(e.getMessage());
			ex.fillInStackTrace();
			throw ex;
		}

	}

	public File getWorkDir() {
		return workDir;
	}

	public void setWorkDir(File workDir) {
		this.workDir = workDir;
	}

	public String getHost() {
		return host;
	}

	public void setHost(String host) {
		this.host = host;
	}

	public int getPortNumber() {
		return portNumber;
	}

	public void setPortNumber(int portNumber) {
		this.portNumber = portNumber;
	}

	/**
	 * addSoapHeaderを取得する
	 * 
	 * @return Returns the addSoapHeader.
	 */
	public boolean isAddSoapHeader() {
		return addSoapHeader;
	}

	/**
	 * addSoapHeaderをセットする
	 * 
	 * @param addSoapHeader
	 *            The addSoapHeader to set.
	 */
	public void setAddSoapHeader(boolean addSoapHeader) {
		this.addSoapHeader = addSoapHeader;
	}

}
