package daruma.bus.processor.misp;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.List;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.DefaultHandler;

import application.startup.SystemProps;

import daruma.bus.message.Message;
import daruma.bus.message.StandardPayload;
import daruma.bus.processor.Processor;
import daruma.bus.processor.ProcessorException;
import daruma.bus.processor.csv.Const;

public class XpathFilter implements Processor {
    
    static public final String PROPS_OUTPUT_ENCODING = "xpathfilter.output.encoding";
    static public final String PROPS_SEPARATOR = "xpathfilter.separator";
    static public final String DEFAULT_OUTPUT_ENCODING = SystemProps.get(PROPS_OUTPUT_ENCODING, System
            .getProperty(SystemProps.PROPS_SYS_OS_ENCODING));
    static public final String DEFAULT_SEPARATOR = SystemProps.get(PROPS_SEPARATOR, ":");
    
    private String outputEncoding = DEFAULT_OUTPUT_ENCODING;
    private String separator = DEFAULT_SEPARATOR;
    
    
    /**
     * ログ
     */
    static private Log log = LogFactory.getLog(XpathFilter.class);

    /**
     * ターゲット名前空間
     */
    private String filterTargetNameSpace = "";

    /**
     * ターゲットタグ名
     */
    private String filterTargetTagName = "";

    /**
     * ターゲットレベル
     */
    private int filterTargetLevel = 0;

    private boolean ONLY_TAG_ENDPOINT = false;
    private String XPATH_SEPARATOR = "/";

    private File workDir = new File(SystemProps.get(SystemProps.PROPS_SYS_WORK_PATH, "./work"));

    public XpathFilter() {
    }

    public XpathFilter(String nameSpace, String tagName) {
        this.filterTargetNameSpace = nameSpace;
        this.filterTargetTagName = tagName;
    }

    public void process(Message message) throws ProcessorException {
        if (message == null) {
            ProcessorException e = new XpathFilterException("A message is null.");
            e.fillInStackTrace();
            throw e;
        }

        String outputEncoding = this.outputEncoding;
       
        if (message.getProperties().containsKey(PROPS_OUTPUT_ENCODING))
            outputEncoding = message.getProperties().getProperty(PROPS_OUTPUT_ENCODING);
        if (message.getProperties().containsKey(PROPS_SEPARATOR))
            this.separator = message.getProperties().getProperty(PROPS_SEPARATOR);

        
        try {
            File workFile = File.createTempFile(XpathFilter.class.getName(), ".xml", workDir);
            OutputStreamWriter out = new OutputStreamWriter(new FileOutputStream(workFile), outputEncoding);
            InputStream in = new BufferedInputStream(message.getPayload().getInputStream());
            transform(in, out);
            in.close();
            out.flush();
            out.close();

            StandardPayload payload = new StandardPayload();
            payload.attach(workFile);
            message.setPayload(payload);

        } catch (Exception e) {
            log.error(e.getMessage(), e);
            XpathFilterException ex = new XpathFilterException(e.getMessage());
            ex.fillInStackTrace();
            throw ex;
        }
    }

    protected void transform(InputStream in, OutputStreamWriter out) throws ProcessorException {
        try {
            XMLReader parser = getXMLReader();
            GenericHandler handler = new GenericHandler(out);
            handler.setSeparator(this.separator);
            parser.setContentHandler(handler);
            parser.parse(new InputSource(in));
        } catch (Exception e) {
            log.debug(e.getMessage(), e);
            ProcessorException exception = new XpathFilterException(e.getMessage());
            exception.setStackTrace(e.getStackTrace());
            throw exception;
        }
    }

    class GenericHandler extends DefaultHandler {
        OutputStreamWriter out;
        
        public GenericHandler(OutputStreamWriter out) {
            super();
            this.out = out;
        }

        private void outputStr(String str) {
            try {
                out.write(str);
            } catch (IOException e) {
            }
        }

        private int currTotalTagLevel = -1;

        private int currTotalTagIndex = 0;

        private int currTargetTagLevel = -1;

        private int currTargetTagIndex = 0;

        private int currLevel = -1;

        private int currIndex = 0;
        
        private String separator = XpathFilter.DEFAULT_SEPARATOR;

        public void startElement(String namespaceURI, String localName, String qName, Attributes atts) throws SAXException {
            currTotalTagLevel++;

            if (currTargetTagLevel < 0) {
                if ((filterTargetNameSpace.equals("") || namespaceURI.equals(filterTargetNameSpace))
                        && (filterTargetTagName.equals("") || localName.equals(filterTargetTagName))) {
                    currTargetTagLevel = 0;
                    currTargetTagIndex = 0;
                }
            }
            else {
                currTargetTagLevel++;
            }

            if (currTargetTagLevel >= filterTargetLevel) {
                if (currTargetTagLevel == filterTargetLevel) {
                    currLevel = 0;
                    currIndex = 0;
                }
                else
                    currLevel++;
                startElement2(namespaceURI, localName, qName, atts);
            }
            else {
                // outputStr("<" + qName + ">");
            }
        }

        public void characters(char[] ch, int start, int length) throws SAXException {
            currTotalTagIndex++;
            if (currTargetTagLevel >= 0) {
                currTargetTagIndex++;
                if (currTargetTagLevel >= filterTargetLevel) {
                    currIndex++;
                    characters2(ch, start, length);
                }
            }
            else {
                // outputChars(ch, start, length);
            }
        }

        public void endElement(String namespaceURI, String localName, String qName) throws SAXException {
            currTotalTagLevel--;

            if (currTargetTagLevel >= filterTargetLevel) {
                endElement2(namespaceURI, localName, qName);
                if (currLevel >= 0)
                    currLevel--;
            }
            else {
                // outputStr("</" + qName + ">");
            }
            if (currTargetTagLevel >= 0)
                currTargetTagLevel--;
        }

        private List currQNameTree = new ArrayList();
        private boolean currTagIsEndPoint = false;
        private String currTagVal = "";

        private void startElement2(String namespaceURI, String localName, String qName, Attributes atts) throws SAXException {
            if (currLevel == 0) {
                currQNameTree.clear();
            }
            else if (!ONLY_TAG_ENDPOINT) {
                String val = currTagVal.trim();
                if (val.length() > 0)
                    outputStr(makeXpath(currQNameTree) + separator + val + "\n");
            }
            currQNameTree.add(qName);
            currTagIsEndPoint = true;
            currTagVal = "";
        }

        private void characters2(char[] ch, int start, int length) throws SAXException {
            currTagVal += new String(ch, start, length);
        }

        private String makeXpath(List qNameList) {
            String fullName = "";
            for (int cnt = 0; cnt < qNameList.size(); cnt++) {
                if (cnt > 0)
                    fullName += XPATH_SEPARATOR;
                fullName += (String) qNameList.get(cnt);
            }
            return fullName;
        }

        private void endElement2(String namespaceURI, String localName, String qName) throws SAXException {
            if ((!ONLY_TAG_ENDPOINT) || currTagIsEndPoint) {
                String val = currTagVal.trim();
                if (val.length() > 0)
                    outputStr(makeXpath(currQNameTree) + separator + val + "\n");
            }
            currTagIsEndPoint = false;
            currTagVal = "";
            if (currQNameTree.size() > 0)
                currQNameTree.remove(currQNameTree.size() - 1);
        }

        public String getSeparator() {
            return separator;
        }

        public void setSeparator(String separator) {
            this.separator = separator;
        }
    }

    private XMLReader getXMLReader() throws ParserConfigurationException, SAXException {
        SAXParserFactory factory = SAXParserFactory.newInstance();
        factory.setFeature(Const.SAX_NS_URI, true);
        SAXParser parser = factory.newSAXParser();
        XMLReader xmlReader = parser.getXMLReader();
        xmlReader.setFeature(Const.SAX_NS_URI, true);
        return xmlReader;
    }

    public String getOutputEncoding() {
        return outputEncoding;
    }

    public String getSeparator() {
        return separator;
    }

    public void setOutputEncoding(String outputEncoding) {
        this.outputEncoding = outputEncoding;
    }

    public void setSeparator(String separator) {
        this.separator = separator;
    }
}
