package daruma.bus.processor.xslt;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.List;
import java.util.Properties;

import javax.xml.transform.OutputKeys;
import javax.xml.transform.Result;
import javax.xml.transform.Templates;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.sax.SAXResult;
import javax.xml.transform.sax.SAXSource;
import javax.xml.transform.sax.SAXTransformerFactory;
import javax.xml.transform.sax.TemplatesHandler;
import javax.xml.transform.sax.TransformerHandler;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xml.serializer.OutputPropertiesFactory;
import org.apache.xml.serializer.Serializer;
import org.apache.xml.serializer.SerializerFactory;
import org.xml.sax.InputSource;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;

import application.startup.SystemProps;

import daruma.bus.message.Message;
import daruma.bus.message.StandardPayload;
import daruma.bus.processor.Processor;
import daruma.bus.processor.ProcessorException;
import daruma.bus.processor.csv.Const;

public class XslTransformer implements Processor {
    static private Log log = LogFactory.getLog(XslTransformer.class);

    static public final String PROPS_XSLT_FILE = "xslttransformer.xslt.file";
    static public final String PROPS_XML_ENCODING = "xslttransformer.xml.encoding";
    static public final String PROPS_OMIT_XML_DECLARATION = "xslttransformer.omit.xml.declaration";
    static public final String PROPS_INDENT = "xslttransformer.indent";
    static public final String PROPS_FEATURE_NAMESAPCE = "xslttransformer.feature.namespace";
    static public final String PROPS_FEATURE_NAMESAPCE_PREFIXES = "xslttransformer.feature.namespace.prefixes";

    static public final String SAX_FEATURE_NAMESAPCE_URI = Const.SAX_NS_URI;
    static public final String SAX_FEATURE_NAMESAPCE_PREFIX_URI = Const.SAX_PRE_URI;

    static public final String DEFAULT_XML_ENCODING = SystemProps.get(SystemProps.DEFAULT_SYS_ENCODING_INTERNAL, "UTF-8");
    static public final String DEFAULT_OMIT_XML_DECLARATION = SystemProps.get(PROPS_OMIT_XML_DECLARATION, "no");
    static public final String DEFAULT_INDENT = SystemProps.get(PROPS_INDENT, "yes");

    private String xsltFile;
    private SAXTransformerFactory saxTransFactory;
    private Properties outputProperties = OutputPropertiesFactory.getDefaultMethodProperties("xml");
    private List parameterList;

    private String xmlEncoding = DEFAULT_XML_ENCODING;
    private String omitXmlDeclaration = DEFAULT_OMIT_XML_DECLARATION;
    private String indent = DEFAULT_INDENT;

    private File workDir = new File(SystemProps.get(SystemProps.PROPS_SYS_WORK_PATH, "./work"));

    /**
     * 
     * 
     */
    public XslTransformer() {

        TransformerFactory transformerFactory = TransformerFactory.newInstance();
        if (transformerFactory.getFeature(SAXSource.FEATURE) && transformerFactory.getFeature(SAXResult.FEATURE)) {
            this.saxTransFactory = ((SAXTransformerFactory) transformerFactory);
        }

    }

    public void process(Message message) throws ProcessorException {

        if (message == null) {
            ProcessorException e = new XslTransformerException("A message is null.");
            e.fillInStackTrace();
            throw e;
        }

        try {
            String xsltFile = this.xsltFile;
            String xmlEncoding = this.xmlEncoding;
            String omitXmlDeclaration = this.omitXmlDeclaration;
            String indent = this.indent;

            if (message.getProperties().containsKey(PROPS_XSLT_FILE))
                xsltFile = message.getProperties().getProperty(PROPS_XSLT_FILE);
            if (message.getProperties().containsKey(PROPS_XML_ENCODING))
                xmlEncoding = message.getProperties().getProperty(PROPS_XML_ENCODING);
            if (message.getProperties().containsKey(PROPS_INDENT))
                indent = message.getProperties().getProperty(PROPS_INDENT);
            if (message.getProperties().containsKey(PROPS_OMIT_XML_DECLARATION))
                omitXmlDeclaration = message.getProperties().getProperty(PROPS_OMIT_XML_DECLARATION);

            this.outputProperties.setProperty(OutputKeys.ENCODING, xmlEncoding);
            this.outputProperties.setProperty(OutputKeys.OMIT_XML_DECLARATION, omitXmlDeclaration);
            this.outputProperties.setProperty(OutputKeys.INDENT, indent);

            File workFile = File.createTempFile(XslTransformer.class.getName(), ".xml", workDir);
            OutputStream out = new BufferedOutputStream(new FileOutputStream(workFile));
            InputStream in = new BufferedInputStream(message.getPayload().getInputStream());
            transform(xsltFile, in, out);
            out.close();
            in.close();
            
            StandardPayload payload = new StandardPayload();
            payload.attach(workFile);
            message.setPayload(payload);

        } catch (Exception e) {
            log.debug(e.getMessage(), e);
            XslTransformerException ex = new XslTransformerException(e.getMessage());
            ex.fillInStackTrace();
            throw ex;
        }

    }

    public void process(InputStream in, OutputStream out) throws ProcessorException {

        String xmlEncoding = this.xmlEncoding;
        String omitXmlDeclaration = this.omitXmlDeclaration;
        String indent = this.indent;

        this.outputProperties.setProperty(OutputKeys.ENCODING, xmlEncoding);
        this.outputProperties.setProperty(OutputKeys.OMIT_XML_DECLARATION, omitXmlDeclaration);
        this.outputProperties.setProperty(OutputKeys.INDENT, indent);

        transform(this.xsltFile, in, out);
    }
    
    protected void transform(String xsltFile, InputStream in, OutputStream out)  throws ProcessorException {
    	/*
    	 * xalanの不具合によって　jdkが1.5以前の場合 SAX方式使えません 
    	 */
    	String jdkVer = System.getProperty("java.version");
    	boolean jdk15 = (jdkVer.compareTo("1.5") >= 0);
    	if(jdk15) transformSax(xsltFile, in, out);
    	else transformDom(xsltFile, in, out);
    }

    protected void transformDom(String xsltFile, InputStream in, OutputStream out) {
    	try
		 {
			 TransformerFactory tFactory = TransformerFactory.newInstance();
			 Transformer transformer = tFactory.newTransformer(new StreamSource(xsltFile));
			 Result result = new StreamResult(out);
			 transformer.setOutputProperties(this.outputProperties);
			 if(parameterList != null)
			 {
				 for(int cnt = 0; cnt < parameterList.size(); cnt+= 2)
				 {
					 String name = (String) parameterList.get(cnt);
					 String value = (String) parameterList.get(cnt + 1);
					 transformer.setParameter(name, value);
				 }
			 }
			 transformer.transform(new StreamSource(in), result);
		 } catch (Exception e) {
			 if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
					e.printStackTrace();
			 }
	     }
	 }
    
    protected void transformSax(String xsltFile, InputStream in, OutputStream out) throws ProcessorException {
        try {
            TemplatesHandler templatesHandler = saxTransFactory.newTemplatesHandler();
            XMLReader reader = XMLReaderFactory.createXMLReader();
            reader.setContentHandler(templatesHandler);

            InputStream xsltFileSource = new FileInputStream(xsltFile);
            reader.parse(new InputSource(xsltFileSource));

            Templates templates = templatesHandler.getTemplates();
            TransformerHandler handler = saxTransFactory.newTransformerHandler(templates);
            if(parameterList != null)
			 {
				 for(int cnt = 0; cnt < parameterList.size(); cnt+= 2)
				 {
					 String name = (String) parameterList.get(cnt);
					 String value = (String) parameterList.get(cnt + 1);
					 handler.getTransformer().setParameter(name, value);
				 }
			 }
            reader.setContentHandler(handler);
            reader.setProperty(Const.SAX_HANDLER_URI, handler);

            java.util.Properties xmlProps = this.outputProperties;
            Serializer serializer = SerializerFactory.getSerializer(xmlProps);
            serializer.setOutputStream(out);

            Result result = new SAXResult(serializer.asContentHandler());
            handler.setResult(result);

            reader.parse(new InputSource(in));

        } catch (Exception e) {
            log.debug(e.getMessage(), e);
            ProcessorException exception = new XslTransformerException(e.getMessage());
            exception.setStackTrace(e.getStackTrace());
            throw exception;

        }

    }

    public String getXsltFile() {
        return xsltFile;
    }

    public void setXsltFile(String xsltFile) {
        this.xsltFile = xsltFile;
    }

    public Properties getOutputProperties() {
        return outputProperties;
    }

    public void setOutputProperties(Properties outputProperties) {
        this.outputProperties = outputProperties;
    }

    public String getIndent() {
        return indent;
    }

    public void setIndent(String indent) {
        this.indent = indent;
    }

    public String getOmitXmlDeclaration() {
        return omitXmlDeclaration;
    }

    public void setOmitXmlDeclaration(String omitXmlDeclaration) {
        this.omitXmlDeclaration = omitXmlDeclaration;
    }

    public String getXmlEncoding() {
        return xmlEncoding;
    }

    public void setXmlEncoding(String xmlEncoding) {
        this.xmlEncoding = xmlEncoding;
    }

	/**
	 * parameterListを取得する
	 * @return Returns the parameterList.
	 */
	public List getParameterList() {
		return parameterList;
	}

	/**
	 * parameterListをセットする
	 * @param parameterList The parameterList to set.
	 */
	public void setParameterList(List parameterList) {
		this.parameterList = parameterList;
	}

    
}
