package daruma.main;

import java.io.File;
import java.util.ArrayList;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import application.runtime.CommandLineHandler;
import application.runtime.StandardApplication;
import application.runtime.SystemContext;
import application.startup.SystemProps;
import daruma.bus.message.StandardMessage;
import daruma.bus.message.StandardPayload;
import daruma.bus.processor.FileWriter;
import daruma.bus.processor.csv.CSVFileReader;
import daruma.bus.processor.csv.MispSchemaAccess;
import daruma.bus.processor.csv.MispSchemaCx;
import daruma.bus.processor.csv.MispSchemaCxAccess;
import daruma.bus.processor.misp.BuiltinFilter;
import daruma.bus.processor.misp.MispSoapClient;
import daruma.bus.processor.misp.XpathFilter;
import daruma.bus.processor.xslt.XslTransformer;
import daruma.common.CommonUtil;

public class CsvInsert extends StandardApplication {

	static protected File workDir = new File(System
			.getProperty(SystemProps.PROPS_SYS_WORK_PATH));

	static private Log log = LogFactory.getLog(CsvInsert.class);

	static public final String PROPS_RETRIEVE_SOAP_BODY_XSLT_TEMPLATE = "csvinsert.retrieve.soap.body.xslt.template";

	static public final String DEFAULT_RETRIEVE_SOAP_BODY_XSLT_TEMPLATE = System
			.getProperty(PROPS_RETRIEVE_SOAP_BODY_XSLT_TEMPLATE, "");


	protected int onStart(String[] parameters) {
		boolean rc = execute(parameters);
		return (rc ? 0 : 1);
	}

	/**
	 * CSVファイルからデータをインサートします。
	 * 
	 * @param parameters
	 */
	public boolean execute(String[] parameters) {
		boolean rc = false;
		// コマンドライン処理クラスの取得
		CommandLineHandler optHandler = (CommandLineHandler) SystemContext
				.getBean("commandLineHelper");
		CommandLine cmd = optHandler.getCommandLineObject(parameters);
		if (cmd == null) {
			// 2006,12,10 hayashi
			System.out.println("オプションの設定が不正です。");
			optHandler.printHelp();
			return rc;
		}
		
		try {
			
			//  2006,12,25 hayashi デバッグモードチェック
			if (cmd.hasOption("debug")) {
				System.setProperty("DEFAULT_DEBUG_MODE","true");
			}
			
			// ホストのチェック
			if (!cmd.hasOption("host")) {
				// 2006,12,10 hayashi
				System.out.println("ホストの設定が不正です。");
				optHandler.printHelp();
				return rc;
			}

			if (!cmd.hasOption("csvcfg")) {
				// 2006,12,10 hayashi
				System.out.println("configファイルの設定が不正です。");
				optHandler.printHelp();
				return rc;
			}

			// cxvcfgファイルをロードする
			MispSchemaCxAccess cxXml = new MispSchemaCxAccess();
			MispSchemaCx schemaCx = cxXml
				.loadSchemaCxFromXml(cmd.getOptionValue("csvcfg"),
							MispSchemaAccess.DFS_DFT_CHARSET);
			
			if (cmd.hasOption("loop")) {
				rc = processLoop(cmd, schemaCx);
			} else {
				// 入力ファイルのチェック
				File inFile = null;
				if (cmd.hasOption("in")) {
					inFile = new File(cmd.getOptionValue("in"));
				} else if (System.in.available() > 0) {
				} else {
					// 2006,12,10 hayashi
					System.out.println("入力ファイルの設定が不正です。");
					optHandler.printHelp();
					return rc;
				}

				String outFile = null;
				if (cmd.hasOption("out")) {
					outFile = cmd.getOptionValue("out");
				}
				rc = processFile(cmd, schemaCx, inFile, outFile);
			}
			
		}catch (Exception e) {
			// 2006,12,10 hayashi
			System.out.println("ホストの接続またはファイルの読み込み、書き込みに失敗しました。");
			if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
				e.printStackTrace();
			}
		}
		return rc;
	}

	private boolean processLoop(CommandLine cmd, MispSchemaCx schemaCx) {
		boolean rc = false;
		try {
			int interval = 10;
			String inDirPath = "./data";
			String inPattern = ".*.csv";
			String trashDirPath = "./trash";

			if (cmd.hasOption("interval")) {
				try {
					interval = Integer.parseInt(cmd.getOptionValue("interval"));
				} catch (Exception e) {
					log.error("intervalのパラメータ不正");
					return rc;
				}
			}

			if (cmd.hasOption("inDir"))
				inDirPath = cmd.getOptionValue("inDir");
			if (cmd.hasOption("inPattern"))
				inPattern = cmd.getOptionValue("inPattern");
			if (cmd.hasOption("trash"))
				trashDirPath = cmd.getOptionValue("trash");

			File inDir;
			try {
				inDir = new File(inDirPath);
			} catch (Exception e) {
				log.error("フォルダ「" + inDirPath + "」が存在しません");
				return rc;
			}

			Pattern pattern;
			try {
				pattern = Pattern.compile(inPattern);
			} catch (Exception e) {
				log.error("パターン不正");
				return rc;
			}

			File trashDir;
			String resultDirPath = trashDirPath + "/result";
			File resultDir;
			String errorDirPath = trashDirPath + "/error";
			File errorDir;
			try {
				trashDir = new File(trashDirPath);
				resultDir = new File(resultDirPath);
				errorDir = new File(errorDirPath);

				if (!trashDir.exists()) {
					trashDir.mkdirs();
				}
				if (!resultDir.exists()) {
					resultDir.mkdirs();
				}
				if (!errorDir.exists()) {
					errorDir.mkdirs();
				}
			} catch (Exception e) {
				log.error("フォルダ「" + trashDirPath + "」が存在しません");
				return rc;
			}

			System.out.println("ループモード開始(時間間隔=" + interval + "秒)");
			int cntTime = 0;
			while (true) {
				cntTime++;
				// System.out.println(cntTime + " 回目開始");
				File[] fileList = inDir.listFiles();
				if (fileList != null) {
					for (int cnt = 0; cnt < fileList.length; cnt++) {
						File currFile = fileList[cnt];
						Matcher matcher = pattern.matcher(currFile.getName());
						if (matcher.find()) {
							System.out.println(currFile.getAbsolutePath()
									+ "を処理...");
							rc = processFile(cmd, schemaCx, currFile,
									resultDirPath + "/" + currFile.getName()
											+ ".result");
							System.out.println(rc ? "処理成功" : "処理エラー");
							String dstFilePath = rc ? (trashDirPath + "/" + currFile
									.getName())
									: (errorDirPath + "/" + currFile.getName());
							File dstFile = new File(dstFilePath);
							dstFile.delete();
							currFile.renameTo(dstFile);
						}
					}
				}
				Thread.sleep(interval * 1000);
			}
		} catch (Exception e) {
			log.error("CsvInsert ProcessLoop error");
			if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
				e.printStackTrace();
			}
		}
		return rc;
	}

	private boolean processFile(CommandLine cmd, MispSchemaCx schemaCx,
			File inFile, String outFile) {
		boolean rc = false;
		try {
			// メッセージ作成
			StandardMessage message = new StandardMessage();

			// Payloadファイル設定
			StandardPayload payload = new StandardPayload();
			message.setPayload(payload);
			if (inFile == null) {
				inFile = CommonUtil.loadFileFromStdin(workDir);
				payload.attach(inFile);
			} else
				payload.load(inFile);

			// デバッグモード設定
			message.setVerboseMode(cmd.hasOption("debug"));

			// 各種プロパティ設定
			Properties properties = new Properties();
			message.setProperties(properties);
			// Mispサーバー
			properties.setProperty(MispSoapClient.PROPS_SERVER_HOST, cmd
					.getOptionValue("host"));
			properties.setProperty(MispSoapClient.PROPS_SERVER_PORT, "5050");
			// 出力ファイル
			if (outFile != null)
				properties.setProperty(FileWriter.PROPS_OUTPUT_PATH, outFile);
			else
				properties.setProperty(FileWriter.PROPS_OUTPUT_PATH,
						FileWriter.VALUE_OUTPUT_PATH_STDOUT);
			// 入力CSVファイル文字コード
			if (cmd.hasOption("encode")) {
				message.getProperties().setProperty(
						CSVFileReader.PROPS_INPUT_ENCODING,
						cmd.getOptionValue("encode"));
			}

			MispSchemaCxAccess cxXml = new MispSchemaCxAccess();

			// XSLTプロセサ
			XslTransformer transformer = new XslTransformer();
			// SOAPクライアントプロセサ
			MispSoapClient soapClient = new MispSoapClient();
			// CSVファイルリーダープロセサ
			CSVFileReader csvReader = new CSVFileReader();
			// ファイル出力プロセサ
			FileWriter fileWriter = new FileWriter();
			// GmlProperty変換プロセサ
			BuiltinFilter builtinFilter = new BuiltinFilter(
					BuiltinFilter.MODE_CSV2XML);
			XpathFilter xpathFilter = new XpathFilter();

			// CSVファイル取り込む
			File csv2xmlXstlFile = File.createTempFile(CsvInsert.class
					.getName(), "_csv2xml.xslt", workDir);
			File delNullTagXsltFile = File.createTempFile(CsvInsert.class
					.getName(), "_delnull.xslt", workDir);

			if (cxXml.createXsltCsv2Xml(schemaCx, csv2xmlXstlFile
					.getAbsolutePath(), "UTF-8", "UTF-8")) {
				transformer.setXsltFile(csv2xmlXstlFile.getCanonicalPath());
				message.setProcessorList(new ArrayList());
				message.getProcessorList().add(csvReader);
				message.getProcessorList().add(transformer);
				message.send();

				// データ処理(空タグ削除、gmlPropertyデータ変換)して 送信する
				cxXml.createXsltDeleteNullTag(schemaCx, delNullTagXsltFile
						.getAbsolutePath(), "UTF-8", "UTF-8");
				transformer.setXsltFile(delNullTagXsltFile.getCanonicalPath());
				message.setProcessorList(new ArrayList());
				message.getProcessorList().add(transformer);
				if (cmd.hasOption("attatchDir"))
					builtinFilter.setAttatchDir(cmd
							.getOptionValue("attatchDir"));
				builtinFilter.setSchemaCx(schemaCx);
				message.getProcessorList().add(builtinFilter);
				message.getProcessorList().add(soapClient);
				message.send();

				// SOAPヘッダーを削除する
				message.setProcessorList(new ArrayList());
				if (cmd.hasOption("onlybody")) {
					transformer.setXsltFile(System.getProperty(
							PROPS_RETRIEVE_SOAP_BODY_XSLT_TEMPLATE,
							DEFAULT_RETRIEVE_SOAP_BODY_XSLT_TEMPLATE));
					message.getProcessorList().add(transformer);
				}
				if (cmd.hasOption("xpath"))
					message.getProcessorList().add(xpathFilter);

				message.getProcessorList().add(fileWriter);
				message.send();
			} else {
				log.error("CSVCFGファイル不正");
			}
			if (!message.isVerboseMode() && message.getPayload() != null)
				message.getPayload().dispose();
			if (!message.isVerboseMode()) {
				csv2xmlXstlFile.delete();
				delNullTagXsltFile.delete();
			}
			rc = true;
		} catch (Exception e) {
			log.error("CsvInsert ProcessFile error");
			if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
				e.printStackTrace();
			}
		}
		return rc;
	}
}
