package daruma.main;

import java.io.File;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.ArrayList;
import java.util.Properties;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import application.runtime.CommandLineHandler;
import application.runtime.StandardApplication;
import application.runtime.SystemContext;
import application.startup.SystemProps;
import daruma.bus.message.MessageException;
import daruma.bus.message.StandardMessage;
import daruma.bus.message.StandardPayload;
import daruma.bus.processor.FileWriter;
import daruma.bus.processor.ProcessorException;
import daruma.bus.processor.csv.Const;
import daruma.bus.processor.misp.FeatureTypeFilter;
import daruma.bus.processor.misp.MispSoapClient;
import daruma.bus.processor.xslt.XslTransformer;
import daruma.common.CommonUtil;
import daruma.common.CommonUtil.FeatureName;

public class DescribeFeatureType extends StandardApplication {
	static private Log log = LogFactory.getLog(DescribeFeatureType.class);

	static protected File workDir = new File(System
			.getProperty(SystemProps.PROPS_SYS_WORK_PATH));

	static public final String PROPS_RETRIEVE_SCHEMA_XSLT_TEMPLATE = "describefeaturetype.retrieve.schema.xslt.template";

	static public final String DEFAULT_RETRIEVE_SCHEMA_XSLT_TEMPLATE = System
			.getProperty(PROPS_RETRIEVE_SCHEMA_XSLT_TEMPLATE, "");


	protected int onStart(String[] parameters) {
		boolean rc = execute(parameters);
		return (rc ? 0 : 1);
	}

	/**
	 * 指定したFeatureTypeのスキーマを表示します。
	 * 
	 * @param parameters
	 */
	public boolean execute(String[] parameters) {
		boolean rc = false;
		// コマンドライン処理クラスの取得
		CommandLineHandler optHandler = (CommandLineHandler) SystemContext
				.getBean("commandLineHelper");
		CommandLine cmd = optHandler.getCommandLineObject(parameters);
		if (cmd == null) {
			// 2006,12,10 hayashi
			System.out.println("オプションの設定が不正です。");
			optHandler.printHelp();
			return rc;
		}
		try {
			StandardMessage message = new StandardMessage();
			StandardPayload payload = new StandardPayload();
			message.setPayload(payload);
			message.setProperties(new Properties());
			//message.setVerboseMode(cmd.hasOption("debug"));
			Properties properties = message.getProperties();
			
			// delete hasOption 			
			int len = parameters.length;
			int i;
			for(i = 0; i < len ; i++){
				if("-debug".equals(parameters[i])){
					System.setProperty("DEFAULT_DEBUG_MODE","true");
					message.setVerboseMode(true);
				}else{
					message.setVerboseMode(false);
				}
			}

			// ホスト
			if (cmd.hasOption("host")) {
				properties.setProperty(MispSoapClient.PROPS_SERVER_HOST, cmd
						.getOptionValue("host"));
				properties
						.setProperty(MispSoapClient.PROPS_SERVER_PORT, "5050");
			} else {
				System.out.println("ホストの設定が不正です。");
				optHandler.printHelp();
				return rc;
			}

			// 出力ファイルのチェック
			if (cmd.hasOption("out")) {
				properties.setProperty(FileWriter.PROPS_OUTPUT_PATH, cmd
						.getOptionValue("out"));
			} else {
				properties.setProperty(FileWriter.PROPS_OUTPUT_PATH,
						FileWriter.VALUE_OUTPUT_PATH_STDOUT);
			}
			message.getProperties().setProperty(
					FileWriter.PROPS_INPUT_ENCODING, "UTF-8");
			message.getProperties().setProperty(
					FileWriter.PROPS_OUTPUT_ENCODING,
					cmd.hasOption("encode") ? cmd.getOptionValue("encode")
							: "UTF-8");

			if (cmd.getArgs().length < 1) {
				// 2006,12,10 hayashi
				System.out.println("オプションの設定が不正です。");
				optHandler.printHelp();
				return rc;
			}

			// FeatureType
			String name = cmd.getArgs()[0];
			String targetNS = null;
			if (cmd.getArgs().length > 1)
				targetNS = cmd.getArgs()[1];
			FeatureName featureName = CommonUtil.getFeatureName(name, targetNS);
			if (featureName == null) {
				log.error("feature名形式不正");
				return rc;
			}

			// DescribeFeature リクエストの作成
			File requestXmlFile = createDescribeFeatureRequestXml(featureName);
			payload.attach(requestXmlFile);

			// プロセッサ
			MispSoapClient soapClient = new MispSoapClient();
			XslTransformer transformer = new XslTransformer();
			FileWriter fileWriter = new FileWriter();
			FeatureTypeFilter featureTypeFilter = new FeatureTypeFilter();

			// XSLT
			transformer.setXsltFile(System.getProperty(
					PROPS_RETRIEVE_SCHEMA_XSLT_TEMPLATE,
					DEFAULT_RETRIEVE_SCHEMA_XSLT_TEMPLATE));

			// 処理
			message.setProcessorList(new ArrayList());
			message.getProcessorList().add(soapClient);
			if (cmd.hasOption("xpath"))
				message.getProcessorList().add(featureTypeFilter);
			else
				message.getProcessorList().add(transformer);
			message.getProcessorList().add(fileWriter);

			message.send();

			if (!message.isVerboseMode() && message.getPayload() != null)
				message.getPayload().dispose();
			if (!message.isVerboseMode())
				requestXmlFile.delete();

			rc = true;

		} catch (Exception e) {
			// 2006,12,10 hayashi
			System.out.println("ホストの接続またはファイルの読み込み、書き込みに失敗しました。");
			if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
				e.printStackTrace();
			}
		}
		return rc;
	}

	/**
	 * リクエスト用XMLを生成します。
	 * 
	 * @param message
	 * @param featureType
	 * @return ファイルを返す
	 * @throws MessageException
	 * @throws ProcessorException
	 * @throws UnsupportedEncodingException
	 * @throws IOException
	 */
	protected File createDescribeFeatureRequestXml(FeatureName featureName)
			throws MessageException, ProcessorException,
			UnsupportedEncodingException, IOException {

		// プロセッサ
		String internalCharset = System
				.getProperty(SystemProps.PROPS_SYS_INTERNAL_ENCODING);
		String msg = "<?xml version='1.0' encoding='" + internalCharset
				+ "'?><" + Const.NS_MISP_PRE + ":DescribeFeatureType "
				+ Const.NS_XMLNS_PRE + ":" + Const.NS_MISP_PRE + "='"
				+ Const.NS_MISP_URI + "'><" + Const.NS_MISP_PRE + ":TypeName "
				+ Const.NS_XMLNS_PRE + ":ns=\"" + featureName.getTargetNS()
				+ "\">ns:" + featureName.getTypeName() + "</"
				+ Const.NS_MISP_PRE + ":TypeName></" + Const.NS_MISP_PRE
				+ ":DescribeFeatureType>";
		File requestFile = CommonUtil.loadFileFromBuffer(msg
				.getBytes(internalCharset), workDir);

		return requestFile;
	}

}
