package daruma.main;

import java.io.File;
import java.util.ArrayList;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import application.runtime.CommandLineHandler;
import application.runtime.StandardApplication;
import application.runtime.SystemContext;
import application.startup.SystemProps;
import daruma.bus.message.StandardMessage;
import daruma.bus.message.StandardPayload;
import daruma.bus.processor.FileWriter;
import daruma.bus.processor.misp.MispSoapClient;
import daruma.bus.processor.misp.XpathFilter;
import daruma.bus.processor.xslt.XslTransformer;
import daruma.common.CommonUtil;

public class XmlInsert extends StandardApplication {

	static protected File workDir = new File(System
			.getProperty(SystemProps.PROPS_SYS_WORK_PATH));

	static public final String PROPS_XML2WFS_XSLT_TEMPLATE = "xmlinsert.xml2wfs.xslt.template";

	static public final String DEFAULT_XML2WFS_XSLT_TEMPLATE = System
			.getProperty(PROPS_XML2WFS_XSLT_TEMPLATE, "");

	static private Log log = LogFactory.getLog(XmlInsert.class);


	protected int onStart(String[] parameters) {
		boolean rc = execute(parameters);
		return (rc ? 0 : 1);
	}

	/**
	 * XMLファイルからデータをインサートします。
	 * 
	 * @param parameters
	 */
	public boolean execute(String[] parameters) {
		boolean rc = false;
		// コマンドライン処理クラスの取得
		CommandLineHandler optHandler = (CommandLineHandler) SystemContext
				.getBean("commandLineHelper");
		CommandLine cmd = optHandler.getCommandLineObject(parameters);
		if (cmd == null) {
			// 2006,12,10 hayashi
			System.out.println("オプションの設定が不正です。");
			optHandler.printHelp();
			return rc;
		}
		try {
			
			//  2006,12,25 hayashi デバッグモードチェック
			if (cmd.hasOption("debug")) {
				System.setProperty("DEFAULT_DEBUG_MODE","true");
			}
			
			// ホストのチェック
			if (!cmd.hasOption("host")) {
				// 2006,12,10 hayashi
				System.out.println("ホストの設定が不正です。");
				optHandler.printHelp();
				return rc;
			}

			if (cmd.hasOption("loop")) {
				rc = processLoop(cmd);
			} else {
				// 入力ファイルのチェック
				File inFile = null;
				if (cmd.hasOption("in")) {
					inFile = new File(cmd.getOptionValue("in"));
				} else if (System.in.available() > 0) {
				} else {
					// 2006,12,10 hayashi
					System.out.println("入力ファイルの設定が不正です。");
					optHandler.printHelp();
					return rc;
				}

				String outFile = null;
				if (cmd.hasOption("out")) {
					outFile = cmd.getOptionValue("out");
				}
				rc = processFile(cmd, inFile, outFile);

			}

		} catch (Exception e) {
			// 2006,12,10 hayashi
			System.out.println("ホストの接続またはファイルの読み込み、書き込みに失敗しました。");
			if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
				e.printStackTrace();
			}
		}
		return rc;
	}

	private boolean processLoop(CommandLine cmd) {
		boolean rc = false;
		try {
			int interval = 10;
			String inDirPath = "./data";
			String inPattern = ".*.xml";
			String trashDirPath = "./trash";

			if (cmd.hasOption("interval")) {
				try {
					interval = Integer.parseInt(cmd.getOptionValue("interval"));
				} catch (Exception e) {
					log.error("intervalのパラメータ不正");
					return rc;
				}
			}

			if (cmd.hasOption("inDir"))
				inDirPath = cmd.getOptionValue("inDir");
			if (cmd.hasOption("inPattern"))
				inPattern = cmd.getOptionValue("inPattern");
			if (cmd.hasOption("trash"))
				trashDirPath = cmd.getOptionValue("trash");

			File inDir;
			try {
				inDir = new File(inDirPath);
			} catch (Exception e) {
				log.error("フォルダ「" + inDirPath + "」が存在しません");
				return rc;
			}

			Pattern pattern;
			try {
				pattern = Pattern.compile(inPattern);
			} catch (Exception e) {
				log.error("パターン不正");
				return rc;
			}

			File trashDir;
			String resultDirPath = trashDirPath + "/result";
			File resultDir;
			String errorDirPath = trashDirPath + "/error";
			File errorDir;
			try {
				trashDir = new File(trashDirPath);
				resultDir = new File(resultDirPath);
				errorDir = new File(errorDirPath);

				if (!trashDir.exists()) {
					trashDir.mkdirs();
				}
				if (!resultDir.exists()) {
					resultDir.mkdirs();
				}
				if (!errorDir.exists()) {
					errorDir.mkdirs();
				}
			} catch (Exception e) {
				log.error("フォルダ「" + trashDirPath + "」が存在しません");
				return rc;
			}

			System.out.println("ループモード開始(時間間隔=" + interval + "秒)");
			int cntTime = 0;
			while (true) {
				cntTime++;
				// System.out.println(cntTime + " 回目開始");
				File[] fileList = inDir.listFiles();
				if (fileList != null) {
					for (int cnt = 0; cnt < fileList.length; cnt++) {
						File currFile = fileList[cnt];
						Matcher matcher = pattern.matcher(currFile.getName());
						if (matcher.find()) {
							System.out.println(currFile.getAbsolutePath()
									+ "を処理...");
							rc = processFile(cmd, currFile, resultDirPath + "/"
									+ currFile.getName() + ".result");
							System.out.println(rc ? "処理成功" : "処理エラー");
							String dstFilePath = rc ? (trashDirPath + "/" + currFile
									.getName())
									: (errorDirPath + "/" + currFile.getName());
							File dstFile = new File(dstFilePath);
							dstFile.delete();
							currFile.renameTo(dstFile);
						}
					}
				}
				Thread.sleep(interval * 1000);
			}
		} catch (Exception e) {
			log.error("XmlInsert ProcessLoop error");
			if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
				e.printStackTrace();
			}
		}
		return rc;
	}

	private boolean processFile(CommandLine cmd, File inFile, String outFile) {
		boolean rc = false;
		try {

			// メッセージ作成
			StandardMessage message = new StandardMessage();

			// Payloadファイル設定
			StandardPayload payload = new StandardPayload();
			message.setPayload(payload);
			if (inFile == null) {
				inFile = CommonUtil.loadFileFromStdin(workDir);
				payload.attach(inFile);
			} else
				payload.load(inFile);

			// デバッグモード設定
			message.setVerboseMode(cmd.hasOption("debug"));

			// 各種プロパティ設定
			Properties properties = new Properties();
			message.setProperties(properties);
			// Mispサーバー
			properties.setProperty(MispSoapClient.PROPS_SERVER_HOST, cmd
					.getOptionValue("host"));
			properties.setProperty(MispSoapClient.PROPS_SERVER_PORT, "5050");
			// 出力ファイル
			if (outFile != null)
				properties.setProperty(FileWriter.PROPS_OUTPUT_PATH, outFile);
			else
				properties.setProperty(FileWriter.PROPS_OUTPUT_PATH,
						FileWriter.VALUE_OUTPUT_PATH_STDOUT);
			// プロセッサ
			XslTransformer transformer = new XslTransformer();
			MispSoapClient soapClient = new MispSoapClient();
			FileWriter fileWriter = new FileWriter();
			XpathFilter xpathFilter = new XpathFilter();

			// XSLT
			transformer
					.setXsltFile(System.getProperty(
							PROPS_XML2WFS_XSLT_TEMPLATE,
							DEFAULT_XML2WFS_XSLT_TEMPLATE));

			message.setProcessorList(new ArrayList());
			message.getProcessorList().add(transformer);
			message.getProcessorList().add(soapClient);
			message.send();

			message.setProcessorList(new ArrayList());
			if (cmd.hasOption("onlybody"))
				message.getProcessorList().add(transformer);
			if (cmd.hasOption("xpath"))
				message.getProcessorList().add(xpathFilter);
			message.getProcessorList().add(fileWriter);
			message.send();

			if (!message.isVerboseMode() && message.getPayload() != null)
				message.getPayload().dispose();
			rc = true;
		} catch (Exception e) {
			log.error("XmlInsert ProcessFile error");
			if(System.getProperty("DEFAULT_DEBUG_MODE") == "true"){
				e.printStackTrace();
			}
		}
		return rc;
	}
}
